#!/bin/env perl

use strict;
use warnings;
use Getopt::Long;

my $DESCRIPTION =
  "\nGroupe avec un tiret dans un corpus les termes formes par plusieurs "
  ."unites donnees dans une liste de termes. Les termes doivent comporter "
  ."au moins deux unites. Lors d'un conflit entre plusieurs termes c'est "
  ."le terme comportant le plus d'unites qui l'emporte et en cas d'egalite "
  ."le premier dans la liste des termes. L'etiquette de la derniere "
  ."unite d'un terme est utilisee comme etiquette du groupement.\n"
  ."Thomas Heitz - 2005 - LRI Paris XI";

my $USAGE =
  "Usage : perl $0 -l liste [-s separateur] [-e etiquette] [-ec] [-t tiret] [-tm taille] < corpus > corpus_resultat\n\n"
  ."Exemple : perl $0 -l liste.txt -s '/' -e \"-X\" -ec -t '-' -tm 4 < entree.txt > sortie.txt\n\n"
  ."-l, --liste=liste :\n"
  ."  liste contenant les unites des termes sans etiquettes et separees "
  ."par des espaces, un terme par ligne\n"
  ."-s, --separateur=separateur :\n"
  ."  separateur entre l'unite et l'etiquette, s'il est egal a une chaine "
  ."vide alors le texte d'entree n'est pas etiquete, ce qui est le cas "
  ."par defaut\n"
  ."-e, --etiquette=etiquette :\n"
  ."  etiquette ajoutee a celle de la dernier unite dans le groupement, "
  ."permet aussi de rajouter une etiquette dans un texte non etiquete\n"
  ."-ec, --etiquettes_completes :\n"
  ."  si cette option est presente alors etiquete le terme avec toutes "
  ."les etiquettes de ses unites et pas seulement l'etiquette de la "
  ."derniere unite comme cela a lieu par defaut\n"
  ."-t, --tiret=tiret :\n"
  ."  expression qui sert de tiret entre les unites du terme groupe, par "
  ."defaut le tiret '-' est utilise\n"
  ."-tm, --taille_min=taille :\n"
  ."  nombre minimum d'unites d'un terme pour ne pas tenir de compte de "
  ."la casse lors d'une comparaison entre termes du texte et de la liste, "
  ."par defaut 2, pour toujours tenir compte de la casse, utilisez 0\n"
  ."--help :\n"
  ."  this help in english";

my $DESCRIPTION_EN =
  "\nGroup with a hyphen in a corpus the terms formed by several "
  ."units given in a terms list. Terms must consist of at less two units. "
  ."In case of conflict between terms that's the term consisting of the "
  ."most units that wins and in case of equality the first in the terms "
  ."list. The tag of last unit of a term is used as tag of the grouping.\n"
  ."Thomas Heitz - 2005 - LRI Paris XI";

my $USAGE_EN =
  "Usage : perl $0 -l list [-s separator] [-e tag] [-ft] [-h hyphen] [-mu minimum] < corpus > result_corpus\n\n"
  ."Example : perl $0 -l list.txt -s '/' -e \"-X\" -ft -h '-' -mu 4 < input.txt > output.txt\n\n"
  ."-l, --list=list :\n"
  ."  list containing terms units without tags and separated with spaces, "
  ."one term per line\n"
  ."-s, --separator=separator :\n"
  ."  separator between unit and tag, if equal to a empty string "
  ."then the input text isn't tagged, that is the default case\n"
  ."-e, --tag=tag :\n"
  ."  tag added to this of the last unit in the grouping, "
  ."allow to add a tag in a non tagged text too\n"
  ."-ft, --full_tags :\n"
  ."  if this option is enabled then tag terms with all tags of its units "
  ."and not only tag of the last unit as the default case\n"
  ."-h, --hyphen=hyphen :\n"
  ."  expression that is used for hyphenate units of grouping terms, "
  ."by default hyphen '-' is used\n"
  ."-mu, --min_units=minimum :\n"
  ."  minimum of units in a term to be case insensitive "
  ."during a comparison between terms of text and of list, "
  ."by default 2, for always be case sensitive, use 0\n"
  ."--aide :\n"
  ."  cette aide en francais";

my $fichier_liste = "";
my $nouvelle_etiquette = "";
my $etiquettes_completes = 0;
my $separateur = "";
my $tiret = '-';
my $taille_min = 2;
my $aide = "";
my $help = "";
GetOptions ('liste|list|l=s' => \$fichier_liste,
            'separateur|separator|s=s', \$separateur,
            'etiquette|tag|e=s' => \$nouvelle_etiquette,
            'etiquettes_completes|full_tags|ft|ec'
            => \$etiquettes_completes,
            'tiret|hyphen|h|t=s', \$tiret,
            'taille_min|min_units|mu|tm=i', \$taille_min,
            'aide' => \$aide,
            'help' => \$help);
if ($help) {
  die $DESCRIPTION_EN . "\n\n" . $USAGE_EN . "\n\n";
}
if ($aide) {
  die $DESCRIPTION . "\n\n" . $USAGE . "\n\n";
}
if (!$fichier_liste) {
  die $DESCRIPTION_EN . "\n\n" . $USAGE_EN . "\n\n";
}

my %termes;
my %termes_index;
my @mots = ();
my @mots_etiquettes = ();
my $ligne = "";
my $mot = "";
my $etiquette = "";
my $remplacements = 0;
my $liste;
my $liste2;
my $liste3;
my $i = 0;
my $j = 0;
my $k = 0;
my $l = 0;
my $m = 0;


# charge la liste de termes
open (F_liste, "$fichier_liste")
  or die ("Le fichier [$fichier_liste] n'existe pas !!\n");

while ($ligne = <F_liste>) {
  chomp($ligne);

  # met les mots spars par des espaces dans la liste @mots
  @mots = ();
  while ($ligne =~ /(\S+)(?:\s+|$)/g) {
    push @mots, $1;
  }

  # DEBUG foreach $mot (@mots) { print "$mot, "; }
  # DEBUG print "\n";

  # met l'index des deux premier mots des termes  jour
  $termes_index{$mots[0]}{$mots[1]} = 1;
  # rajoute une liste  la liste des listes de mots des termes
  push @{$termes{$mots[0]}{$mots[1]}{$#mots}}, [@mots];

  # si le terme est suffisamment grand pour ne pas tenir compte de la casse
  if ($#mots >= $taille_min && $taille_min != 0) {

    # cas majuscule en dbut du premier mot puis minuscules
    @mots = map(lc, @mots);
    $mots[0] = ucfirst($mots[0]);
    $termes_index{$mots[0]}{$mots[1]} = 1;
    push @{$termes{$mots[0]}{$mots[1]}{$#mots}}, [@mots];

    # cas majuscule en dbut de chaque mot puis minuscules
    @mots = map(lc, @mots);
    @mots = map(ucfirst, @mots);
    $termes_index{$mots[0]}{$mots[1]} = 1;
    push @{$termes{$mots[0]}{$mots[1]}{$#mots}}, [@mots];

    # cas tout en minuscules
    @mots = map(lc, @mots);
    $termes_index{$mots[0]}{$mots[1]} = 1;
    push @{$termes{$mots[0]}{$mots[1]}{$#mots}}, [@mots];

    # cas tout en majuscules
    @mots = map(uc, @mots);
    $termes_index{$mots[0]}{$mots[1]} = 1;
    push @{$termes{$mots[0]}{$mots[1]}{$#mots}}, [@mots];
  }
}
close F_liste;

# DEBUG print "\n";
# DEBUG # parcours par taille dcroissante les listes de mots
# DEBUG foreach $i (keys %termes) {
# DEBUG   foreach $j (keys %{$termes{$i}}) {
# DEBUG     foreach $k (sort {$b <=> $a} keys %{$termes{$i}{$j}}) {
# DEBUG       print "$i, $j, (";
# DEBUG       foreach $liste (@{$termes{$i}{$j}{$k}}) {
# DEBUG         foreach $liste2 ($liste) {
# DEBUG           print "[ ", join(", ", @$liste2), " ], ";
# DEBUG         }
# DEBUG       }
# DEBUG       print ")\n";
# DEBUG     }
# DEBUG   }
# DEBUG }


# lit l'entre ligne  ligne
while ($ligne = <STDIN>) {
  chomp($ligne);

  # si la ligne est vide passe  l'itration suivante
  if ($ligne =~ /^\s*$/) { next; }

  # dcoupe la ligne en expressions spares par des espaces
  @mots_etiquettes = split('\s', $ligne);

  # rcupre la partie mots des expressions
  if ($separateur) {
    for $i (0..$#mots_etiquettes) {
      $mots[$i] = (split($separateur, $mots_etiquettes[$i]))[0];
    }
  } else {
    @mots = @mots_etiquettes;
  }

  $ligne = ""; $i = 0;
 SORTIE: while ($i < $#mots_etiquettes) {

    # si les deux mots courants dbutent un terme connu
    if (exists $termes_index{$mots[$i]}{$mots[$i+1]}) {
      # DEBUG print "$mots[$i], $mots[$i+1]\n\n";

      # parcours les listes de mots par ordre de taille dcroissante
      foreach $j (sort {$b <=> $a}
                  keys %{$termes{$mots[$i]}{$mots[$i+1]}}) {
        # DEBUG print "j = $j\n";

        # parcours les listes de mots d'une taille donne
        foreach $liste (@{$termes{$mots[$i]}{$mots[$i+1]}{$j}}) {

          # parcours la liste des mots
          foreach $liste2 ($liste) {
            # DEBUG print "[ ", join(", ", @$liste2), " ]\n";

            $k = $i; $l = 0;
            # parcours la liste des mots
            foreach $liste3 (@$liste2) {
              # DEBUG print "l = $l, k = $k\n";
              if ($k > $#mots || @$liste2[$l] ne $mots[$k]) {
                # quitte la boucle si la fin de la phrase est atteinte
                # ou si les mots ne correspondent pas
                last;
              }
              # DEBUG print "@$liste2[$l] = $mots[$k]\n";

              # si tous les mots du texte correspondent au terme
              if ($l == $j) {
                # DEBUG print "l = $l == $j = j\n\n";
                for $m ($i..($i+$j)) {
                  # groupe les mots du texte avec des tirets
                  $ligne .= "$mots[$m]$tiret";
                }
                # retire le tiret final en trop
                chop($ligne);

                # rajoute l'tiquette du dernier mot
                if ($separateur) {
                  if ($etiquettes_completes) {
                    $etiquette = "";
                    for $m ($i..($i+$j)) {
                      $etiquette .=
                        (split($separateur, $mots_etiquettes[$m]))[1]
                          . $tiret;
                    }
                    chop($etiquette);
                  } else {
                    $etiquette =
                      (split($separateur, $mots_etiquettes[$i+$j]))[1];
                  }
                  $ligne .= "$separateur$etiquette$nouvelle_etiquette ";

                } elsif ($nouvelle_etiquette) {
                  $ligne .= "/$nouvelle_etiquette ";

                } else {
                  $ligne .= " ";
                }
                # avance de la longueur du terme dans les mots
                $i = $i+$j+1;
                # incrmente le nombre de remplacements effectus
                $remplacements++;
                # revient  la boucle principale
                next SORTIE;
              }

              $l++; $k++;
            }
          }
        }
      }
      # si le terme n'a pas t trouv alors ne modifie pas le texte
      # d'entre; lorsque le terme est trouv on ne passe pas ici
      $ligne .= "$mots_etiquettes[$i] ";

    } else {
      # sinon ne modifie pas le texte d'entre
      $ligne .= "$mots_etiquettes[$i] ";
    }

    $i++;
  }

  # rajoute le dernier mot de la ligne
  $ligne .= "$mots_etiquettes[$i] ";

  print "$ligne\n";
}

print STDERR "$remplacements remplacements effectues / replacements done\n";
