
package extractTerms;

import java.util.*;
import java.io.*;
import java.util.regex.*;
import java.text.*;

/**
 * A first method, extract, extract pairs of adjacent units and estimate
 * their "affinity" with several functions like the mutual information
 * function. Units triple are processed too. But the middle unit is
 * considered as informationless.
 * 
 * <p>A second method wich is static, replace, enable the replacement of
 * separate units encapsulated in an Occurrence object with a term that
 * group the unit. The inverse operation is also possible.
 *
 * @author Thomas Heitz for LRI Paris XI University
 * @version 0.8 22/09/2004
 * @since 27/06/2003
 */
public class ExtractTerms {

    /** Contain the number of occurrences of each unit in the input file
        indexed by units. */
    private Hashtable occFor1Unit;

    /** Contain Occurrence objects indexed by units pair.
        @see Occurrence */
    private Hashtable occFor2Units;

    /** Contain total number of occurrences for each units pair grouped by
        first unit indexed by first unit. */
    private Hashtable occFor2Units1stUnitKnown;

    /** Contain total number of occurrences for each units pair grouped by
        second unit indexed by second unit. */
    private Hashtable occFor2Units1stUnitKnownRevIndex;
    
    /** Contain Occurrence objects indexed by units triple.
        @see Occurrence */
    private Hashtable occFor3Units;

    /** Contain total number of occurrences for each units triple grouped
        by first unit indexed by first unit. */
    private Hashtable occFor3Units1stUnitKnown;

    /** Contain total number of occurrences for each units triple grouped
        by third unit indexed by third unit. */
    private Hashtable occFor3Units1stUnitKnownRevIndex;
    
    /** Sorted Occurrence objects by relevancy of pairs.
        @see Occurrence */
    private TreeSet results2Units;

    /** Sorted Occurrence objects by relevancy of triples.
        @see Occurrence */
    private TreeSet results3Units;

    /** Total number of units without double contained in the input file. */
    private int totalUnitsWithoutDouble;

    /** Total number of units pair without double contained in the input
     * file. */
    private int totalUnitsPairWithoutDouble;

    /** Total number of units pair without triple contained in the input
     * file. */
    private int totalUnitsTripleWithoutDouble;

    /** Total number of results for units pair with pruning. */
    private int totalResults2Units;

    /** Total number of results for units triple with pruning. */
    private int totalResults3Units;

    /** Total number of units pairs with double contained in the input
     * file. */
    private int totalUnitsPair;

    /** Total number of units triple with double contained in the input
     * file. */
    private int totalUnitsTriple;

    /** Total number of documents in the input file. */
    private int totalDocuments;

    // see constructor
    private static String inputFileName;
    private static Pattern patternUnit;
    private static Pattern patternSentPhraSeparator;
    private static Pattern patternNameOfText;
    private static Pattern pattern2Units, pattern2UnitsReversed;
    private static int tagKept2Units;
    private static Pattern pattern3Units, pattern3UnitsReversed;
    private static int tagKept3Units;
    private static int formula;
    private static boolean tfidf;
    private static int pruning2Units, pruning3Units;
    private static int maxResults2Units, maxResults3Units;
    private static String separator;
    private static int relation2Units;
    private static int relation3Units;

    /**
     * Two regular expressions for matching pair and triple of units must
     * be provided. It is strongly advised to provide as an input file a
     * tagged file so you could analyse only certain syntaxic relations as
     * Adjective-Noun for example.
     *
     * @param _inputFileName name of input file containing text to analyse
     * @param _regexUnit regular expression used for splitting input text
     * into units
     * @param _regexSentPhraSeparator regular expression for separator of
     * sentence or phrase
     * @param _regexNameOfText regular expression used for extract name of
     * text in the corpus
     * @param _regex2Units1 regular expression of analysed units pair for
     * the first unit
     * @param _regex2Units2 regular expression of analysed units pair for
     * the second unit
     * @param _tagKept2Units tag to keep between the 2 units when group
     * @param _regex3Units1 regular expression of analysed units triple for
     * the first unit
     * @param _regex3Units2 regular expression of analysed units triple for
     * the second unit
     * @param _regex3Units3 regular expression of analysed units triple for
     * the third unit
     * @param _tagKept3Units tag to keep between the 3 units when group
     * @param _formula formula for calculating relevancy
     * @param _tfidf true if relevancy measures are based on tf*idf values
     * rather than number of occurrence
     * @param _pruning2Units units pair that occur stricly less than pruning
     * time doesn't appear in the results
     * @param _pruning3Units units triple that occur stricly less than
     * @param _maxResults2Units maximum number of results of units pair
     * @param _maxResults3Units maximum number of results of units triple
     * pruning time doesn't appear in the results
     * @param _separator separator between word and tag
     * @param _relation2Units position of the element in the list of 2 units
     * selector that was selected for this term ; -1 if you don't want
     * extraction of units pair
     * @param _relation3Units position of the element in the list of 3 units
     * selector that was selected for this term ; -1 if you don't want
     * extraction of units triple
     * @param _occFirst if the measure takes first into account the number of
     * occurrence
     * @see ExtractTermsExample#main(java.lang.String[]) ExtractTermsExample
     * for an example
     */
    public ExtractTerms(String _inputFileName,
                        String _regexUnit,
                        String _regexNameOfText,
                        String _regexSentPhraSeparator,
                        String _regex2Units1,
                        String _regex2Units2,
                        int _tagKept2Units,
                        String _regex3Units1,
                        String _regex3Units2,
                        String _regex3Units3,
                        int _tagKept3Units,
                        int _formula,
                        boolean _tfidf,
                        int _pruning2Units, int _pruning3Units,
                        int _maxResults2Units,
                        int _maxResults3Units,
                        String _separator,
                        int _relation2Units,
                        int _relation3Units,
                        boolean _occFirst) {


        // objects initialisation
        inputFileName = _inputFileName;
        patternUnit = Pattern.compile(_regexUnit);
        patternNameOfText = Pattern.compile(_regexNameOfText);
        patternSentPhraSeparator = Pattern.compile(_regexSentPhraSeparator);
        pattern2Units = Pattern.compile(_regex2Units1+_regex2Units2);
        pattern2UnitsReversed =
            Pattern.compile(_regex2Units2+_regex2Units1);
        tagKept2Units = _tagKept2Units;
        pattern3Units =
            Pattern.compile(_regex3Units1+_regex3Units2+_regex3Units3);
        pattern3UnitsReversed =
            Pattern.compile(_regex3Units3+_regex3Units2+_regex3Units1);
        tagKept3Units = _tagKept3Units;
        formula = _formula;
        tfidf = _tfidf;
        pruning2Units = _pruning2Units;
        pruning3Units = _pruning3Units;
        maxResults2Units = _maxResults2Units;
        maxResults3Units = _maxResults3Units;
        separator = _separator;
        relation2Units = _relation2Units;
        relation3Units = _relation3Units;

        // disable the extraction of units pair or triple
        if (relation2Units == -1) { pattern2Units = Pattern.compile(""); }
        if (relation3Units == -1) { pattern3Units = Pattern.compile(""); }

        occFor1Unit = new Hashtable();
        occFor2Units = new Hashtable();
        occFor3Units = new Hashtable();
        occFor2Units1stUnitKnown = new Hashtable();
        occFor2Units1stUnitKnownRevIndex = new Hashtable();
        occFor3Units1stUnitKnown = new Hashtable();
        occFor3Units1stUnitKnownRevIndex = new Hashtable();
        if (_occFirst) {
            results2Units =
                new TreeSet(new ComparatorOccurrenceFirstMeasure());
            results3Units =
                new TreeSet(new ComparatorOccurrenceFirstMeasure());
        } else {
            results2Units = new TreeSet(new ComparatorOccurrenceMeasure());
            results3Units = new TreeSet(new ComparatorOccurrenceMeasure());
        }
    }

    /**
     * Returns total number of units without double contained in the input
     * file.
     *
     * @return total number of units without double.
     */
    public int getTotalUnitsWithoutDouble() {
        return totalUnitsWithoutDouble;
    }

    /**
     * Returns total number of units pair without double contained in the
     * input file.
     *
     * @return total number of units pair without double.
     */
    public int getTotalUnitsPairWithoutDouble() {
        return totalUnitsPairWithoutDouble;
    }

    /**
     * Returns total number of units triple without double contained in the
     * input file.
     *
     * @return total number of units triple without double.
     */
    public int getTotalUnitsTripleWithoutDouble() {
        return totalUnitsTripleWithoutDouble;
    }

    /**
     * Returns sorted results by relevancy of units pair.
     *
     * @return list of sorted Occurrence results.
     */
    public TreeSet getResults2Units() {
        return results2Units;
    }

    /**
     * Returns sorted results by relevancy of units triple.
     *
     * @return list of sorted Occurrence results.
     */
    public TreeSet getResults3Units() {
        return results3Units;
    }

    /**
     * Returns total results for units pair with pruning.
     *
     * @return number of results for units pair with pruning.
     */
    public int getTotalResults2Units() {
        return totalResults2Units;
    }

    /**
     * Returns total results for units triple with pruning.
     *
     * @return number of results for units triple with pruning.
     */
    public int getTotalResults3Units() {
        return totalResults3Units;
    }

    /**
     * Returns total number of units pair with double contained in the
     * input file.
     *
     * @return total number of units pair with double.
     */
    public int getTotalUnitsPair() {
        return totalUnitsPair;
    }

    /**
     * Returns total number of units triple with double contained in the
     * input file.
     *
     * @return total number of units triple with double.
     */
    public int getTotalUnitsTriple() {
        return totalUnitsTriple;
    }

    /**
     * Returns total number of documents in the input file.
     *
     * @return total number of documents.
     */
    public int getTotalDocuments() {
        return totalDocuments;
    }

    /**
     * Returns pattern of regular expression used for splitting input text
     * into units.
     *
     * @return pattern of regular expression used for splitting input text
     * into units.
     */
    public Pattern getPatternUnit() {
        return patternUnit;
    }

    /**
     * Returns pattern of regular expression used for analysed units pair.
     *
     * @return pattern of regular expression used for analysed units pair.
     */
    public Pattern getPattern2Units() {
        return pattern2Units;
    }

    /**
     * Returns pattern of regular expression used for analysed units triple.
     *
     * @return pattern of regular expression used for analysed units triple.
     */
    public Pattern getPattern3Units() {
        return pattern3Units;
    }

    /**
     * Returns pattern of regular expression for the separator of sentence
     * or phrases.
     *
     * @return pattern for the separator of sentence or phrases.
     */
    public Pattern getPatternSentPhraSeparator() {
        return patternSentPhraSeparator;
    }

    /**
     * Returns pattern of regular expression used for extract name of
     * text in the corpus.
     *
     * @return pattern for name of text
     */
    public Pattern getPatternNameOfText() {
        return patternNameOfText;
    }

    /**
     * Returns pruning threshold for units pair.
     *
     * @return value of pruning threshold.
     */
    public int getPruning2Units() {
        return pruning2Units;
    }

    /**
     * Returns pruning threshold for units triple.
     *
     * @return value of pruning threshold.
     */
    public int getPruning3Units() {
        return pruning3Units;
    }

    /**
     * Returns separator between word and tag.
     *
     * @return separator.
     */
    public String getSeparator() {
        return separator;
    }

    /**
     * Returns input file name.
     *
     * @return input file name
     */
    public String getInputFileName() {
        return inputFileName;
    }

    /**
     * Set input file name.
     *
     * @param _inputFileName input file name
     */
    public void setInputFileName(String _inputFileName) {
        inputFileName = _inputFileName;
    }

    /** Shortcut for shorting formulas. */
    private static double log(double a) {
        return Math.log(a);
    }

    /** Shortcut for shorting formulas. */
    private static double pow(double a, double b) {
        return Math.pow(a, b);
    }

    /**
     * Extract terms from the input file and sort them by relevancy with
     * a relevancy measure formula.
     *
     * <p>
     * example :
     *
     * <pre><code>
     * ExtractTerms et = new ExtractTerms(...);
     * et.extract();
     * Iterator it = et.getResults2Units().iterator();
     * System.out.println("Relevancy results for units pair");
     * while (it.hasNext()) { System.out.println(it.next()); }
     *</code></pre>
     *
     * <p>No results is return in return parameters but in fields of the
     * ExtractTerms object. You can get results with getResults2Units() and
     * getResults3Units().</p>
     *
     * @param currentSkippedTerms contain the terms (String) to skip
     * @param iteration number of the iteration when this term was
     * extracted
     * @param extractionJProgressBar progress bar of the extraction ; null
     * if not used
     * @return false if an error as occurred, true otherwise.
     */
    public boolean extract(TreeSet currentSkippedTerms, int iteration,
        javax.swing.JProgressBar extractionJProgressBar) {

        String lineContent;
        String unit, previousUnit, previous2Unit;
        Integer occNb;
        Occurrence occ;
        Vector vect;
        Vector vectContext;
        HashSet setNameFile;
        int line;
        String nameOfText;
        Matcher matcherNameOfText;
        BufferedReader br;
        Matcher matcherUnit;


        // open the input file in reading mode
        try {
            br = new BufferedReader(new FileReader(inputFileName));

        } catch (FileNotFoundException fnfe) {
            System.out.println("ExtractTerms.extract : "+fnfe.getMessage());
            return false;
        }

        // objects initialisation
        totalUnitsPair = 0;
        totalUnitsTriple = 0;
        totalDocuments = 0;
        unit = null;
        previousUnit = null;
        previous2Unit = null;
        matcherUnit = null;
        matcherNameOfText = null;
        nameOfText = "";
        line = 1;


        // read the input file line by line only one time and
        // fill the hashtables with units and their occurrences
        try {

        lineContent = br.readLine();

        while (lineContent != null) {

            // update the progress bar each 50 lines
            if ((line % 50 == 0) && extractionJProgressBar != null) {
                extractionJProgressBar.setValue(line);
                extractionJProgressBar.update(
                    (extractionJProgressBar).getGraphics());
            }

            matcherNameOfText = patternNameOfText.matcher(lineContent);
            // get the name of text if the line match regex
            if  (matcherNameOfText.matches()) {
                nameOfText = matcherNameOfText.group(1);
                totalDocuments++;
                // skip the line
                lineContent = br.readLine();
                line++;
                continue;
            }

            matcherUnit = patternUnit.matcher(lineContent);
            previous2Unit = null;
            previousUnit = null;

            while(matcherUnit.find()) {
                unit = matcherUnit.group();

                // if a sentence/phrase separator is found skip it
                if (patternSentPhraSeparator.matcher(unit).matches()) {
                    previousUnit = null;
                    continue;
                }

                // set the current number of occurrences for the unit
                occNb = (Integer)occFor1Unit.get(unit);
                if (occNb != null) {
                    // update the number of occurrences
                    occFor1Unit.put(unit, new Integer(occNb.intValue()+1));
                } else {
                    // add a new unit
                    occFor1Unit.put(unit, new Integer(1));
                }

                // wait to find the second unit before filling
                // occurrences for units pair
                if (previousUnit == null) {
                    previousUnit = unit;
                    continue;
                }

                // skip the pair if it doesn't match
                if (pattern2Units.matcher(previousUnit+unit).matches()) {

                    totalUnitsPair++;

                    // set the number of occurrences, line and name
                    // of text for the units pair
                    occ = (Occurrence)occFor2Units.get(previousUnit+unit);
                    if (occ != null) {
                        // update the number of occurrences
                        occ.incrementOcc();
                    } else {
                        // create a new occurrence
                        occ = new Occurrence(previousUnit, unit);
                    }
                    // save line of input file for the units pair
                    occ.addLine(line);
                    // save the name of text
                    occ.addNameOfText(nameOfText);
                    occFor2Units.put(previousUnit+unit, occ);

                    // set the total number of occurrences for each units
                    // pair grouped by first unit
                    occNb =
                        (Integer)occFor2Units1stUnitKnown.get(previousUnit);
                    if (occNb != null) {
                        // update the number of occurrences
                        occFor2Units1stUnitKnown.put(previousUnit,
                            new Integer(occNb.intValue()+1));
                    } else {
                        // add a new unit
                        occFor2Units1stUnitKnown.put(
                            previousUnit,new Integer(1));
                    }

                    // set the total number of occurrences for each units
                    // pair grouped by second unit
                    occNb =
                        (Integer)occFor2Units1stUnitKnownRevIndex.get(unit);
                    if (occNb != null) {
                        // update the number of occurrences
                        occFor2Units1stUnitKnownRevIndex.put(
                            unit, new Integer(occNb.intValue()+1));
                    } else {
                        // add a new unit
                        occFor2Units1stUnitKnownRevIndex.put(
                            unit, new Integer(1));
                    }
                }

                // wait to find the third unit before filling
                // occurrences for units triple
                if (previous2Unit == null) {
                    previous2Unit = previousUnit;
                    previousUnit = unit;
                    continue;
                }

                // skip the triple if it doesn't match
                if (pattern3Units.matcher(
                        previous2Unit+previousUnit+unit).matches()) {

                    totalUnitsTriple++;

                    // set the current number of occurrences and number of
                    // context for the units triple
                    occ = (Occurrence)occFor3Units.get(
                        previous2Unit+previousUnit+unit);
                    if (occ != null) {
                        // update the number of occurrences
                        occ.incrementOcc();
                    } else {
                        // create a new occurrence
                        occ = new Occurrence(previous2Unit, previousUnit,
                                             unit);
                    }
                    // save line of input file for the units triple
                    occ.addLine(line);
                    // save the name of text
                    occ.addNameOfText(nameOfText);
                    occFor3Units.put(previous2Unit+previousUnit+unit, occ);

                    // set the total number of occurrences for each units
                    // triple grouped by first unit
                    occNb = (Integer)
                        occFor3Units1stUnitKnown.get(previous2Unit);
                    if (occNb != null) {
                        // update the number of occurrences
                        occFor3Units1stUnitKnown.put(previous2Unit,
                            new Integer(occNb.intValue()+1));
                    } else {
                        // add a new unit
                        occFor3Units1stUnitKnown.put(
                            previous2Unit, new Integer(1));
                    }

                    // set the total number of occurrences for each units
                    // triple grouped by third unit
                    occNb =
                        (Integer)occFor3Units1stUnitKnownRevIndex.get(unit);
                    if (occNb != null) {
                        // update the number of occurrences
                        occFor3Units1stUnitKnownRevIndex.put(
                            unit, new Integer(occNb.intValue()+1));
                    } else {
                        // add a new unit
                        occFor3Units1stUnitKnownRevIndex.put(
                            unit, new Integer(1));
                    }
                }

                previous2Unit = previousUnit;
                previousUnit = unit;
            }
            lineContent = br.readLine();
            line++;
        }

        // close the input file
        br.close();

        } catch(IOException ioe){
            System.out.println("ExtractTerms.extract : "+ioe.getMessage());
            return false;
        }

        totalUnitsWithoutDouble = occFor1Unit.size();
        totalUnitsPairWithoutDouble = occFor2Units.size();
        totalUnitsTripleWithoutDouble = occFor3Units.size();

        Iterator it, it2;
        Map.Entry me;
        String str;
        double p1, p2, a, b, c, d;
        double n, cst, nMin, nMax;

        // calculate relevancy measure for units pair
        // the informations of the previous hashtables are used
        // each measure results are put in sorted sets
        it = occFor2Units.entrySet().iterator();
        while(it.hasNext()) {
            me = (Map.Entry)it.next();
            str = (String)me.getKey();
            occ = (Occurrence)me.getValue();

            if (currentSkippedTerms.contains(occ.getKey())) { continue; }

            a = occ.getOccInt().doubleValue();
            if (a < pruning2Units) { continue; }

            // calculate tf*idf value of units pair
            if (tfidf && totalDocuments > 0) {
                a = 0;
                cst = log(totalDocuments/occ.getNamesOfText().size())
                    / log(2);
                it2 = occ.getNamesOfText().values().iterator();
                while(it2.hasNext()) {
                    a += ((Integer)it2.next()).doubleValue() * cst;
                }
            }
            p1 = ((Integer)occFor1Unit.get(occ.getUnit1())).doubleValue();
            p2 = ((Integer)occFor1Unit.get(occ.getUnit2())).doubleValue();
            b = ((Integer)occFor2Units1stUnitKnown.get(occ.getUnit1()))
                .doubleValue() - a;
            // to avoid infinite value with 0*log(0)
            if (b <= 0) { b = Double.MIN_VALUE; }
            c = ((Integer)occFor2Units1stUnitKnownRevIndex
                 .get(occ.getUnit2())).doubleValue() - a;
            if (c <= 0) { c = Double.MIN_VALUE; }
            d = totalUnitsPair - a - b - c;
            if (d <= 0) { d = Double.MIN_VALUE; }

            n = 0.0;
            switch (formula) {

            case 0:
                // mutual information
                n = log(a/(p1*p2)) / log(2);
                break;
 
            case 1:
                // vraisemblance
                n = a*log(a) + b*log(b) + c*log(c) + d*log(d)
                    - (a+b)*log(a+b) - (a+c)*log(a+c)
                    - (b+d)*log(b+d) - (c+d)*log(c+d)
                    + totalUnitsPair*log(totalUnitsPair);
                break;

            case 2:
                // cubic mutual information
                n = log(pow(a/(p1*p2), 3)) / log(2);
                break;
            }

            if (!Double.isNaN(n)) {
                occ.setRelation(relation2Units);
                occ.setIteration(iteration);
                occ.setMeasure(n);
                results2Units.add(occ);
            } else {
                assert false : "occ = "+occ+"a = "+a+", b = "+b+", c = "
                    +c+", d = "+d+", p1 = "+p1+", p2 = "+p2;
            }
        }

        totalResults2Units = results2Units.size();

        // remove the less relevant terms if max size is exceeded
        if (results2Units.size() > maxResults2Units) {
            int m = results2Units.size() - maxResults2Units;
            for (int i = 0 ; i < m; i++) {
                results2Units.remove(results2Units.last());
            }
        }

        if (!results2Units.isEmpty()) {
            // normalize the results between 0 and 1
            nMin = Double.POSITIVE_INFINITY;
            nMax = Double.NEGATIVE_INFINITY;
            it = results2Units.iterator();
            while(it.hasNext()) {
                n = ((Occurrence)it.next()).getMeasure();
                if (nMax < n) { nMax = n; }
                if (nMin > n) { nMin = n; }
            }
            it = results2Units.iterator();
            while(it.hasNext()) {
                occ = (Occurrence)it.next();
                occ.setMeasure(((occ.getMeasure())/(nMax-nMin))
                                - (nMin/(nMax-nMin)));
            }
        }

        // no more need now so clear it
        occFor2Units1stUnitKnown.clear();
        occFor2Units1stUnitKnownRevIndex.clear();

 
        // calculate relevancy measure for units triple
        // the middle unit is ignored in calculus
        // the informations of the previous hashtables are used
        // each measure results are put in sorted sets
        it = occFor3Units.entrySet().iterator();
        while(it.hasNext()) {
            me = (Map.Entry)it.next();
            str = (String)me.getKey();
            occ = (Occurrence)me.getValue();

            if (currentSkippedTerms.contains(occ.getKey())) { continue; }

            a = occ.getOccInt().doubleValue();
            if (a < pruning3Units) { continue; }

            // calculate tf*idf value of units pair
            if (tfidf && totalDocuments > 0) {
                a = 0;
                cst = log(totalDocuments/occ.getNamesOfText().size())
                    / log(2);
                it2 = occ.getNamesOfText().values().iterator();
                while(it2.hasNext()) {
                    a += ((Integer)it2.next()).doubleValue() * cst;
                }
            }
            p1 = ((Integer)occFor1Unit.get(occ.getUnit1())).doubleValue();
            p2 = ((Integer)occFor1Unit.get(occ.getUnit3())).doubleValue();
            b = ((Integer)occFor3Units1stUnitKnown.get(occ.getUnit1()))
                .doubleValue() - a;
            // to avoid infinite value with 0*log(0)
            if (b <= 0) { b = Double.MIN_VALUE; }
            c = ((Integer)occFor3Units1stUnitKnownRevIndex
                 .get(occ.getUnit3())).doubleValue() - a;
            if (c <= 0) { c = Double.MIN_VALUE; }
            d = totalUnitsTriple - a - b - c;
            if (d <= 0) { d = Double.MIN_VALUE; }

            n = 0.0;
            switch (formula) {

            case 0:
                // mutual information
                n = log(a/(p1*p2)) / log(2);
                break;

            case 1:
                // vraisemblance
                n = a*log(a) + b*log(b) + c*log(c) + d*log(d)
                    - (a+b)*log(a+b) - (a+c)*log(a+c)
                    - (b+d)*log(b+d) - (c+d)*log(c+d)
                    + totalUnitsTriple*log(totalUnitsTriple);
                break;

            case 2:
                // cubic mutual information
                n = log(pow(a/(p1*p2), 3)) / log(2);
                break;
            }

            if (!Double.isNaN(n)) {
                occ.setRelation(relation3Units);
                occ.setIteration(iteration);
                occ.setMeasure(n);
                results3Units.add(occ);
            } else {
                assert false : "occ = "+occ+"a = "+a+", b = "+b+", c = "
                    +c+", d = "+d+", p1 = "+p1+", p2 = "+p2;
            }
        }

        totalResults3Units = results3Units.size();

        // remove the less relevant terms if max size is exceeded
        if (results3Units.size() > maxResults3Units) {
            int m = results3Units.size() - maxResults3Units;
            for (int i = 0 ; i < m; i++) {
                results3Units.remove(results3Units.last());
            }
        }

        if (!results3Units.isEmpty()) {
            // normalize the results between 0 and 1
            nMin = Double.POSITIVE_INFINITY;
            nMax = Double.NEGATIVE_INFINITY;
            it = results3Units.iterator();
            while(it.hasNext()) {
                n = ((Occurrence)it.next()).getMeasure();
                if (nMax < n) { nMax = n; }
                if (nMin > n) { nMin = n; }
            }
            it = results3Units.iterator();
            while(it.hasNext()) {
                occ = (Occurrence)it.next();
                occ.setMeasure((occ.getMeasure()/(nMax-nMin))
                                - (nMin/(nMax-nMin)));
            }
        }

        // no more need now so clear it
        occFor3Units1stUnitKnown.clear();
        occFor3Units1stUnitKnownRevIndex.clear();

        return true;
    } // method extract


    /**
     * Comparator for Vector for Occurence object.
     *
     * <p>
     * First compare by descending measure value, then by ascending unit1,
     * unit2 and unit3.
     */
    private class ComparatorOccurrenceMeasure implements Comparator {

        public int compare(Object o1, Object o2) {

            Occurrence occ1 = (Occurrence)o1;
            Occurrence occ2 = (Occurrence)o2;

            if (!occ2.getMeasureDbl().equals(occ1.getMeasureDbl())) {
                return occ2.getMeasureDbl().compareTo(occ1.getMeasureDbl());

            } else if (!occ1.getUnit1().equals(occ2.getUnit1())) {
                return occ1.getUnit1().compareTo(occ2.getUnit1());

            } else if (!occ1.getUnit2().equals(occ2.getUnit2())) {
                return occ1.getUnit2().compareTo(occ2.getUnit2());
 
            } else if (occ1.is2Units() && occ2.is2Units()) {
                return 0;

            } else if (occ1.is3Units() && occ2.is3Units()) {

                if (!occ1.getUnit3().equals(occ2.getUnit3())) {
                    return occ1.getUnit3().compareTo(occ2.getUnit3());
                } else {
                    return 0;
                }

            } else { // comparison of 2 and 3 units Occurrence
                return -1;
            }
        }

        public boolean equals(Object obj) {

            return this.equals(obj);
        }
    }

    /**
     * Comparator for Vector for Occurence object.
     *
     * <p>
     * First compare by number of occurrence, then by descending measure
     * value and eventually by ascending unit1, unit2 and unit3.
     */
    private class ComparatorOccurrenceFirstMeasure implements Comparator {

        public int compare(Object o1, Object o2) {

            Occurrence occ1 = (Occurrence)o1;
            Occurrence occ2 = (Occurrence)o2;

            if (!occ2.getOccInt().equals(occ1.getOccInt())) {
                return occ2.getOccInt().compareTo(occ1.getOccInt());
            } else if (!occ2.getMeasureDbl().equals(occ1.getMeasureDbl())) {
                return occ2.getMeasureDbl().compareTo(occ1.getMeasureDbl());

            } else if (!occ1.getUnit1().equals(occ2.getUnit1())) {
                return occ1.getUnit1().compareTo(occ2.getUnit1());

            } else if (!occ1.getUnit2().equals(occ2.getUnit2())) {
                return occ1.getUnit2().compareTo(occ2.getUnit2());
 
            } else if (occ1.is2Units() && occ2.is2Units()) {
                return 0;

            } else if (occ1.is3Units() && occ2.is3Units()) {

                if (!occ1.getUnit3().equals(occ2.getUnit3())) {
                    return occ1.getUnit3().compareTo(occ2.getUnit3());
                } else {
                    return 0;
                }

            } else { // comparison of 2 and 3 units Occurrence
                return -1;
            }
        }

        public boolean equals(Object obj) {

            return this.equals(obj);
        }
    }

    /**
     * Comparator for Vector of (Integer, String)
     * <p>
     * First compare by Integer value, then by String value.
     */
    private static class ComparatorIntegerString implements Comparator {

        public int compare(Object o1, Object o2) {

            Vector v1 = (Vector)o1;
            Vector v2 = (Vector)o2;

            if (!((Integer)v1.firstElement()).equals(v2.firstElement())) {
                return ((Integer)v1.firstElement())
                    .compareTo(v2.firstElement());
            } else {
                return ((String)v1.lastElement())
                    .compareTo(v2.lastElement());
            }
        }

        public boolean equals(Object obj) {

            return this.equals(obj);
        }
    }

    /**
     * Determine the mode of replacement in the replace method.
     *
     * @see ExtractTerms#replace(java.util.LinkedHashSet, java.lang.String,
     * java.lang.String, java.lang.String, int) replace
     */
    public static final int GROUP_MODE = 0;

    /**
     * Determine the mode of replacement in the replace method.
     *
     * @see ExtractTerms#replace(java.util.LinkedHashSet, java.lang.String,
     * java.lang.String, java.lang.String, int) replace
     */
    public static final int UNGROUP_MODE = 1;

    /**
     * Replace in the file inputFileName adjacent units contained into units
     * with un/grouped units and write a new file named newFile.
     *
     * <p>unit1 unit2...unitn <-> unit1-unit2-...-unitn
     *
     * @param units list of Occurrence which contain the units pair or
     * triple ungrouped with their lines.
     * @param wordSeparator used for grouping units
     * @param inputFileName input file name
     * @param newFileName name of new file written
     * @param mode GROUP_MODE or UNGROUP_MODE
     * @see Occurrence
     */
    public static void replace(LinkedHashSet units, String wordSeparator,
                               String inputFileName, String newFileName,
                               int mode) {

        Iterator it, it2;
        String str, strReplacement, strPrecedent;
        TreeSet occUnitsTreeSet;
        Vector linesVect;
        String lineContent;
        Occurrence occ;
        int line, lineOcc;
        BufferedReader br;
        PrintWriter out;
        Matcher matcher, matcher2;
        Vector vect, vect1, vect2;
        String [] unitsArray;
        int i;

        File newFile = new File(newFileName);
        // needed if input file and output is the same file
        File tempFile = new File(newFileName+".tmp");
        File inputFile = new File(inputFileName);
        occUnitsTreeSet = new TreeSet(new ComparatorIntegerString());

        // construct a list of units sorted by ascending lines of
        // input file where they occur
        it = units.iterator();
        while(it.hasNext()) {
            occ = (Occurrence)it.next();

            // add each line where the units occurs
            it2 = occ.getLinesIterator();
            while(it2.hasNext()) {
                vect = new Vector();
                vect.add((Integer)it2.next());
                vect.add(occ.getKey());
                // add and sort by line the units
                occUnitsTreeSet.add(vect);
            }
        }

        // no units or units without lines was found in Vector units
        if (occUnitsTreeSet.size() == 0) {
            System.out.println("ExtractTerms.replace : can't find provided "
                               +"Occurrences in the input file.");
            System.out.println(units);
            return;
        }

        try {

        // open input file in reading
        br = new BufferedReader(new FileReader(inputFile));

        // open output file in writing
        out = new PrintWriter(new BufferedWriter(
                                  new FileWriter(tempFile)));

        // search units in input file
        it = occUnitsTreeSet.iterator();
        line = 0;
        strReplacement = "";
        lineContent = "";
        strPrecedent = null;
        while(it.hasNext()) {
            vect = (Vector)it.next();
            lineOcc = ((Integer)vect.get(0)).intValue();
            str = (String)vect.get(1);
            // replace '\' with '\\' to escape the escape character
            str = str.replaceAll("\\\\", "\\\\\\\\");
            str = str.replaceAll("\\$", "\\\\\\$");

            // write the line with un/grouped units
            if (line < lineOcc && !strReplacement.equals("")) {
                out.println(lineContent);
                strReplacement = "";
                strPrecedent = null;
            }

            // copy the line without modifications
            while (line < lineOcc-1) {
                line++;
                out.println(br.readLine());
            }

            // read the line where replacement occurs
            if (line == lineOcc-1 && strReplacement.equals("")) {
                line++;
                lineContent = br.readLine();
            }

            if  (strPrecedent == null || !str.equals(strPrecedent)) {
                // create the remplacement string for the separated units
                strReplacement = "";
                matcher = patternUnit.matcher(str);
                unitsArray = new String [6];
                Arrays.fill(unitsArray, null);
                i = -1;
                // split the occurrence in words [0,2,4] and tags [1,3,5]
                while (matcher.find()) {
                    unitsArray[++i] =
                    
                        Occurrence.getWord(matcher.group(), separator);
                    unitsArray[++i] =
                        Occurrence.getTag(matcher.group(), separator);
                }
                strReplacement = unitsArray[0]+wordSeparator+unitsArray[2];

                if (unitsArray[5] == null) {
                    strReplacement += separator;
                    switch (tagKept2Units) {
                        case 1: strReplacement += unitsArray[1]; break;
                        case 2: strReplacement += unitsArray[3]; break;
                    }
                } else {
                    strReplacement += wordSeparator+unitsArray[4]+separator;
                    switch (tagKept3Units) {
                        case 1: strReplacement += unitsArray[1]; break;
                        case 2: strReplacement += unitsArray[3]; break;
                        case 3: strReplacement += unitsArray[5]; break;
                    }
                }
                strReplacement += " ";
                // avoid a error if there is $g with g a number between 0
                // and 9 in the replacement string because each occurrence
                // of $g will be replaced by the result of evaluating
                // group(g)
                // so replace "$" with "\$" before to do replaceAll
                strReplacement =
                    strReplacement.replaceAll("\\$", "\\\\\\$");

                // replace '\' with '\\' to escape the escape character
                strReplacement =
                    strReplacement.replaceAll("\\\\", "\\\\\\\\");
            }

            // create the replacement line
            switch (mode) {

            case GROUP_MODE:
                lineContent =
                    lineContent.replaceAll("\\Q"+str+"\\E", strReplacement);
                break;

            case UNGROUP_MODE:
                lineContent =
                    lineContent.replaceAll("\\Q"+strReplacement+"\\E", str);
                break;
            }
            strPrecedent = str;
        }

        // write the last line with grouped units
        if (!strReplacement.equals("")) {
            out.println(lineContent);
        }

        // copy the remaining lines of the input file
        lineContent = br.readLine();
        while (lineContent != null) {
            out.println(lineContent);
            lineContent = br.readLine();
        }

        // close the input and output files
        br.close();
        out.close();

        if (newFile.exists()) {
            newFile.delete();
        }
        // rename the temporary output file to newFileName
        tempFile.renameTo(newFile);

        } catch (FileNotFoundException fnfe) {
            System.out.println("ExtractTerms.replace : Error in opening "
                               +"file. ("+inputFile+")");
            return;

        } catch (IOException ioe) {
            System.out.println("ExtractTerms.replace : Error in writing "
                               +"file. ("+tempFile+")");
            return;
        }

        return;
    }

} // class ExtractTerms
