
package extractTerms;

/**
 * Local rule used for improving relevancy of extraction. These rules are
 * designed to be used on Occurrence object. The result of the application
 * of these rules is true or false.
 *
 * <p>Grammar of the local rules language :
 *
 * <pre>
 * SOURCE -> E
 * E -> F | not E | E and E | E or E
 * F -> (UNIT, PART, STRING, STATUS, TYPE)
 * UNIT -> 1 | 2 | 3
 * PART -> prefix | suffix | contain | tag | contain_regex | tag_regex
 * STRING -> [^SEPARATOR]+
 * SEPARATOR -> '/' or any other character
 * STATUS -> ACCEPTED | REJECTED
 * TYPE - > any string
 * </pre>
 *
 * @author Thomas Heitz for LRI Paris XI University
 * @version 0.2 28/07/2004
 * @since 07/08/2003
 * @see Occurrence
 */
public class LocalRule implements java.io.Serializable {

    // see constructor
    private final int operator;
    private LocalRule nestedRule1 = null;
    private LocalRule nestedRule2 = null;

    /** Denote a rule with no operator. */
    public static final int NO_OPERATOR = 0;

    /** Unary operator. */
    public static final int NOT = 1;

    /** Binary operator. */
    public static final int OR = 2;

    /** Binary operator. */
    public static final int AND = 3;

    // see constructor
    private final int unit;
    private final int part;
    private final String string;
    private final char separator;
    private final int status;
    private final String type;

    /** Type of part. */
    public static final int PREFIX = 0;

    /** Type of part. */
    public static final int SUFFIX = 1;

    /** Type of part. */
    public static final int CONTAIN = 2;

    /** Type of part. */
    public static final int TAG = 3;

    /** Type of part. */
    public static final int CONTAIN_REGEX = 4;

    /** Type of part. */
    public static final int TAG_REGEX = 5;
    
    /** Type of status */
    public static final int ACCEPTED  = 1;

    /** Type of status */
    public static final int REJECTED  = 2;

    /**
     * Local rule constructor for the operators NOT, AND and OR which don't
     * need to have filled all the fields of the object.
     *
     * @see LocalRule#LocalRule(int, LocalRule, LocalRule, int, int,
     * java.lang.String, char, int, String) full constructor
     */
    public LocalRule(int _operator,
                     LocalRule _nestedRule1, LocalRule _nestedRule2) {

        if (_operator < NO_OPERATOR || _operator > AND) {
            throw new IllegalArgumentException(
                "This operator doesn't exist.");
        }

        if ((_operator == NOT && _nestedRule1 == null)
            || (_operator == AND && _nestedRule1 == null
                && _nestedRule2 == null)
            || (_operator == OR && _nestedRule1 == null
                && _nestedRule2 == null)) {
                throw new NullPointerException(
                    "Nested LocalRule musn't be null.");
        }

        if ((_operator == NOT
               && (_nestedRule1.getStatus() < ACCEPTED
                || _nestedRule1.getStatus() > REJECTED
                || _nestedRule1.getType() == null))
            || ((_operator == AND || _operator == OR)
               && (_nestedRule1.getStatus() < ACCEPTED
                || _nestedRule1.getStatus() > REJECTED
                || _nestedRule1.getType() == null
                || _nestedRule2.getStatus() < ACCEPTED
                || _nestedRule2.getStatus() > REJECTED
                || _nestedRule2.getType() == null))) {
                throw new IllegalArgumentException(
                    "Status or type of the nested rules aren't valid.");
        }

        if ((_operator == AND || _operator == OR)
               && (_nestedRule1.getStatus() != _nestedRule2.getStatus()
                || _nestedRule1.getType() != _nestedRule2.getType())) {
            throw new IllegalArgumentException(
            "The nested rules must have the same status and type.");
        }

        // get the status and type from the nested rule
        if (_operator != NO_OPERATOR) {
            status = _nestedRule1.getStatus();
            type = _nestedRule1.getType();
        } else {
             status = Integer.MAX_VALUE;
             type = null;
        }

        operator = _operator;
        nestedRule1 = _nestedRule1;
        nestedRule2 = _nestedRule2;
        unit = Integer.MAX_VALUE;
        part = Integer.MAX_VALUE;
        string = null;
        separator = Character.MAX_VALUE;
    }

    /**
     * Creates a new instance of LocalRule.
     *
     * @param _operator AND, OR, NOT or NO_OPERATOR if this is the last
     * operand of the rule.
     * @param _nestedRule1 a nested LocalRule with AND, OR, NOT operator.
     * @param _nestedRule2 a nested LocalRule with AND, OR operator.
     * @param _unit unit of application of the rule. 1, 2 or 3 only if
     * units triple.
     * @param _part part of application of the rule. PREFIX, SUFFIX,
     * CONTAIN or TAG...
     * @param _string string to use with part of the rule to test the
     * Occurrence. Invariant : musn't contains SEPARATOR char.
     * @param _separator used to split unit in word and tag.
     * @param _status ACCEPTED or REJECTED status of the result of the rule
     * @param _type type of the Occurrence objects selected by the rule
     */
    public LocalRule(int _operator,
                     LocalRule _nestedRule1, LocalRule _nestedRule2,
                     int _unit, int _part, String _string,
                     char _separator, int _status, String _type) {
    
        if (_operator < NO_OPERATOR || _operator > AND) {
            throw new IllegalArgumentException(
                "This operator doesn't exist.");
        }
        
        if (_operator == NO_OPERATOR) {
            if (_unit < 1 || _unit > 3) {
                throw new IllegalArgumentException(
                    "Unit must be 1, 2 or 3.");
            }
        
            if (_part < PREFIX || _part > TAG_REGEX) {
                throw new IllegalArgumentException(
                    "Part must be PREFIX, SUFFIX, CONTAIN, TAG...");
            }

            if (_string.indexOf(_separator) != -1) {
                throw new IllegalArgumentException(
                    "String musn't contain a SEPARATOR character.");
            }

            if (_status < ACCEPTED || _status > REJECTED) {
                throw new IllegalArgumentException(
                    "Status must be ACCEPTED or REJECTED.");
            }

            if (_type == null || _type.trim().length() == 0) {
                throw new NullPointerException(
                    "Type musn't be null or the empty string.");
            }
        }

        if ((_operator == NOT && _nestedRule1 == null)
            || (_operator == AND && _nestedRule1 == null
                && _nestedRule2 == null)
            || (_operator == OR && _nestedRule1 == null
                && _nestedRule2 == null)) {
                throw new NullPointerException(
                    "Nested LocalRule musn't be null.");
        }

        operator = _operator;
        nestedRule1 = _nestedRule1;
        nestedRule2 = _nestedRule2;
        unit = _unit;
        part = _part;
        string = _string;
        separator = _separator;
        status = _status;
        type = _type;
    }

    /**
     * Returns the number of the unit where the rule is applied.
     * @return number of the unit
     */
    public int getUnit() {
        return unit;
    }

    /**
     * Returns the type of part that define how the rule is applied.
     * @return type of part
     */
    public int getPart() {
        return part;
    }

    /**
     * Returns the string to use with part of the rule.
     * @return the string of rule
     */
    public String getString() {
        return string;
    }

    /**
     * Returns the separator of the unit.
     * @return the separator of the unit
     */
    public int getSeparator() {
        return separator;
    }

    /**
     * Returns the status of the rule.
     * @return the status of the rule
     */
    public int getStatus() {
        return status;
    }

    /**
     * Returns the type of the rule.
     * @return the type of the rule
     */
    public String getType() {
        return type;
    }

    /**
     * Returns the operator of the rule.
     * @return the operator of the rule
     */
    public int getOperator() {
        return operator;
    }

    /**
     * Returns the first nested local rule of this rule.
     * @return the first nested local rule
     */
    public LocalRule getNestedRule1() {
        return nestedRule1;
    }

    /**
     * Returns the secaond nested local rule of this rule.
     * @return the second nested local rule
     */
    public LocalRule getNestedRule2() {
        return nestedRule2;
    }

    /**
     * Returns true if the rule apply to the Occurrence.
     *
     * @return true only and only if the rule apply to the Occurrence.
     * @param occ Occurrence onto apply rules
     */
    public boolean apply(Occurrence occ) {

        switch (operator) {
            case NO_OPERATOR:
                return this.applyRule(occ);

            case OR:
                return nestedRule1.apply(occ) || nestedRule2.apply(occ);

            case AND:
                return nestedRule1.apply(occ) && nestedRule2.apply(occ);

            case NOT:
                return !nestedRule1.apply(occ);
        }

        return false;
    }

    /**
     * Returns true if the partial rule apply to the Occurrence.
     *
     * @return true only and only if the rule apply to the Occurrence.
     * @param occ Occurrence onto apply rules
     */
    private boolean applyRule(Occurrence occ) {

        String unitContent, word, tag;
        int index;

        unitContent = "";
        // get the desired unit
        switch (unit) {
            case 1: unitContent = occ.getUnit1(); break;
            case 2: unitContent = occ.getUnit2(); break;
            case 3: unitContent = occ.getUnit3(); break;
        }
        
        // get the 2 strings around the separator
        // the word and the tag
        index = unitContent.indexOf(separator);
        word = unitContent.substring(0, index).trim();
        tag = unitContent.substring(index+1).trim();

        // test the part of the unit
        switch (part) {
            case CONTAIN:
                return (word.indexOf(string) != -1);

            case PREFIX:
                return word.startsWith(string);

            case SUFFIX:
                return word.endsWith(string);

            case TAG:
                return tag.equals(string);

            case CONTAIN_REGEX:
                return word.matches(string);

            case TAG_REGEX:
                return tag.matches(string);
        }

        return false;
    }

    /**
     * Returns the string representation of a LocalRule.
     *
     * @return the string representation of a LocalRule
     */
    public String toString() {

        String s;

        s = "";
        switch (operator) {
            case NO_OPERATOR:
                switch (part) {
                    case CONTAIN: s += "containing"; break;
                    case PREFIX: s += "prefixed"; break;
                    case SUFFIX: s += "suffixed"; break;
                    case TAG: s += "tagged"; break;
                    case CONTAIN_REGEX: s += "containing regex"; break;
                    case TAG_REGEX: s += "tagged regex"; break;
                }
                return "[unit "+unit+" "+s+" \""+string+"\"]";

            case OR:
                return "( "+nestedRule1.toString()+" || "
                    +nestedRule2.toString()+" )";

            case AND:
                return "( "+nestedRule1.toString()+" && "
                    +nestedRule2.toString()+" )";

            case NOT:
                return "! "+nestedRule1.toString();
        }
        return null;
    }

} // end of class LocalRule
