
package extractTerms;

import java.util.*;
import java.text.*;

/**
 * Contain an occurrence of units (String) pair or triple.
 *
 * <p>
 * Number of occurrence of units pair or triple (int), lines positions in
 * the input file (TreeSet of Integers) and names of text (Hashtable of
 * Integers) wich contain the occurrence.
 *
 * <p>
 * Relation (int), iteration (int) and typeName (String) of the occurrence.
 *
 * <p>
 * May contain a measure (double).
 *
 * @author Thomas Heitz for LRI Paris XI University
 * @version 0.4 17/06/2004
 * @since 18/07/2003
 */
public class Occurrence implements java.io.Serializable, Comparable {


    // see constructor
    private double measure;
    private int occ;
    private String unit1;
    private String unit2;
    private String unit3;
    private TreeSet lines;
    private Hashtable namesOfText;
    private int relation;
    private int iteration;
    private String typeName;

    /**
     * Partial Occurrence with just strings of units filled.
     *
     * @see #Occurrence(java.lang.String, java.lang.String,
     * java.lang.String, int, java.util.TreeSet, java.util.Hashtable,
     * double, int, int, String) full constructor
     */
    public Occurrence(String _unit1, String _unit2) {

        if (_unit1 == null || _unit2 == null) {
            throw new NullPointerException("Occurrence constructor required"
                      +" that the units parameters are not null.");
        }

        occ = 1;
        unit1 = _unit1;
        unit2 = _unit2;
        unit3 = null;
        lines = new TreeSet();
        namesOfText = new Hashtable();
        measure = Double.NaN;
        relation = -1;
        iteration = -1;
        typeName = null;
    }

    /**
     * Partial Occurrence with just strings of units filled.
     *
     * @see #Occurrence(java.lang.String, java.lang.String,
     * java.lang.String, int, java.util.TreeSet, java.util.Hashtable,
     * double, int, int, String) full constructor
     */
    public Occurrence(String _unit1, String _unit2, String _unit3) {

        if (_unit1 == null || _unit2 == null || _unit3 == null) {
            throw new NullPointerException("Occurrence constructor required"
                      +" that the units parameters are not null.");
        }

        occ = 1;
        unit1 = _unit1;
        unit2 = _unit2;
        unit3 = _unit3;
        lines = new TreeSet();
        namesOfText = new Hashtable();
        measure = Double.NaN;
        relation = -1;
        iteration = -1;
        typeName = null;
    }

    /**
     * Partial Occurrence with all fields filled.
     *
     * @see #Occurrence(java.lang.String, java.lang.String,
     * java.lang.String, int, java.util.TreeSet, java.util.Hashtable,
     * double, int, int, String) full constructor
     */
    public Occurrence(String _unit1, String _unit2,
                      int _occ, TreeSet _lines, Hashtable _nameOfText,
                      double _measure, int _relation, int _iteration,
                      String _typeName) {
        occ = _occ;
        unit1 = _unit1;
        unit2 = _unit2;
        unit3 = null;
        lines = _lines;
        namesOfText = _nameOfText;
        measure = _measure;
        relation = _relation;
        iteration = _iteration;
        typeName = _typeName;
    }

    /**
     * Full Occurrence with all fields filled.
     *
     * @param _unit1 string of the first unit of the occurrence ; couldn't
     * be null.
     * @param _unit2 string of the second unit of the occurrence ; couldn't
     * be null.
     * @param _unit3 string of the third unit of the occurrence ; couldn't
     * be null.
     * @param _occ number of occurrence for these concatenated
     * units ; superior or equal to 1.
     * @param _lines contain lines (Integer) of occurrence.
     * @param _nameOfText contain number of occurrence (Integer) indexed by
     * name of text ; invariant : the sum of all number of occurrence is
     * equal to occ field.
     * @param _measure contain the value of the measure.
     * @param _relation position of the element in the list of units
     * selector that was selected for this term
     * @param _iteration number of the iteration when this term was
     * extracted
     * @param _typeName three first letters of the type name according to
     * the list of types in the Expression panel
     */
    public Occurrence(String _unit1, String _unit2, String _unit3,
                      int _occ, TreeSet _lines, Hashtable _nameOfText,
                      double _measure, int _relation, int _iteration,
                      String _typeName) {
        occ = _occ;
        unit1 = _unit1;
        unit2 = _unit2;
        unit3 = _unit3;
        lines = _lines;
        namesOfText = _nameOfText;
        measure = _measure;
        relation = _relation;
        iteration = _iteration;
        typeName = _typeName;
    }

    /**
     * Returns the number of occurrence.
     *
     * @return the number occurence always superior or equal to 1.
     */
    public int getOcc() {
        return occ;
    }

    /**
     * Same as getOcc() but of Integer type.
     *
     * @return the number occurence of Integer type.
     * @see #getOcc()
     */
    public Integer getOccInt() {
        return new Integer(occ);
    }

    /**
     * Returns the first unit of the term.
     *
     * @return the non-null string of the first unit.
     */
    public String getUnit1() {
        return unit1;
    }

    /**
     * Returns the second unit of the term.
     *
     * @return the non-null string of the second unit.
     */
    public String getUnit2() {
        return unit2;
    }

    /**
     * Returns the third unit of the term.
     *
     * @return the string of the third unit ; it may be null if the term is
     * a couple of units.
     */
    public String getUnit3() {
        return unit3;
    }

    /**
     * Returns the concatenation of the term's units.
     *
     * @return the non-null string concatenation.
     */
    public String getKey() {
        if (unit3 == null) {
            return unit1+unit2;
        } else {
            return unit1+unit2+unit3;
        }
    }

    /**
     * Returns the value of the measure for this occurrence.
     *
     * @return the value of measure ; may be equal to Double.NaN if the
     * measure value of the term wasn't added.
     */
    public double getMeasure() {
        return measure;
    }

    /**
     * Same as getMeasure() but of Double type.
     *
     * @return the value measure of Double type.
     * @see #getMeasure()
     */
    public Double getMeasureDbl() {
        return new Double(measure);
    }

    /**
     * Returns the String of the formatted value of the measure for this
     * occurrence.
     *
     * @return the String value of measure ; may be the empty string if the
     * measure value of the term wasn't added.
     */
    public String getMeasureString() {

        if (measure == Double.NaN) {
            return "";
        } else {
            return new DecimalFormat("0.0").format(measure);
        }
    }

    /**
     * Returns the list of lines of this occurrence in the text where it was
     * extract.
     *
     * @return an ascending sorted list of lines (Integer) ; may be empty if
     * the lines of the occurrence wasn't added.
     */
    public TreeSet getLines() {
        return lines;
    }

    /**
     * Returns an iterator to read successively the value of lines of this
     * occurrence.
     *
     * @return an iterator of lines (Integer) ; may be empty if the lines of
     * the occurrence wasn't added.
     */
    public Iterator getLinesIterator() {
        return lines.iterator();
    }

    /**
     * Returns the table of names of text where occurrence appears. Each
     * name is linked to the number of occurrence for this text name.
     *
     * @return an hash table of number of occurrence (Integer) indexed by
     * name of text.
     */
    public Hashtable getNamesOfText() {
        return namesOfText;
    }

    /**
     * Returns the position of the element in the list of units selector
     * that was selected for this term.
     *
     * @return the number of the relation.
     */
    public int getRelation() {
        return relation;
    }

    /**
     * Returns the number of the iteration when this term was extracted.
     *
     * @return the number of the iteration.
     */
    public int getIteration() {
        return iteration;
    }

    /**
     * Returns the type name of this term.
     *
     * @return the type name of this term.
     */
    public String getTypeName() {
        return typeName;
    }

    /**
     * Returns true if it is a 2 units occurrence.
     *
     * @return true if and only if it is a 2 units occurrence ; invariant :
     * o.is2Units() != o.is3Units().
     */
    public boolean is2Units() {
        return unit3 == null;
    }

    /**
     * Returns true if it is a 3 units occurrence.
     *
     * @return true if and only if it is a 3 units occurrence ; invariant :
     * o.is2Units() != o.is3Units().
     */
    public boolean is3Units() {
        return unit3 != null;
    }

    /**
     * Returns true if this occurrence has at less one line in the list of
     * its lines where it appears.
     *
     * @return true if and only if one or more lines is contain in the list
     * of lines.
     */
    public boolean hasLines() {
        return (lines.size() > 0);
    }

    /**
     * Set the value of the measure.
     */
    public void setMeasure(double _measure) {
        measure = _measure;
    }

    /**
     * Increment by 1 the number of occurrence for this occurrence.
     *
     * @return the new number of occurrence.
     */
    public int incrementOcc() {
        return ++occ;
    }

    /**
     * Add a line number if it doesn't already exist.
     *
     * @return true if the set did not already contain the line.
     */
    public boolean addLine(int line) {
        return lines.add(new Integer(line));
    }

    /**
     * Add a name of text if it doesn't already exist.
     *
     * @return true if the Hashtable did not already contain the
     * _nameOfText.
     */
    public boolean addNameOfText(String _nameOfText) {
        Integer n = (Integer)namesOfText.get(_nameOfText);

        if (n == null) {
            namesOfText.put(_nameOfText, new Integer(1));
        } else {
            namesOfText.put(_nameOfText, new Integer(n.intValue()+1));
        }
        return (n == null);
    }

    /**
     * Set the number of the iteration.
     */
    public void setIteration(int _iteration) {
        iteration = _iteration;
    }

    /**
     * Set the number of the relation.
     */
    public void setRelation(int _relation) {
        relation = _relation;
    }

    /**
     * Set the name of the type.
     */
    public void setTypeName(String _typeName) {
        typeName = _typeName;
    }

    /**
     * Returns the string representation of an occurrence.
     *
     * @return the string representation ; only the units or if the format
     * is define the units plus the measure value, the number of occurrences
     * and the number of texts where present.
     */
    public String toString() {

        String s;

        s = "";
        s += this.getMeasureString()+" ";
        s += (this.is2Units()) ? unit1+unit2 : unit1+unit2+unit3;
        if (namesOfText.size() > 0) {
            s += " ("+occ+", "+namesOfText.size()+")";
        }
        if (typeName != null) { s += " ["+typeName+"]"; }

        return s;
    }

    /**
     * Returns the string before the separator contains in the unit.
     *
     * @param unit the unit to split
     * @param separator the separator used to split the unit in two
     * @return the string before the separator contains in the unit
     */
    public static String getWord(String unit, String separator) {

        int index = unit.indexOf(separator);
        if (index == -1) {
            return unit.trim();
        } else {
            return unit.substring(0, index).trim();
        }
    }

    /**
     * Returns the string after the separator contains in the unit.
     *
     * @param unit the unit to split
     * @param separator the separator used to split the unit in two
     * @return the string after the separator contains in the unit
     */
    public static String getTag(String unit, String separator) {

        int index = unit.indexOf(separator);
        if (index == -1) {
            return "";
        } else {
            return unit.substring(index+1).trim();
        }
    }

    /**
     * Compares an Occurrence with another Occurrence on their units content
     * in ascending alphabetical order.
     *
     * @param obj the reference object with which to compare.
     * @return a negative integer, zero, or a positive integer as this
     * object is less than, equal to, or greater than the specified object.
     */
    public int compareTo(Object obj) {

        Occurrence occ = (Occurrence)obj;

        if (!unit1.equals(occ.getUnit1())) {
            return unit1.compareTo(occ.getUnit1());

        } else if (!unit2.equals(occ.getUnit2())) {
            return unit2.compareTo(occ.getUnit2());

        } else if (this.is2Units() && occ.is2Units()) {
            return 0;

        } else if (this.is3Units() && occ.is3Units()) {

            if (!unit3.equals(occ.getUnit3())) {
                return unit3.compareTo(occ.getUnit3());
            } else {
                return 0;
            }

        } else { // comparison of 2 and 3 units Occurrence
            return -1;
        }
    }

    /**
     * Test the equality of 2 Occurrence objects on their units content.
     *
     * @param obj the reference object with which to compare.
     * @return true if this object is the same as the obj argument; false
     * otherwise.
     */
    public boolean equals(Object obj) {

        Occurrence occ = (Occurrence)obj;

        return (
            (this.is2Units() && occ.is2Units()
             && unit1.equals(occ.getUnit1())
             && unit2.equals(occ.getUnit2()))

         || (this.is3Units() && occ.is3Units()
             && unit1.equals(occ.getUnit1())
             && unit2.equals(occ.getUnit2())
             && unit3.equals(occ.getUnit3()))
        );
    }

} // end of class Occurrence
