
package extractTerms;

import java.io.*;
import java.nio.channels.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.text.*;
import java.util.*;
import java.util.regex.*;
import java.text.*;
import java.util.prefs.*;
import extractTerms.*;

/**
 * Graphical user interface for ExtractTerms class.
 *
 *<p>Main features are :
 *
 * <ul>
 * <li>Search of already formed terms with an hyphen</li>
 * <li>Import of lists of accepted/rejected terms previously to any
 * extraction</li>
 * <li>Edit/test/load/save all parameters of extraction</li>
 * <li>Creation of rules that are automatically applied to extraction
 * results</li>
 * <li>Take into account precedent choices to accept/reject new extracted
 * terms and undefined terms automatically</li>
 * <li>Display of context for any term or string</li>
 * <li>Automatic saving of all extracted terms and possibility to quit,
 * launch again the programm and to return to last state of extraction</li>
 * <li>Possibility to save extraction results and list of terms in ASCII
 * format</li>
 * <li>Modification of status of extracted terms at any moment</li>
 * <li>Multiple filters for the display of lists of terms</li>
 * <li>Contextual help and help under cursor</li>
 * <li>Multilingual interface (english, french currently)</li>
 * </ul>
 *
 * @author Thomas Heitz for LRI Paris XI University
 * @version 0.9 10/04/2004
 * @since 27/06/2003
 * @see ExtractTerms
 * @see LocalRule
 * @see Occurrence
 */
public class ExtractTermsGUI extends javax.swing.JFrame {

    /** Link to the main frame used in the dialog boxes. */
    private JFrame frame = this;

    /** Link to the help frame. */
    private JFrame helpFrame;

    /** Link to the currently used ExtractTerms object. */
    private ExtractTerms et;

    /** Preferences of the program */
    private Preferences prefs;

    /** Count of the current iteration. */
    private int iteration;

    /** Number of lines of the input file */
    private int inputFileLines;

    /** Group of language radio button */
    private ButtonGroup languageButtonGroup;

    /** Link to the ResourceBundle of this program. */
    private ResourceBundle messages;

    /** LinkedHashSet of Occurrence of accepted terms for each iteration in
     * order of iteration. Filled during extraction. */
    private LinkedHashSet acceptedTerms;

    /** Contain Occurrence of accepted terms for the current
     * iteration. Filled during extraction. Needed for grouping
     * of terms and cleared after grouping. */
    private LinkedHashSet currentAcceptedTerms;

    /** LinkedHashSet of Occurrence of rejected terms for each iteration in
     * order of iteration. Filled during extraction. */
    private LinkedHashSet rejectedTerms;

    /** LinkedHashSet of Occurrence of undefined terms for each iteration in
     * order of iteration. Filled during extraction. */
    private LinkedHashSet undefinedTerms;

    /** Contain the Strings of accepted terms for the current
     * iteration. Filled during extraction. Needed for removed these terms
     * from currentSkippedTermsTreeSet after grouping. Cleared after
     * grouping. */
    private TreeSet currentAcceptedTermsTreeSet;

    /** Contain the Strings of accepted, rejected and undefined terms for
     * all iterations but accepted terms only for the current
     * iteration. Filled during extraction. Needed for extraction. */
    private TreeSet currentSkippedTermsTreeSet;

    /** Contain the status of each row of the results table. */
    private int [] rowsStatus;

    /** Undefined status of a cell in the results table. */
    private static final int UNDEFINED = 0;

    /** Accepted status of a cell in the results table. */
    private static final int ACCEPTED  = 1;

    /** Rejected status of a cell in the results table. */
    private static final int REJECTED  = 2;

    /** Contain pair of LocalRule object and their result status. */
    private Vector localRules2Units;
    
    /** Contain pair of LocalRule object and their result status. */
    private Vector localRules3Units;


    /*-*************
     * Constructor *
     ***************/

    /**
     * Creates new form Terms with menus, tabbed panes and state
     * line. Contain all user interface controls to easily change the
     * parameters of the extraction of terms and to see the results.
     */
    public ExtractTermsGUI() {

        // preferences of the program
        prefs = Preferences.userNodeForPackage(ExtractTermsGUI.class);

        if (prefs.get("language", "automatic").equals("automatic")){
            // do nothing
        } else if (prefs.get("language", "english").equals("english")) {
            Locale.setDefault(new Locale("en", "EN"));
        } else if (prefs.get("language", "francais").equals("francais")) {
            Locale.setDefault(new Locale("fr", "FR"));
        }

        // used for internationalisation of text
        messages = ResourceBundle.getBundle("messages");

        // iteration
        iteration = 0;

        // number of lines of the input file
        inputFileLines = 0;

        // initialisation of the memory tables
        acceptedTerms = new LinkedHashSet();
        currentAcceptedTerms = new LinkedHashSet();
        rejectedTerms = new LinkedHashSet();
        undefinedTerms = new LinkedHashSet();
        currentAcceptedTermsTreeSet = new TreeSet();
        currentSkippedTermsTreeSet = new TreeSet();

        // initialisation of the rows status of the results table
        rowsStatus = new int [0];

        // initialisation of the rules
        localRules2Units = new Vector();
        localRules3Units = new Vector();

        // initialisation of graphical user interface components
        initComponents();

        // group the language menu items radio buttons
        languageButtonGroup = new ButtonGroup();
        languageButtonGroup.add(automaticjRadioButtonMenuItem);
        languageButtonGroup.add(englishjRadioButtonMenuItem);
        languageButtonGroup.add(francaisjRadioButtonMenuItem);
        // select interface language according to the system properties
        if (Locale.getDefault().equals(new Locale("fr", "FR"))) {
            francaisjRadioButtonMenuItem.setSelected(true);
        } else if (Locale.getDefault().equals(new Locale("en", "EN"))) {
            englishjRadioButtonMenuItem.setSelected(true);
        } else {
            automaticjRadioButtonMenuItem.setSelected(true);
        }

        // hide the menu items of the extraction table pop up menu
        for (int i = 0; i < acceptjMenu.getItemCount(); i++) {
            acceptjMenu.getItem(i).setVisible(false);
        }
        for (int i = 0; i < rejectjMenu.getItemCount(); i++) {
            rejectjMenu.getItem(i).setVisible(false);
        }

        // hide the extraction progress bar
        extractionjProgressBar.setVisible(false);

        // set help titles of tabbed panes
        mainjTabbedPane.setTitleAt(0, messages.getString("Expressions"));
        mainjTabbedPane.setTitleAt(1, messages.getString("Preprocessing"));
        mainjTabbedPane.setTitleAt(2, messages.getString("Extraction"));
        mainjTabbedPane.setTitleAt(3, messages.getString("Rules"));
        mainjTabbedPane.setTitleAt(4,
            messages.getString("Accepted/Rejected_Terms"));
        mainjTabbedPane.setTitleAt(5,
            messages.getString("Undefined_terms"));

        // set the tooltip texts of the tabbed indexes
        mainjTabbedPane.setToolTipTextAt(0, messages.getString(
        "Here_you_define_the_several_expressions_and_formulas_needed_"
        +"for_extraction."));
        mainjTabbedPane.setToolTipTextAt(1, messages.getString(
        "Operations_of_this_stage_are_executed_only_before_the_first_"
        +"iteration."));
        mainjTabbedPane.setToolTipTextAt(2, messages.getString(
        "The_main_panel_where_you_execute_and_control_the_extraction."));
        mainjTabbedPane.setToolTipTextAt(3, messages.getString(
        "Interface_of_rules_creation_that_are_applied_after_each_"
        +"extraction."));
        mainjTabbedPane.setToolTipTextAt(4, messages.getString(
        "Here_you_manage_the_accepted_and_rejected_terms_of_the_precedent_"
        +"extractions."));
        mainjTabbedPane.setToolTipTextAt(5, messages.getString(
        "Here_you_manage_the_undefined_terms_of_the_precedent_"
        +"extractions."));

        // select some elements of the interface at the start of the program
        mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
        statejLabel.setText(
            messages.getString("First_select_an_iteration."));

        // update files list
        updateFilesList();

        // preprocessing panel initialisation
        alreadyFormedTermsjList.setModel(new DefaultListModel());

        // expressions panel initialisation
        regexjTable.setModel(new ExpressionsTableModel(10, 6) {
            Class[] types = new Class [] {String.class, String.class,
                String.class, String.class, String.class, String.class};
            public Class getColumnClass(int columnIndex) {
                return types [columnIndex];
            }
        });
        regex2jTable.setModel(new ExpressionsTableModel(100, 9) {
            Class[] types = new Class [] {String.class, String.class,
                String.class, Integer.class, String.class, String.class,
                String.class, String.class, Integer.class};
            public Class getColumnClass(int columnIndex) {
                return types [columnIndex];
            }
        });
        formulasjTable.setModel(new ExpressionsTableModel(10, 2) {
            Class[] types = new Class [] {String.class, String.class};
            public Class getColumnClass(int columnIndex) {
                return types [columnIndex];
            }
        });
        termsTypesjTable.setModel(new ExpressionsTableModel(50, 2) {
            Class[] types = new Class [] {String.class, String.class};
            public Class getColumnClass(int columnIndex) {
                return types [columnIndex];
            }
        });
        regexjTable.setDefaultRenderer(regexjTable.getColumnClass(0),
                                       new ExpressionsCellRenderer());
        regex2jTable.setDefaultRenderer(regexjTable.getColumnClass(0),
                                       new ExpressionsCellRenderer());
        formulasjTable.setDefaultRenderer(regexjTable.getColumnClass(0),
                                       new ExpressionsCellRenderer());
        termsTypesjTable.setDefaultRenderer(regexjTable.getColumnClass(0),
                                       new ExpressionsCellRenderer());
        // name the tables columns
        ((DefaultTableModel)regexjTable.getModel())
            .setColumnIdentifiers(new String [] {
                messages.getString("Unit"),
                messages.getString("Definition"),
                messages.getString("Sentence/Phrase_separator"),
                messages.getString("Definition"),
                messages.getString("Name_of_text"),
                messages.getString("Selector")});

        ((DefaultTableModel)regex2jTable.getModel())
            .setColumnIdentifiers(new String [] {
                messages.getString("Units_pair"),
                messages.getString("Selector")+" 1",
                messages.getString("Selector")+" 2",
                messages.getString("Tag_kept"),
                messages.getString("Units_triple"),
                messages.getString("Selector")+" 1",
                messages.getString("Selector")+" 2",
                messages.getString("Selector")+" 3",
                messages.getString("Tag_kept")});

        ((DefaultTableModel)formulasjTable.getModel())
            .setColumnIdentifiers(new String [] {
                messages.getString("Name"),
                messages.getString("Formula")});

        ((DefaultTableModel)termsTypesjTable.getModel())
            .setColumnIdentifiers(new String [] {
                messages.getString("Accepted_types"),
                messages.getString("Rejected_types")});

        regexjTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        regex2jTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        formulasjTable.setSelectionMode(
            ListSelectionModel.SINGLE_SELECTION);

        // editors for the cells
        JComboBox number2JComboBox = new JComboBox();
        number2JComboBox.addItem(new Integer(1));
        number2JComboBox.addItem(new Integer(2));
        ((TableColumn)regex2jTable.getColumnModel().getColumn(3))
            .setCellEditor(new DefaultCellEditor(number2JComboBox));
        JComboBox number3JComboBox = new JComboBox();
        number3JComboBox.addItem(new Integer(1));
        number3JComboBox.addItem(new Integer(2));
        number3JComboBox.addItem(new Integer(3));
        ((TableColumn)regex2jTable.getColumnModel().getColumn(8))
            .setCellEditor(new DefaultCellEditor(number3JComboBox));

        // extraction panel initialisation
        // set the min, max, default values of the spinners
        nbIntegersSpinner.setModel(new SpinnerNumberModel(1, 1, 10, 1));
        nbDecimalsSpinner.setModel(new SpinnerNumberModel(2, 1, 10, 1));
        twoUnitsResultsSpinner.setModel(
            new SpinnerNumberModel(100, 10, 100000, 20));
        threeUnitsResultsSpinner.setModel(
            new SpinnerNumberModel(100, 10, 100000, 20));
        twoUnitsSpinner.setModel(new SpinnerNumberModel(4, 1, 100, 1));
        threeUnitsSpinner.setModel(new SpinnerNumberModel(4, 1, 100, 1));
        resultsjTable.getSelectionModel().addListSelectionListener(
           new ListSelectionListenerResultsTable(resultsjTable));
        optionsComboBox.setRenderer(new IterationCellRenderer());

        // rules panel initialisation
        lrUnitjComboBox.addItem(new Integer(1));
        lrUnitjComboBox.addItem(new Integer(2));
        lrUnitjComboBox.addItem(new Integer(3));
        lrPartjComboBox.addItem(messages.getString("Prefixed"));
        lrPartjComboBox.addItem(messages.getString("Suffixed"));
        lrPartjComboBox.addItem(messages.getString("Containing"));
        lrPartjComboBox.addItem(messages.getString("Tagged"));
        lrPartjComboBox.addItem(messages.getString("Containing_regex"));
        lrPartjComboBox.addItem(messages.getString("Tagged_regex"));
        localRules2UnitsjList.setModel(new DefaultListModel());
        localRules3UnitsjList.setModel(new DefaultListModel());

        // accepted/rejected terms panel initialisation
        acceptedTermsjTable.setModel(new UneditableDefaultTableModel());
        acceptedTermsjTable.getSelectionModel().addListSelectionListener(
           new ListSelectionListenerResultsTable(acceptedTermsjTable));

        rejectedTermsjTable.setModel(new UneditableDefaultTableModel());
        rejectedTermsjTable.getSelectionModel().addListSelectionListener(
           new ListSelectionListenerResultsTable(rejectedTermsjTable));

        // undefined terms panel initialisation
        undefinedTermsjTable.setModel(new UneditableDefaultTableModel());
        undefinedTermsjTable.getSelectionModel().addListSelectionListener(
           new ListSelectionListenerResultsTable(undefinedTermsjTable));

        // precedent terms JDialog
        precedentTermsjSpinner.setModel(
            new SpinnerNumberModel(1, 1, 50, 1));
        precedentTermsjDialog.pack();
    }


    /*-*********************************************
     * Initialisation of elements of the interface *
     ***********************************************/

    /**
     * This method is called from within the constructor to initialize the
     * form.
     *
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        resultsjPopupMenu = new javax.swing.JPopupMenu();
        acceptjMenu = new javax.swing.JMenu();
        acceptTypejMenuItem1 = new javax.swing.JMenuItem();
        acceptTypejMenuItem2 = new javax.swing.JMenuItem();
        acceptTypejMenuItem3 = new javax.swing.JMenuItem();
        acceptTypejMenuItem4 = new javax.swing.JMenuItem();
        acceptTypejMenuItem5 = new javax.swing.JMenuItem();
        acceptTypejMenuItem6 = new javax.swing.JMenuItem();
        acceptTypejMenuItem7 = new javax.swing.JMenuItem();
        acceptTypejMenuItem8 = new javax.swing.JMenuItem();
        acceptTypejMenuItem9 = new javax.swing.JMenuItem();
        acceptTypejMenuItem10 = new javax.swing.JMenuItem();
        acceptTypejMenuItem11 = new javax.swing.JMenuItem();
        acceptTypejMenuItem12 = new javax.swing.JMenuItem();
        acceptTypejMenuItem13 = new javax.swing.JMenuItem();
        acceptTypejMenuItem14 = new javax.swing.JMenuItem();
        acceptTypejMenuItem15 = new javax.swing.JMenuItem();
        acceptTypejMenuItem16 = new javax.swing.JMenuItem();
        acceptTypejMenuItem17 = new javax.swing.JMenuItem();
        acceptTypejMenuItem18 = new javax.swing.JMenuItem();
        acceptTypejMenuItem19 = new javax.swing.JMenuItem();
        acceptTypejMenuItem20 = new javax.swing.JMenuItem();
        acceptTypejMenuItem21 = new javax.swing.JMenuItem();
        acceptTypejMenuItem22 = new javax.swing.JMenuItem();
        acceptTypejMenuItem23 = new javax.swing.JMenuItem();
        acceptTypejMenuItem24 = new javax.swing.JMenuItem();
        acceptTypejMenuItem25 = new javax.swing.JMenuItem();
        acceptTypejMenuItem26 = new javax.swing.JMenuItem();
        acceptTypejMenuItem27 = new javax.swing.JMenuItem();
        acceptTypejMenuItem28 = new javax.swing.JMenuItem();
        acceptTypejMenuItem29 = new javax.swing.JMenuItem();
        acceptTypejMenuItem30 = new javax.swing.JMenuItem();
        acceptTypejMenuItem31 = new javax.swing.JMenuItem();
        acceptTypejMenuItem32 = new javax.swing.JMenuItem();
        acceptTypejMenuItem33 = new javax.swing.JMenuItem();
        acceptTypejMenuItem34 = new javax.swing.JMenuItem();
        acceptTypejMenuItem35 = new javax.swing.JMenuItem();
        acceptTypejMenuItem36 = new javax.swing.JMenuItem();
        acceptTypejMenuItem37 = new javax.swing.JMenuItem();
        acceptTypejMenuItem38 = new javax.swing.JMenuItem();
        acceptTypejMenuItem39 = new javax.swing.JMenuItem();
        acceptTypejMenuItem40 = new javax.swing.JMenuItem();
        acceptTypejMenuItem41 = new javax.swing.JMenuItem();
        acceptTypejMenuItem42 = new javax.swing.JMenuItem();
        acceptTypejMenuItem43 = new javax.swing.JMenuItem();
        acceptTypejMenuItem44 = new javax.swing.JMenuItem();
        acceptTypejMenuItem45 = new javax.swing.JMenuItem();
        acceptTypejMenuItem46 = new javax.swing.JMenuItem();
        acceptTypejMenuItem47 = new javax.swing.JMenuItem();
        acceptTypejMenuItem48 = new javax.swing.JMenuItem();
        acceptTypejMenuItem49 = new javax.swing.JMenuItem();
        acceptTypejMenuItem50 = new javax.swing.JMenuItem();
        rejectjMenu = new javax.swing.JMenu();
        rejectTypejMenuItem1 = new javax.swing.JMenuItem();
        rejectTypejMenuItem2 = new javax.swing.JMenuItem();
        rejectTypejMenuItem3 = new javax.swing.JMenuItem();
        rejectTypejMenuItem4 = new javax.swing.JMenuItem();
        rejectTypejMenuItem5 = new javax.swing.JMenuItem();
        rejectTypejMenuItem6 = new javax.swing.JMenuItem();
        rejectTypejMenuItem7 = new javax.swing.JMenuItem();
        rejectTypejMenuItem8 = new javax.swing.JMenuItem();
        rejectTypejMenuItem9 = new javax.swing.JMenuItem();
        rejectTypejMenuItem10 = new javax.swing.JMenuItem();
        rejectTypejMenuItem11 = new javax.swing.JMenuItem();
        rejectTypejMenuItem12 = new javax.swing.JMenuItem();
        rejectTypejMenuItem13 = new javax.swing.JMenuItem();
        rejectTypejMenuItem14 = new javax.swing.JMenuItem();
        rejectTypejMenuItem15 = new javax.swing.JMenuItem();
        rejectTypejMenuItem16 = new javax.swing.JMenuItem();
        rejectTypejMenuItem17 = new javax.swing.JMenuItem();
        rejectTypejMenuItem18 = new javax.swing.JMenuItem();
        rejectTypejMenuItem19 = new javax.swing.JMenuItem();
        rejectTypejMenuItem20 = new javax.swing.JMenuItem();
        rejectTypejMenuItem21 = new javax.swing.JMenuItem();
        rejectTypejMenuItem22 = new javax.swing.JMenuItem();
        rejectTypejMenuItem23 = new javax.swing.JMenuItem();
        rejectTypejMenuItem24 = new javax.swing.JMenuItem();
        rejectTypejMenuItem25 = new javax.swing.JMenuItem();
        rejectTypejMenuItem26 = new javax.swing.JMenuItem();
        rejectTypejMenuItem27 = new javax.swing.JMenuItem();
        rejectTypejMenuItem28 = new javax.swing.JMenuItem();
        rejectTypejMenuItem29 = new javax.swing.JMenuItem();
        rejectTypejMenuItem30 = new javax.swing.JMenuItem();
        rejectTypejMenuItem31 = new javax.swing.JMenuItem();
        rejectTypejMenuItem32 = new javax.swing.JMenuItem();
        rejectTypejMenuItem33 = new javax.swing.JMenuItem();
        rejectTypejMenuItem34 = new javax.swing.JMenuItem();
        rejectTypejMenuItem35 = new javax.swing.JMenuItem();
        rejectTypejMenuItem36 = new javax.swing.JMenuItem();
        rejectTypejMenuItem37 = new javax.swing.JMenuItem();
        rejectTypejMenuItem38 = new javax.swing.JMenuItem();
        rejectTypejMenuItem39 = new javax.swing.JMenuItem();
        rejectTypejMenuItem40 = new javax.swing.JMenuItem();
        rejectTypejMenuItem41 = new javax.swing.JMenuItem();
        rejectTypejMenuItem42 = new javax.swing.JMenuItem();
        rejectTypejMenuItem43 = new javax.swing.JMenuItem();
        rejectTypejMenuItem44 = new javax.swing.JMenuItem();
        rejectTypejMenuItem45 = new javax.swing.JMenuItem();
        rejectTypejMenuItem46 = new javax.swing.JMenuItem();
        rejectTypejMenuItem47 = new javax.swing.JMenuItem();
        rejectTypejMenuItem48 = new javax.swing.JMenuItem();
        rejectTypejMenuItem49 = new javax.swing.JMenuItem();
        rejectTypejMenuItem50 = new javax.swing.JMenuItem();
        undefinejMenuItem = new javax.swing.JMenuItem();
        searchStringjMenuItem = new javax.swing.JMenuItem();
        precedentTermsjDialog = new javax.swing.JDialog();
        jLabel23 = new javax.swing.JLabel();
        precedentTermsjSpinner = new javax.swing.JSpinner();
        precedentTermsAcceptjButton = new javax.swing.JButton();
        precedentTermsRejectjButton = new javax.swing.JButton();
        precedentTermsCanceljButton = new javax.swing.JButton();
        mainjTabbedPane = new javax.swing.JTabbedPane();
        regexFormulasjPanel = new javax.swing.JPanel();
        regexFormulasUtilsjPanel = new javax.swing.JPanel();
        jLabel26 = new javax.swing.JLabel();
        separatorjTextField = new javax.swing.JTextField();
        regexTestjToggleButton = new javax.swing.JToggleButton();
        regexHelpjButton = new javax.swing.JButton();
        regexHelpTagsjButton = new javax.swing.JButton();
        regexFormulasTablesjPanel = new javax.swing.JPanel();
        regexjScrollPane = new javax.swing.JScrollPane();
        regexjTable = new javax.swing.JTable();
        regex2jScrollPane = new javax.swing.JScrollPane();
        regex2jTable = new javax.swing.JTable();
        jPanel1 = new javax.swing.JPanel();
        formulasjScrollPane = new javax.swing.JScrollPane();
        formulasjTable = new javax.swing.JTable();
        termsTypesjScrollPane = new javax.swing.JScrollPane();
        termsTypesjTable = new javax.swing.JTable();
        preProcessingjPanel = new javax.swing.JPanel();
        preProcessing11jPanel = new javax.swing.JPanel();
        jLabel25 = new javax.swing.JLabel();
        alreadyFormedTermsjScrollPane = new javax.swing.JScrollPane();
        alreadyFormedTermsjList = new javax.swing.JList();
        preProcessing12jPanel = new javax.swing.JPanel();
        searchAlreadyFormedTermsjButton = new javax.swing.JButton();
        removeAlreadyFormedTermsjButton = new javax.swing.JButton();
        editTagsAlreadyFormedTermsjButton = new javax.swing.JButton();
        helpTagsAlreadyFormedTermsjButton = new javax.swing.JButton();
        preProcessing2jPanel = new javax.swing.JPanel();
        acceptListjComboBox = new javax.swing.JComboBox();
        jLabel17 = new javax.swing.JLabel();
        jLabel18 = new javax.swing.JLabel();
        rejectListjComboBox = new javax.swing.JComboBox();
        rejectListjButton = new javax.swing.JButton();
        acceptListjButton = new javax.swing.JButton();
        extractionjSplitPane = new javax.swing.JSplitPane();
        extraction2jSplitPane = new javax.swing.JSplitPane();
        resultsjScrollPane = new javax.swing.JScrollPane();
        resultsjTable = new javax.swing.JTable();
        optionsjPanel = new javax.swing.JPanel();
        extractButton = new javax.swing.JButton();
        precedentTermsjButton = new javax.swing.JButton();
        startAgainjButton = new javax.swing.JButton();
        jLabel19 = new javax.swing.JLabel();
        optionsDirjComboBox = new javax.swing.JComboBox();
        jLabel15 = new javax.swing.JLabel();
        optionsComboBox = new javax.swing.JComboBox();
        jLabel2 = new javax.swing.JLabel();
        outputFileComboBox = new javax.swing.JComboBox();
        outputFileChooserjButton = new javax.swing.JButton();
        savejButton = new javax.swing.JButton();
        jLabel3 = new javax.swing.JLabel();
        unitComboBox = new javax.swing.JComboBox();
        jLabel4 = new javax.swing.JLabel();
        sentPhrComboBox = new javax.swing.JComboBox();
        jLabel14 = new javax.swing.JLabel();
        nameOfTextjComboBox = new javax.swing.JComboBox();
        jLabel5 = new javax.swing.JLabel();
        twoUnitsComboBox = new javax.swing.JComboBox();
        jLabel6 = new javax.swing.JLabel();
        threeUnitsComboBox = new javax.swing.JComboBox();
        tfidfjCheckBox = new javax.swing.JCheckBox();
        jLabel7 = new javax.swing.JLabel();
        measuresjComboBox = new javax.swing.JComboBox();
        jLabel13 = new javax.swing.JLabel();
        jLabel10 = new javax.swing.JLabel();
        twoUnitsSpinner = new javax.swing.JSpinner();
        jLabel11 = new javax.swing.JLabel();
        threeUnitsSpinner = new javax.swing.JSpinner();
        jLabel16 = new javax.swing.JLabel();
        twoUnitsResultsSpinner = new javax.swing.JSpinner();
        threeUnitsResultsSpinner = new javax.swing.JSpinner();
        jLabel12 = new javax.swing.JLabel();
        jLabel8 = new javax.swing.JLabel();
        nbIntegersSpinner = new javax.swing.JSpinner();
        jLabel9 = new javax.swing.JLabel();
        nbDecimalsSpinner = new javax.swing.JSpinner();
        occFirstjCheckBox = new javax.swing.JCheckBox();
        contextjScrollPane = new javax.swing.JScrollPane();
        contextjTextPane = new javax.swing.JTextPane();
        rulesjPanel = new javax.swing.JPanel();
        localRulesjPanel = new javax.swing.JPanel();
        jLabel20 = new javax.swing.JLabel();
        localRulesListsjPanel = new javax.swing.JPanel();
        localRules2UnitsjScrollPane = new javax.swing.JScrollPane();
        localRules2UnitsjList = new javax.swing.JList();
        localRules3UnitsjScrollPane = new javax.swing.JScrollPane();
        localRules3UnitsjList = new javax.swing.JList();
        localRulesInputjPanel = new javax.swing.JPanel();
        lrTypejComboBox = new javax.swing.JComboBox();
        jLabel22 = new javax.swing.JLabel();
        lrUnitjComboBox = new javax.swing.JComboBox();
        lrPartjComboBox = new javax.swing.JComboBox();
        jLabel24 = new javax.swing.JLabel();
        lrStringjTextField = new javax.swing.JTextField();
        lrAdd2UnitsjButton = new javax.swing.JButton();
        lrAdd3UnitsjButton = new javax.swing.JButton();
        localRulesActionsjPanel = new javax.swing.JPanel();
        lrNotjButton = new javax.swing.JButton();
        lrAndjButton = new javax.swing.JButton();
        lrOrjButton = new javax.swing.JButton();
        lrDecomposejButton = new javax.swing.JButton();
        lrUpjButton = new javax.swing.JButton();
        lrDownjButton = new javax.swing.JButton();
        lrRemovejButton = new javax.swing.JButton();
        contextualRulesjPanel = new javax.swing.JPanel();
        jLabel21 = new javax.swing.JLabel();
        termsjSplitPane = new javax.swing.JSplitPane();
        termsjPanel = new javax.swing.JPanel();
        termsTablesjPanel = new javax.swing.JPanel();
        acceptedTermsjScrollPane = new javax.swing.JScrollPane();
        acceptedTermsjTable = new javax.swing.JTable();
        rejectedTermsjScrollPane = new javax.swing.JScrollPane();
        rejectedTermsjTable = new javax.swing.JTable();
        termsButtonsjPanel = new javax.swing.JPanel();
        acceptedTermsTypesjComboBox = new javax.swing.JComboBox();
        acceptedTermsRelationjComboBox = new javax.swing.JComboBox();
        acceptedTermsIterationjComboBox = new javax.swing.JComboBox();
        rejectedTermsRelationjComboBox = new javax.swing.JComboBox();
        rejectedTermsIterationjComboBox = new javax.swing.JComboBox();
        rejectedTermsTypesjComboBox = new javax.swing.JComboBox();
        termsContextjScrollPane = new javax.swing.JScrollPane();
        termsContextjTextPane = new javax.swing.JTextPane();
        undefinedTermsjSplitPane = new javax.swing.JSplitPane();
        undefinedTermsjPanel = new javax.swing.JPanel();
        undefinedTermsjScrollPane = new javax.swing.JScrollPane();
        undefinedTermsjTable = new javax.swing.JTable();
        undefinedTermsButtonsjPanel = new javax.swing.JPanel();
        undefinedTermsRelationjComboBox = new javax.swing.JComboBox();
        undefinedTermsIterationjComboBox = new javax.swing.JComboBox();
        precedentTermsUndefinedTermsjButton = new javax.swing.JButton();
        undefinedTermsContextjScrollPane = new javax.swing.JScrollPane();
        undefinedTermsContextjTextPane = new javax.swing.JTextPane();
        statejPanel = new javax.swing.JPanel();
        statejLabel = new javax.swing.JLabel();
        extractionjProgressBar = new javax.swing.JProgressBar();
        menuBar = new javax.swing.JMenuBar();
        programMenu = new javax.swing.JMenu();
        selectNewInputFilejMenuItem = new javax.swing.JMenuItem();
        saveResultsjMenuItem = new javax.swing.JMenuItem();
        languagejMenu = new javax.swing.JMenu();
        automaticjRadioButtonMenuItem = new javax.swing.JRadioButtonMenuItem();
        englishjRadioButtonMenuItem = new javax.swing.JRadioButtonMenuItem();
        francaisjRadioButtonMenuItem = new javax.swing.JRadioButtonMenuItem();
        exitMenuItem = new javax.swing.JMenuItem();
        displayjMenu = new javax.swing.JMenu();
        displayMeasuresjCheckBoxMenuItem = new javax.swing.JCheckBoxMenuItem();
        displayTagsjCheckBoxMenuItem = new javax.swing.JCheckBoxMenuItem();
        displayOccjCheckBoxMenuItem = new javax.swing.JCheckBoxMenuItem();
        displayExpertModejCheckBoxMenuItem = new javax.swing.JCheckBoxMenuItem();
        helpMenu = new javax.swing.JMenu();
        contexualHelpjMenuItem = new javax.swing.JMenuItem();
        contentsMenuItem = new javax.swing.JMenuItem();
        firstStartjMenuItem = new javax.swing.JMenuItem();
        helpUnderCursorjCheckBoxMenuItem = new javax.swing.JCheckBoxMenuItem();
        aboutMenuItem = new javax.swing.JMenuItem();

        acceptjMenu.setText(messages.getString("Accept"));
        acceptTypejMenuItem1.setText("Item");
        acceptTypejMenuItem1.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem1);

        acceptTypejMenuItem2.setText("Item");
        acceptTypejMenuItem2.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem2);

        acceptTypejMenuItem3.setText("Item");
        acceptTypejMenuItem3.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem3);

        acceptTypejMenuItem4.setText("Item");
        acceptTypejMenuItem4.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem4);

        acceptTypejMenuItem5.setText("Item");
        acceptTypejMenuItem5.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem5);

        acceptTypejMenuItem6.setText("Item");
        acceptTypejMenuItem6.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem6);

        acceptTypejMenuItem7.setText("Item");
        acceptTypejMenuItem7.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem7);

        acceptTypejMenuItem8.setText("Item");
        acceptTypejMenuItem8.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem8);

        acceptTypejMenuItem9.setText("Item");
        acceptTypejMenuItem9.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem9);

        acceptTypejMenuItem10.setText("Item");
        acceptTypejMenuItem10.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem10);

        acceptTypejMenuItem11.setText("Item");
        acceptTypejMenuItem11.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem11);

        acceptTypejMenuItem12.setText("Item");
        acceptTypejMenuItem12.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem12);

        acceptTypejMenuItem13.setText("Item");
        acceptTypejMenuItem13.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem13);

        acceptTypejMenuItem14.setText("Item");
        acceptTypejMenuItem14.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem14);

        acceptTypejMenuItem15.setText("Item");
        acceptTypejMenuItem15.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem15);

        acceptTypejMenuItem16.setText("Item");
        acceptTypejMenuItem16.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem16);

        acceptTypejMenuItem17.setText("Item");
        acceptTypejMenuItem17.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem17);

        acceptTypejMenuItem18.setText("Item");
        acceptTypejMenuItem18.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem18);

        acceptTypejMenuItem19.setText("Item");
        acceptTypejMenuItem19.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem19);

        acceptTypejMenuItem20.setText("Item");
        acceptTypejMenuItem20.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem20);

        acceptTypejMenuItem21.setText("Item");
        acceptTypejMenuItem21.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem21);

        acceptTypejMenuItem22.setText("Item");
        acceptTypejMenuItem22.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem22);

        acceptTypejMenuItem23.setText("Item");
        acceptTypejMenuItem23.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem23);

        acceptTypejMenuItem24.setText("Item");
        acceptTypejMenuItem24.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem24);

        acceptTypejMenuItem25.setText("Item");
        acceptTypejMenuItem25.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem25);

        acceptTypejMenuItem26.setText("Item");
        acceptTypejMenuItem26.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem26);

        acceptTypejMenuItem27.setText("Item");
        acceptTypejMenuItem27.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem27);

        acceptTypejMenuItem28.setText("Item");
        acceptTypejMenuItem28.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem28);

        acceptTypejMenuItem29.setText("Item");
        acceptTypejMenuItem29.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem29);

        acceptTypejMenuItem30.setText("Item");
        acceptTypejMenuItem30.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem30);

        acceptTypejMenuItem31.setText("Item");
        acceptTypejMenuItem31.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem31);

        acceptTypejMenuItem32.setText("Item");
        acceptTypejMenuItem32.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem32);

        acceptTypejMenuItem33.setText("Item");
        acceptTypejMenuItem33.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem33);

        acceptTypejMenuItem34.setText("Item");
        acceptTypejMenuItem34.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem34);

        acceptTypejMenuItem35.setText("Item");
        acceptTypejMenuItem35.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem35);

        acceptTypejMenuItem36.setText("Item");
        acceptTypejMenuItem36.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem36);

        acceptTypejMenuItem37.setText("Item");
        acceptTypejMenuItem37.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem37);

        acceptTypejMenuItem38.setText("Item");
        acceptTypejMenuItem38.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem38);

        acceptTypejMenuItem39.setText("Item");
        acceptTypejMenuItem39.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem39);

        acceptTypejMenuItem40.setText("Item");
        acceptTypejMenuItem40.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem40);

        acceptTypejMenuItem41.setText("Item");
        acceptTypejMenuItem41.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem41);

        acceptTypejMenuItem42.setText("Item");
        acceptTypejMenuItem42.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem42);

        acceptTypejMenuItem43.setText("Item");
        acceptTypejMenuItem43.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem43);

        acceptTypejMenuItem44.setText("Item");
        acceptTypejMenuItem44.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem44);

        acceptTypejMenuItem45.setText("Item");
        acceptTypejMenuItem45.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem45);

        acceptTypejMenuItem46.setText("Item");
        acceptTypejMenuItem46.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem46);

        acceptTypejMenuItem47.setText("Item");
        acceptTypejMenuItem47.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem47);

        acceptTypejMenuItem48.setText("Item");
        acceptTypejMenuItem48.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem48);

        acceptTypejMenuItem49.setText("Item");
        acceptTypejMenuItem49.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem49);

        acceptTypejMenuItem50.setText("Item");
        acceptTypejMenuItem50.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptTypejMenuItemActionPerformed(evt);
            }
        });

        acceptjMenu.add(acceptTypejMenuItem50);

        resultsjPopupMenu.add(acceptjMenu);

        rejectjMenu.setText(messages.getString("Reject"));
        rejectTypejMenuItem1.setText("Item");
        rejectTypejMenuItem1.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem1);

        rejectTypejMenuItem2.setText("Item");
        rejectTypejMenuItem2.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem2);

        rejectTypejMenuItem3.setText("Item");
        rejectTypejMenuItem3.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem3);

        rejectTypejMenuItem4.setText("Item");
        rejectTypejMenuItem4.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem4);

        rejectTypejMenuItem5.setText("Item");
        rejectTypejMenuItem5.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem5);

        rejectTypejMenuItem6.setText("Item");
        rejectTypejMenuItem6.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem6);

        rejectTypejMenuItem7.setText("Item");
        rejectTypejMenuItem7.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem7);

        rejectTypejMenuItem8.setText("Item");
        rejectTypejMenuItem8.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem8);

        rejectTypejMenuItem9.setText("Item");
        rejectTypejMenuItem9.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem9);

        rejectTypejMenuItem10.setText("Item");
        rejectTypejMenuItem10.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem10);

        rejectTypejMenuItem11.setText("Item");
        rejectTypejMenuItem11.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem11);

        rejectTypejMenuItem12.setText("Item");
        rejectTypejMenuItem12.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem12);

        rejectTypejMenuItem13.setText("Item");
        rejectTypejMenuItem13.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem13);

        rejectTypejMenuItem14.setText("Item");
        rejectTypejMenuItem14.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem14);

        rejectTypejMenuItem15.setText("Item");
        rejectTypejMenuItem15.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem15);

        rejectTypejMenuItem16.setText("Item");
        rejectTypejMenuItem16.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem16);

        rejectTypejMenuItem17.setText("Item");
        rejectTypejMenuItem17.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem17);

        rejectTypejMenuItem18.setText("Item");
        rejectTypejMenuItem18.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem18);

        rejectTypejMenuItem19.setText("Item");
        rejectTypejMenuItem19.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem19);

        rejectTypejMenuItem20.setText("Item");
        rejectTypejMenuItem20.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem20);

        rejectTypejMenuItem21.setText("Item");
        rejectTypejMenuItem21.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem21);

        rejectTypejMenuItem22.setText("Item");
        rejectTypejMenuItem22.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem22);

        rejectTypejMenuItem23.setText("Item");
        rejectTypejMenuItem23.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem23);

        rejectTypejMenuItem24.setText("Item");
        rejectTypejMenuItem24.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem24);

        rejectTypejMenuItem25.setText("Item");
        rejectTypejMenuItem25.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem25);

        rejectTypejMenuItem26.setText("Item");
        rejectTypejMenuItem26.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem26);

        rejectTypejMenuItem27.setText("Item");
        rejectTypejMenuItem27.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem27);

        rejectTypejMenuItem28.setText("Item");
        rejectTypejMenuItem28.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem28);

        rejectTypejMenuItem29.setText("Item");
        rejectTypejMenuItem29.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem29);

        rejectTypejMenuItem30.setText("Item");
        rejectTypejMenuItem30.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem30);

        rejectTypejMenuItem31.setText("Item");
        rejectTypejMenuItem31.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem31);

        rejectTypejMenuItem32.setText("Item");
        rejectTypejMenuItem32.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem32);

        rejectTypejMenuItem33.setText("Item");
        rejectTypejMenuItem33.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem33);

        rejectTypejMenuItem34.setText("Item");
        rejectTypejMenuItem34.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem34);

        rejectTypejMenuItem35.setText("Item");
        rejectTypejMenuItem35.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem35);

        rejectTypejMenuItem36.setText("Item");
        rejectTypejMenuItem36.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem36);

        rejectTypejMenuItem37.setText("Item");
        rejectTypejMenuItem37.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem37);

        rejectTypejMenuItem38.setText("Item");
        rejectTypejMenuItem38.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem38);

        rejectTypejMenuItem39.setText("Item");
        rejectTypejMenuItem39.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem39);

        rejectTypejMenuItem40.setText("Item");
        rejectTypejMenuItem40.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem40);

        rejectTypejMenuItem41.setText("Item");
        rejectTypejMenuItem41.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem41);

        rejectTypejMenuItem42.setText("Item");
        rejectTypejMenuItem42.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem42);

        rejectTypejMenuItem43.setText("Item");
        rejectTypejMenuItem43.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem43);

        rejectTypejMenuItem44.setText("Item");
        rejectTypejMenuItem44.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem44);

        rejectTypejMenuItem45.setText("Item");
        rejectTypejMenuItem45.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem45);

        rejectTypejMenuItem46.setText("Item");
        rejectTypejMenuItem46.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem46);

        rejectTypejMenuItem47.setText("Item");
        rejectTypejMenuItem47.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem47);

        rejectTypejMenuItem48.setText("Item");
        rejectTypejMenuItem48.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem48);

        rejectTypejMenuItem49.setText("Item");
        rejectTypejMenuItem49.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem49);

        rejectTypejMenuItem50.setText("Item");
        rejectTypejMenuItem50.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectTypejMenuItemActionPerformed(evt);
            }
        });

        rejectjMenu.add(rejectTypejMenuItem50);

        resultsjPopupMenu.add(rejectjMenu);

        undefinejMenuItem.setText(messages.getString("Undefine"));
        undefinejMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                undefinejMenuItemActionPerformed(evt);
            }
        });

        resultsjPopupMenu.add(undefinejMenuItem);

        searchStringjMenuItem.setText(messages.getString("Search"));
        searchStringjMenuItem.setToolTipText(messages.getString("search_a_string_in_the_input_file"));
        searchStringjMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                searchStringjMenuItemActionPerformed(evt);
            }
        });

        resultsjPopupMenu.add(searchStringjMenuItem);

        precedentTermsjDialog.getContentPane().setLayout(new java.awt.FlowLayout());

        precedentTermsjDialog.setTitle(messages.getString("Precedent_terms"));
        precedentTermsjDialog.setLocationRelativeTo(this);
        precedentTermsjDialog.setModal(true);
        precedentTermsjDialog.setResizable(false);
        jLabel23.setText(messages.getString("Threshold"));
        precedentTermsjDialog.getContentPane().add(jLabel23);

        precedentTermsjDialog.getContentPane().add(precedentTermsjSpinner);

        precedentTermsAcceptjButton.setText(messages.getString("Accept"));
        precedentTermsAcceptjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                precedentTermsAcceptjButtonActionPerformed(evt);
            }
        });

        precedentTermsjDialog.getContentPane().add(precedentTermsAcceptjButton);

        precedentTermsRejectjButton.setText(messages.getString("Reject"));
        precedentTermsRejectjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                precedentTermsRejectjButtonActionPerformed(evt);
            }
        });

        precedentTermsjDialog.getContentPane().add(precedentTermsRejectjButton);

        precedentTermsCanceljButton.setText(messages.getString("Cancel"));
        precedentTermsCanceljButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                precedentTermsCanceljButtonActionPerformed(evt);
            }
        });

        precedentTermsjDialog.getContentPane().add(precedentTermsCanceljButton);

        setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        setTitle(messages.getString("EXIT"));
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                exitForm(evt);
            }
            public void windowOpened(java.awt.event.WindowEvent evt) {
                formWindowOpened(evt);
            }
        });

        mainjTabbedPane.setPreferredSize(new java.awt.Dimension(900, 675));
        mainjTabbedPane.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                mainjTabbedPaneStateChanged(evt);
            }
        });

        regexFormulasjPanel.setLayout(new java.awt.BorderLayout());

        regexFormulasjPanel.setToolTipText(messages.getString("Here_you_define_the_several_expressions_and_formulas_needed_for_extraction."));
        regexFormulasUtilsjPanel.setPreferredSize(new java.awt.Dimension(10, 40));
        jLabel26.setText(messages.getString("Separator"));
        regexFormulasUtilsjPanel.add(jLabel26);

        separatorjTextField.setColumns(1);
        separatorjTextField.setToolTipText(messages.getString("separator_between_word_and_tag"));
        regexFormulasUtilsjPanel.add(separatorjTextField);

        regexTestjToggleButton.setMnemonic(messages.getString("Test_regular_expression").charAt(0));
        regexTestjToggleButton.setText(messages.getString("Test_regular_expression"));
        regexTestjToggleButton.setToolTipText(messages.getString("test_selected_regular_expression_on_the_first_10"));
        regexTestjToggleButton.setPreferredSize(new java.awt.Dimension(230, 26));
        regexTestjToggleButton.setRolloverEnabled(true);
        regexTestjToggleButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                regexTestjToggleButtonActionPerformed(evt);
            }
        });

        regexFormulasUtilsjPanel.add(regexTestjToggleButton);

        regexHelpjButton.setMnemonic(messages.getString("Help_on_regular_expressions").charAt(0));
        regexHelpjButton.setText(messages.getString("Help_on_regular_expressions"));
        regexHelpjButton.setPreferredSize(new java.awt.Dimension(240, 26));
        regexHelpjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                regexHelpjButtonActionPerformed(evt);
            }
        });

        regexFormulasUtilsjPanel.add(regexHelpjButton);

        regexHelpTagsjButton.setMnemonic(messages.getString("Help_on_tags").charAt(0));
        regexHelpTagsjButton.setText(messages.getString("Help_on_tags"));
        regexHelpTagsjButton.setPreferredSize(new java.awt.Dimension(180, 26));
        regexHelpTagsjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                regexHelpTagsjButtonActionPerformed(evt);
            }
        });

        regexFormulasUtilsjPanel.add(regexHelpTagsjButton);

        regexFormulasjPanel.add(regexFormulasUtilsjPanel, java.awt.BorderLayout.NORTH);

        regexFormulasTablesjPanel.setLayout(new java.awt.GridLayout(3, 1));

        regexjScrollPane.setMinimumSize(new java.awt.Dimension(200, 200));
        regexjTable.setFont(new java.awt.Font("Courier New", 0, 12));
        regexjTable.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {
                {null, null, null, null, null, null},
                {null, null, null, null, null, null},
                {null, null, null, null, null, null},
                {null, null, null, null, null, null},
                {null, null, null, null, null, null},
                {null, null, null, null, null, null},
                {null, null, null, null, null, null},
                {null, null, null, null, null, null},
                {null, null, null, null, null, null},
                {null, null, null, null, null, null}
            },
            new String [] {
                "Unit", "Definition", "Sentence/Phrase separator", "Definition", "Name of text", "Selector"
            }
        ) {
            Class[] types = new Class [] {
                java.lang.String.class, java.lang.String.class, java.lang.String.class, java.lang.String.class, java.lang.String.class, java.lang.String.class
            };

            public Class getColumnClass(int columnIndex) {
                return types [columnIndex];
            }
        });
        regexjTable.setMinimumSize(new java.awt.Dimension(200, 200));
        regexjTable.setPreferredSize(new java.awt.Dimension(525, 160));
        regexjTable.setRowSelectionAllowed(false);
        regexjScrollPane.setViewportView(regexjTable);

        regexFormulasTablesjPanel.add(regexjScrollPane);

        regex2jTable.setFont(new java.awt.Font("Courier New", 0, 12));
        regex2jTable.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null},
                {null, null, null, null, null, null, null, null, null}
            },
            new String [] {
                "Units pair", "Selector 1", "Selector 2", "Keep", "Units triple", "Selector 1", "Selector 2", "Selector 3", "Keep"
            }
        ) {
            Class[] types = new Class [] {
                java.lang.String.class, java.lang.String.class, java.lang.String.class, java.lang.Integer.class, java.lang.String.class, java.lang.String.class, java.lang.String.class, java.lang.String.class, java.lang.Integer.class
            };

            public Class getColumnClass(int columnIndex) {
                return types [columnIndex];
            }
        });
        regex2jTable.setRowSelectionAllowed(false);
        regex2jScrollPane.setViewportView(regex2jTable);

        regexFormulasTablesjPanel.add(regex2jScrollPane);

        jPanel1.setLayout(new java.awt.GridLayout(1, 2));

        formulasjTable.setFont(new java.awt.Font("Courier New", 0, 12));
        formulasjTable.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null}
            },
            new String [] {
                "Name", "Formula"
            }
        ) {
            Class[] types = new Class [] {
                java.lang.String.class, java.lang.String.class
            };

            public Class getColumnClass(int columnIndex) {
                return types [columnIndex];
            }
        });
        formulasjTable.setPreferredSize(null);
        formulasjTable.setRowSelectionAllowed(false);
        formulasjScrollPane.setViewportView(formulasjTable);

        jPanel1.add(formulasjScrollPane);

        termsTypesjTable.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null},
                {null, null}
            },
            new String [] {
                "Accepted types", "Rejected types"
            }
        ));
        termsTypesjScrollPane.setViewportView(termsTypesjTable);

        jPanel1.add(termsTypesjScrollPane);

        regexFormulasTablesjPanel.add(jPanel1);

        regexFormulasjPanel.add(regexFormulasTablesjPanel, java.awt.BorderLayout.CENTER);

        mainjTabbedPane.addTab(messages.getString("Expressions"), regexFormulasjPanel);

        preProcessingjPanel.setLayout(new java.awt.GridLayout(1, 2));

        preProcessingjPanel.setToolTipText(messages.getString("Operations_of_this_stage_are_executed_only_before_the_first_iteration."));
        preProcessing11jPanel.setLayout(new java.awt.BorderLayout());

        jLabel25.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
        jLabel25.setText(messages.getString("Terms_already_formed"));
        preProcessing11jPanel.add(jLabel25, java.awt.BorderLayout.NORTH);

        alreadyFormedTermsjScrollPane.setViewportView(alreadyFormedTermsjList);

        preProcessing11jPanel.add(alreadyFormedTermsjScrollPane, java.awt.BorderLayout.CENTER);

        preProcessing12jPanel.setLayout(new java.awt.GridLayout(2, 2));

        searchAlreadyFormedTermsjButton.setMnemonic(messages.getString("Search_terms").charAt(0));
        searchAlreadyFormedTermsjButton.setText(messages.getString("Search_terms"));
        searchAlreadyFormedTermsjButton.setToolTipText(messages.getString("search_terms_already_formed_in_the_input_file"));
        searchAlreadyFormedTermsjButton.setPreferredSize(new java.awt.Dimension(170, 26));
        searchAlreadyFormedTermsjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                searchAlreadyFormedTermsjButtonActionPerformed(evt);
            }
        });

        preProcessing12jPanel.add(searchAlreadyFormedTermsjButton);

        removeAlreadyFormedTermsjButton.setMnemonic(messages.getString("Remove_terms").charAt(0));
        removeAlreadyFormedTermsjButton.setText(messages.getString("Remove_terms"));
        removeAlreadyFormedTermsjButton.setToolTipText(messages.getString("remove_selected_terms_in_this_list"));
        removeAlreadyFormedTermsjButton.setMinimumSize(new java.awt.Dimension(185, 26));
        removeAlreadyFormedTermsjButton.setPreferredSize(new java.awt.Dimension(150, 26));
        removeAlreadyFormedTermsjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                removeAlreadyFormedTermsjButtonActionPerformed(evt);
            }
        });

        preProcessing12jPanel.add(removeAlreadyFormedTermsjButton);

        editTagsAlreadyFormedTermsjButton.setMnemonic(messages.getString("Edit_tags").charAt(0));
        editTagsAlreadyFormedTermsjButton.setText(messages.getString("Edit_tags"));
        editTagsAlreadyFormedTermsjButton.setToolTipText(messages.getString("Edit_the_tags_of_the_selected_terms"));
        editTagsAlreadyFormedTermsjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                editTagsAlreadyFormedTermsjButtonActionPerformed(evt);
            }
        });

        preProcessing12jPanel.add(editTagsAlreadyFormedTermsjButton);

        helpTagsAlreadyFormedTermsjButton.setMnemonic(messages.getString("Help_on_tags").charAt(0));
        helpTagsAlreadyFormedTermsjButton.setText(messages.getString("Help_on_tags"));
        helpTagsAlreadyFormedTermsjButton.setPreferredSize(new java.awt.Dimension(180, 26));
        helpTagsAlreadyFormedTermsjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                helpTagsAlreadyFormedTermsjButtonActionPerformed(evt);
            }
        });

        preProcessing12jPanel.add(helpTagsAlreadyFormedTermsjButton);

        preProcessing11jPanel.add(preProcessing12jPanel, java.awt.BorderLayout.SOUTH);

        preProcessingjPanel.add(preProcessing11jPanel);

        preProcessing2jPanel.setLayout(null);

        acceptListjComboBox.setEditable(true);
        acceptListjComboBox.setToolTipText(messages.getString("file_containing_list_of_terms_to_accept"));
        preProcessing2jPanel.add(acceptListjComboBox);
        acceptListjComboBox.setBounds(30, 220, 260, 28);

        jLabel17.setText(messages.getString("Accept_list"));
        preProcessing2jPanel.add(jLabel17);
        jLabel17.setBounds(100, 190, 170, 19);

        jLabel18.setText(messages.getString("Reject_list"));
        preProcessing2jPanel.add(jLabel18);
        jLabel18.setBounds(100, 270, 170, 19);

        rejectListjComboBox.setEditable(true);
        rejectListjComboBox.setToolTipText(messages.getString("file_containing_list_of_terms_to_reject"));
        preProcessing2jPanel.add(rejectListjComboBox);
        rejectListjComboBox.setBounds(30, 310, 260, 28);

        rejectListjButton.setText("...");
        rejectListjButton.setToolTipText(messages.getString("Display_file_chooser"));
        rejectListjButton.setMargin(new java.awt.Insets(1, 1, 1, 1));
        rejectListjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectListjButtonActionPerformed(evt);
            }
        });

        preProcessing2jPanel.add(rejectListjButton);
        rejectListjButton.setBounds(290, 310, 20, 27);

        acceptListjButton.setText("...");
        acceptListjButton.setToolTipText(messages.getString("Display_file_chooser"));
        acceptListjButton.setMargin(new java.awt.Insets(1, 1, 1, 1));
        acceptListjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptListjButtonActionPerformed(evt);
            }
        });

        preProcessing2jPanel.add(acceptListjButton);
        acceptListjButton.setBounds(290, 220, 20, 27);

        preProcessingjPanel.add(preProcessing2jPanel);

        mainjTabbedPane.addTab(messages.getString("Preprocessing"), preProcessingjPanel);

        extractionjSplitPane.setDividerSize(5);
        extractionjSplitPane.setOrientation(javax.swing.JSplitPane.VERTICAL_SPLIT);
        extractionjSplitPane.setToolTipText(messages.getString("The_main_panel_where_you_execute_and_control_the_extraction."));
        extraction2jSplitPane.setDividerSize(5);
        extraction2jSplitPane.setResizeWeight(0.5);
        extraction2jSplitPane.setPreferredSize(new java.awt.Dimension(880, 540));
        resultsjTable.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {

            },
            new String [] {

            }
        ));
        resultsjTable.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                resultsjTableKeyReleased(evt);
            }
        });
        resultsjTable.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseReleased(java.awt.event.MouseEvent evt) {
                resultsjTableMouseReleased(evt);
            }
        });

        resultsjScrollPane.setViewportView(resultsjTable);

        extraction2jSplitPane.setLeftComponent(resultsjScrollPane);

        optionsjPanel.setLayout(null);

        optionsjPanel.setPreferredSize(new java.awt.Dimension(425, 520));
        extractButton.setMnemonic(messages.getString("Extract_terms").charAt(0));
        extractButton.setText(messages.getString("Extract_terms"));
        extractButton.setToolTipText(messages.getString("launch_the_analyse_of_the_input_file"));
        extractButton.setBorder(new javax.swing.border.EtchedBorder());
        extractButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                extractButtonActionPerformed(evt);
            }
        });

        optionsjPanel.add(extractButton);
        extractButton.setBounds(10, 10, 140, 30);

        precedentTermsjButton.setMnemonic(messages.getString("Precedent_terms").charAt(0));
        precedentTermsjButton.setText(messages.getString("Precedent_terms"));
        precedentTermsjButton.setToolTipText(messages.getString("Use_the_precedent_accepted_and_rejected_terms_to_accept_and_reject_new_terms."));
        precedentTermsjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                precedentTermsjButtonActionPerformed(evt);
            }
        });

        optionsjPanel.add(precedentTermsjButton);
        precedentTermsjButton.setBounds(290, 10, 140, 30);

        startAgainjButton.setMnemonic(messages.getString("Start_again").charAt(0));
        startAgainjButton.setText(messages.getString("Start_again"));
        startAgainjButton.setToolTipText(messages.getString("Start_the_extraction_again_form_the_beginning."));
        startAgainjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                startAgainjButtonActionPerformed(evt);
            }
        });

        optionsjPanel.add(startAgainjButton);
        startAgainjButton.setBounds(160, 10, 120, 30);

        jLabel19.setText(messages.getString("Input"));
        optionsjPanel.add(jLabel19);
        jLabel19.setBounds(10, 60, 90, 19);

        optionsDirjComboBox.setToolTipText(messages.getString("Directory_of_input_file"));
        optionsDirjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                optionsDirjComboBoxActionPerformed(evt);
            }
        });

        optionsjPanel.add(optionsDirjComboBox);
        optionsDirjComboBox.setBounds(110, 50, 310, 28);

        jLabel15.setText(messages.getString("Iterations"));
        optionsjPanel.add(jLabel15);
        jLabel15.setBounds(10, 100, 90, 19);

        optionsComboBox.setFont(new java.awt.Font("Courier New", 0, 12));
        optionsComboBox.setToolTipText(messages.getString("Total_number_of_iterations."));
        optionsComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                optionsComboBoxActionPerformed(evt);
            }
        });

        optionsjPanel.add(optionsComboBox);
        optionsComboBox.setBounds(110, 90, 310, 28);

        jLabel2.setText(messages.getString("Results"));
        optionsjPanel.add(jLabel2);
        jLabel2.setBounds(10, 140, 90, 19);

        outputFileComboBox.setEditable(true);
        outputFileComboBox.setFont(new java.awt.Font("Courier New", 0, 12));
        outputFileComboBox.setToolTipText(messages.getString("name_of_output_file_containing_results"));
        optionsjPanel.add(outputFileComboBox);
        outputFileComboBox.setBounds(100, 130, 290, 28);

        outputFileChooserjButton.setText("...");
        outputFileChooserjButton.setToolTipText(messages.getString("Display_file_chooser"));
        outputFileChooserjButton.setMargin(new java.awt.Insets(1, 1, 1, 1));
        outputFileChooserjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                outputFileChooserjButtonActionPerformed(evt);
            }
        });

        optionsjPanel.add(outputFileChooserjButton);
        outputFileChooserjButton.setBounds(390, 130, 20, 27);

        savejButton.setText(messages.getString("Save").substring(0,1));
        savejButton.setToolTipText(messages.getString("save_the_results_in_file"));
        savejButton.setMargin(new java.awt.Insets(1, 1, 1, 1));
        savejButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                savejButtonActionPerformed(evt);
            }
        });

        optionsjPanel.add(savejButton);
        savejButton.setBounds(410, 130, 20, 27);

        jLabel3.setText(messages.getString("Unit_definition"));
        optionsjPanel.add(jLabel3);
        jLabel3.setBounds(150, 170, 160, 19);

        unitComboBox.setFont(new java.awt.Font("Courier New", 0, 12));
        unitComboBox.setToolTipText(messages.getString("regular_expression_used_for_splitting"));
        optionsjPanel.add(unitComboBox);
        unitComboBox.setBounds(20, 190, 390, 28);

        jLabel4.setText(messages.getString("Sentence_phrase_separators_definition"));
        optionsjPanel.add(jLabel4);
        jLabel4.setBounds(70, 220, 320, 19);

        sentPhrComboBox.setFont(new java.awt.Font("Courier New", 0, 12));
        sentPhrComboBox.setToolTipText(messages.getString("regular_expression_for_separator_of_sentence"));
        optionsjPanel.add(sentPhrComboBox);
        sentPhrComboBox.setBounds(20, 240, 390, 28);

        jLabel14.setText(messages.getString("Name_of_text_selector"));
        optionsjPanel.add(jLabel14);
        jLabel14.setBounds(140, 270, 220, 19);

        nameOfTextjComboBox.setFont(new java.awt.Font("Courier New", 0, 12));
        nameOfTextjComboBox.setToolTipText(messages.getString("regular_expression_used_for_extract_name_of_file"));
        optionsjPanel.add(nameOfTextjComboBox);
        nameOfTextjComboBox.setBounds(20, 290, 390, 28);

        jLabel5.setText(messages.getString("Units_pair_selector"));
        optionsjPanel.add(jLabel5);
        jLabel5.setBounds(130, 320, 200, 19);

        twoUnitsComboBox.setFont(new java.awt.Font("Courier New", 0, 12));
        twoUnitsComboBox.setToolTipText(messages.getString("regular_expression_of_analysed_units_pair"));
        twoUnitsComboBox.addFocusListener(new java.awt.event.FocusAdapter() {
            public void focusGained(java.awt.event.FocusEvent evt) {
                twoUnitsComboBoxFocusGained(evt);
            }
        });

        optionsjPanel.add(twoUnitsComboBox);
        twoUnitsComboBox.setBounds(20, 340, 390, 28);

        jLabel6.setText(messages.getString("Units_triple_selector"));
        optionsjPanel.add(jLabel6);
        jLabel6.setBounds(130, 370, 210, 19);

        threeUnitsComboBox.setFont(new java.awt.Font("Courier New", 0, 12));
        threeUnitsComboBox.setToolTipText(messages.getString("regular_expression_of_analysed_units_triplet"));
        threeUnitsComboBox.addFocusListener(new java.awt.event.FocusAdapter() {
            public void focusGained(java.awt.event.FocusEvent evt) {
                threeUnitsComboBoxFocusGained(evt);
            }
        });

        optionsjPanel.add(threeUnitsComboBox);
        threeUnitsComboBox.setBounds(20, 390, 390, 28);

        tfidfjCheckBox.setSelected(true);
        tfidfjCheckBox.setText(messages.getString("tf*idf"));
        tfidfjCheckBox.setToolTipText(messages.getString("relevancy_measure_are_based_on_tf*idf_values_rather_than_number_of_occurrence"));
        optionsjPanel.add(tfidfjCheckBox);
        tfidfjCheckBox.setBounds(40, 420, 70, 20);

        jLabel7.setText(messages.getString("Relevancy_measure"));
        optionsjPanel.add(jLabel7);
        jLabel7.setBounds(130, 420, 210, 19);

        measuresjComboBox.setFont(new java.awt.Font("Courier New", 0, 12));
        measuresjComboBox.setToolTipText(messages.getString("measure_used_for_the_sorting_of_extraction_results"));
        optionsjPanel.add(measuresjComboBox);
        measuresjComboBox.setBounds(20, 440, 390, 28);

        jLabel13.setText(messages.getString("Pruning_threshold"));
        optionsjPanel.add(jLabel13);
        jLabel13.setBounds(10, 470, 130, 19);

        jLabel10.setText(messages.getString("Units_pair"));
        optionsjPanel.add(jLabel10);
        jLabel10.setBounds(10, 490, 100, 19);

        twoUnitsSpinner.setToolTipText(messages.getString("units_pair_that_occur_strictly_less_than_pruning_time"));
        optionsjPanel.add(twoUnitsSpinner);
        twoUnitsSpinner.setBounds(110, 490, 50, 24);

        jLabel11.setText(messages.getString("Units_triple"));
        optionsjPanel.add(jLabel11);
        jLabel11.setBounds(10, 510, 100, 19);

        threeUnitsSpinner.setToolTipText(messages.getString("units_triple_that_occur_stricly_less_than_pruning_time"));
        optionsjPanel.add(threeUnitsSpinner);
        threeUnitsSpinner.setBounds(110, 510, 50, 24);

        jLabel16.setText(messages.getString("Results_number"));
        optionsjPanel.add(jLabel16);
        jLabel16.setBounds(170, 470, 120, 19);

        twoUnitsResultsSpinner.setToolTipText(messages.getString("maximum_results_displayed_for_the_units_pair"));
        optionsjPanel.add(twoUnitsResultsSpinner);
        twoUnitsResultsSpinner.setBounds(180, 490, 70, 24);

        threeUnitsResultsSpinner.setToolTipText(messages.getString("maximum_results_displayed_for_the_units_triple"));
        optionsjPanel.add(threeUnitsResultsSpinner);
        threeUnitsResultsSpinner.setBounds(180, 510, 70, 24);

        jLabel12.setText(messages.getString("Results_format"));
        optionsjPanel.add(jLabel12);
        jLabel12.setBounds(290, 470, 120, 19);

        jLabel8.setText(messages.getString("Nb_integers"));
        optionsjPanel.add(jLabel8);
        jLabel8.setBounds(280, 490, 90, 19);

        nbIntegersSpinner.setToolTipText(messages.getString("number_of_integers_for_results"));
        optionsjPanel.add(nbIntegersSpinner);
        nbIntegersSpinner.setBounds(370, 490, 40, 24);

        jLabel9.setText(messages.getString("Nb_decimals"));
        optionsjPanel.add(jLabel9);
        jLabel9.setBounds(280, 510, 90, 19);

        nbDecimalsSpinner.setToolTipText(messages.getString("number_of_decimals_for_results"));
        optionsjPanel.add(nbDecimalsSpinner);
        nbDecimalsSpinner.setBounds(370, 510, 40, 24);

        occFirstjCheckBox.setSelected(true);
        occFirstjCheckBox.setText(messages.getString("Occurrence_first"));
        occFirstjCheckBox.setToolTipText(messages.getString("Measure_takes_first_into_account_the_number_of_occurrence."));
        optionsjPanel.add(occFirstjCheckBox);
        occFirstjCheckBox.setBounds(270, 420, 150, 20);

        extraction2jSplitPane.setRightComponent(optionsjPanel);

        extractionjSplitPane.setTopComponent(extraction2jSplitPane);

        contextjTextPane.setEditable(false);
        contextjScrollPane.setViewportView(contextjTextPane);

        extractionjSplitPane.setBottomComponent(contextjScrollPane);

        mainjTabbedPane.addTab(messages.getString("Extraction"), extractionjSplitPane);

        rulesjPanel.setLayout(new java.awt.GridLayout(2, 1));

        rulesjPanel.setToolTipText(messages.getString("Interface_of_rules_creation_that_are_applied_after_each_extraction."));
        localRulesjPanel.setLayout(new java.awt.BorderLayout());

        jLabel20.setText(messages.getString("Local_rules"));
        localRulesjPanel.add(jLabel20, java.awt.BorderLayout.NORTH);

        localRulesListsjPanel.setLayout(new java.awt.GridLayout(2, 1));

        localRules2UnitsjScrollPane.setViewportView(localRules2UnitsjList);

        localRulesListsjPanel.add(localRules2UnitsjScrollPane);

        localRules3UnitsjScrollPane.setViewportView(localRules3UnitsjList);

        localRulesListsjPanel.add(localRules3UnitsjScrollPane);

        localRulesjPanel.add(localRulesListsjPanel, java.awt.BorderLayout.CENTER);

        lrTypejComboBox.setToolTipText(messages.getString("Type_of_the_term"));
        localRulesInputjPanel.add(lrTypejComboBox);

        jLabel22.setText(messages.getString("Unit"));
        localRulesInputjPanel.add(jLabel22);

        lrUnitjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrUnitjComboBoxActionPerformed(evt);
            }
        });

        localRulesInputjPanel.add(lrUnitjComboBox);

        localRulesInputjPanel.add(lrPartjComboBox);

        jLabel24.setText(messages.getString("String"));
        localRulesInputjPanel.add(jLabel24);

        lrStringjTextField.setColumns(15);
        localRulesInputjPanel.add(lrStringjTextField);

        lrAdd2UnitsjButton.setMnemonic('2');
        lrAdd2UnitsjButton.setText(messages.getString("Add_2_units_rule"));
        lrAdd2UnitsjButton.setToolTipText(messages.getString("Add_a_2_units_rule"));
        lrAdd2UnitsjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrAdd2UnitsjButtonActionPerformed(evt);
            }
        });

        localRulesInputjPanel.add(lrAdd2UnitsjButton);

        lrAdd3UnitsjButton.setMnemonic('3');
        lrAdd3UnitsjButton.setText(messages.getString("Add_3_units_rule"));
        lrAdd3UnitsjButton.setToolTipText(messages.getString("Add_a_3_units_rule"));
        lrAdd3UnitsjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrAdd3UnitsjButtonActionPerformed(evt);
            }
        });

        localRulesInputjPanel.add(lrAdd3UnitsjButton);

        localRulesjPanel.add(localRulesInputjPanel, java.awt.BorderLayout.SOUTH);

        localRulesActionsjPanel.setLayout(new java.awt.GridLayout(7, 1));

        lrNotjButton.setMnemonic(messages.getString("Not").charAt(0));
        lrNotjButton.setText(messages.getString("Not"));
        lrNotjButton.setToolTipText(messages.getString("Negation_of_the_selected_rules"));
        lrNotjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrNotjButtonActionPerformed(evt);
            }
        });

        localRulesActionsjPanel.add(lrNotjButton);

        lrAndjButton.setMnemonic(messages.getString("And").charAt(0));
        lrAndjButton.setText(messages.getString("And"));
        lrAndjButton.setToolTipText(messages.getString("Conjonction_of_the_selected_pair_of_rules"));
        lrAndjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrAndjButtonActionPerformed(evt);
            }
        });

        localRulesActionsjPanel.add(lrAndjButton);

        lrOrjButton.setMnemonic(messages.getString("Or").charAt(0));
        lrOrjButton.setText(messages.getString("Or"));
        lrOrjButton.setToolTipText(messages.getString("Disjonction_of_the_selected_pair_of_rules"));
        lrOrjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrOrjButtonActionPerformed(evt);
            }
        });

        localRulesActionsjPanel.add(lrOrjButton);

        lrDecomposejButton.setMnemonic(messages.getString("Decompose").charAt(0));
        lrDecomposejButton.setText(messages.getString("Decompose"));
        lrDecomposejButton.setToolTipText(messages.getString("Decompose_the_selected_rules"));
        lrDecomposejButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrDecomposejButtonActionPerformed(evt);
            }
        });

        localRulesActionsjPanel.add(lrDecomposejButton);

        lrUpjButton.setMnemonic(messages.getString("Up").charAt(0));
        lrUpjButton.setText(messages.getString("Up"));
        lrUpjButton.setToolTipText(messages.getString("Shift_the_selected_rules_up"));
        lrUpjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrUpjButtonActionPerformed(evt);
            }
        });

        localRulesActionsjPanel.add(lrUpjButton);

        lrDownjButton.setMnemonic(messages.getString("Down").charAt(0));
        lrDownjButton.setText(messages.getString("Down"));
        lrDownjButton.setToolTipText(messages.getString("Shift_the_selected_rules_down"));
        lrDownjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrDownjButtonActionPerformed(evt);
            }
        });

        localRulesActionsjPanel.add(lrDownjButton);

        lrRemovejButton.setMnemonic(messages.getString("Remove").charAt(0));
        lrRemovejButton.setText(messages.getString("Remove"));
        lrRemovejButton.setToolTipText(messages.getString("Remove_the_selected_rules"));
        lrRemovejButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                lrRemovejButtonActionPerformed(evt);
            }
        });

        localRulesActionsjPanel.add(lrRemovejButton);

        localRulesjPanel.add(localRulesActionsjPanel, java.awt.BorderLayout.EAST);

        rulesjPanel.add(localRulesjPanel);

        jLabel21.setText(messages.getString("Contextual_rules"));
        contextualRulesjPanel.add(jLabel21);

        rulesjPanel.add(contextualRulesjPanel);

        mainjTabbedPane.addTab(messages.getString("Rules"), rulesjPanel);

        termsjSplitPane.setDividerSize(5);
        termsjSplitPane.setOrientation(javax.swing.JSplitPane.VERTICAL_SPLIT);
        termsjSplitPane.setToolTipText(messages.getString("Here_you_manage_the_accepted_and_rejected_terms_of_the_precedent_extractions."));
        termsjSplitPane.setMinimumSize(new java.awt.Dimension(26, 300));
        termsjPanel.setLayout(new java.awt.BorderLayout());

        termsjPanel.setMinimumSize(new java.awt.Dimension(24, 500));
        termsTablesjPanel.setLayout(new java.awt.GridLayout(1, 2));

        acceptedTermsjTable.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {

            },
            new String [] {

            }
        ));
        acceptedTermsjTable.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                acceptedTermsjTableKeyReleased(evt);
            }
        });
        acceptedTermsjTable.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseReleased(java.awt.event.MouseEvent evt) {
                acceptedTermsjTableMouseReleased(evt);
            }
        });

        acceptedTermsjScrollPane.setViewportView(acceptedTermsjTable);

        termsTablesjPanel.add(acceptedTermsjScrollPane);

        rejectedTermsjTable.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {

            },
            new String [] {

            }
        ));
        rejectedTermsjTable.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                rejectedTermsjTableKeyReleased(evt);
            }
        });
        rejectedTermsjTable.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseReleased(java.awt.event.MouseEvent evt) {
                rejectedTermsjTableMouseReleased(evt);
            }
        });

        rejectedTermsjScrollPane.setViewportView(rejectedTermsjTable);

        termsTablesjPanel.add(rejectedTermsjScrollPane);

        termsjPanel.add(termsTablesjPanel, java.awt.BorderLayout.CENTER);

        termsButtonsjPanel.setLayout(new java.awt.GridBagLayout());

        termsButtonsjPanel.setMaximumSize(new java.awt.Dimension(32767, 50));
        termsButtonsjPanel.setPreferredSize(new java.awt.Dimension(337, 36));
        acceptedTermsTypesjComboBox.setToolTipText(messages.getString("Choose_a_type_to_display_in_the_table."));
        acceptedTermsTypesjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptedTermsTypesjComboBoxActionPerformed(evt);
            }
        });

        termsButtonsjPanel.add(acceptedTermsTypesjComboBox, new java.awt.GridBagConstraints());

        acceptedTermsRelationjComboBox.setToolTipText(messages.getString("Choose_a_relation_to_display_in_the_table."));
        acceptedTermsRelationjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptedTermsRelationjComboBoxActionPerformed(evt);
            }
        });

        termsButtonsjPanel.add(acceptedTermsRelationjComboBox, new java.awt.GridBagConstraints());

        acceptedTermsIterationjComboBox.setToolTipText(messages.getString("Choose_an_iteration_to_display_in_the_table."));
        acceptedTermsIterationjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                acceptedTermsIterationjComboBoxActionPerformed(evt);
            }
        });

        termsButtonsjPanel.add(acceptedTermsIterationjComboBox, new java.awt.GridBagConstraints());

        rejectedTermsRelationjComboBox.setToolTipText(messages.getString("Choose_a_relation_to_display_in_the_table."));
        rejectedTermsRelationjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectedTermsRelationjComboBoxActionPerformed(evt);
            }
        });

        termsButtonsjPanel.add(rejectedTermsRelationjComboBox, new java.awt.GridBagConstraints());

        rejectedTermsIterationjComboBox.setToolTipText(messages.getString("Choose_an_iteration_to_display_in_the_table."));
        rejectedTermsIterationjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectedTermsIterationjComboBoxActionPerformed(evt);
            }
        });

        termsButtonsjPanel.add(rejectedTermsIterationjComboBox, new java.awt.GridBagConstraints());

        rejectedTermsTypesjComboBox.setToolTipText(messages.getString("Choose_a_type_to_display_in_the_table."));
        rejectedTermsTypesjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rejectedTermsTypesjComboBoxActionPerformed(evt);
            }
        });

        termsButtonsjPanel.add(rejectedTermsTypesjComboBox, new java.awt.GridBagConstraints());

        termsjPanel.add(termsButtonsjPanel, java.awt.BorderLayout.SOUTH);

        termsjSplitPane.setTopComponent(termsjPanel);

        termsContextjTextPane.setEditable(false);
        termsContextjScrollPane.setViewportView(termsContextjTextPane);

        termsjSplitPane.setBottomComponent(termsContextjScrollPane);

        mainjTabbedPane.addTab(messages.getString("Accepted/Rejected_Terms"), termsjSplitPane);

        undefinedTermsjSplitPane.setDividerSize(5);
        undefinedTermsjSplitPane.setOrientation(javax.swing.JSplitPane.VERTICAL_SPLIT);
        undefinedTermsjSplitPane.setToolTipText(messages.getString("Here_you_manage_the_undefined_terms_of_the_precedent_extractions."));
        undefinedTermsjPanel.setLayout(new java.awt.BorderLayout());

        undefinedTermsjTable.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {

            },
            new String [] {

            }
        ));
        undefinedTermsjTable.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                undefinedTermsjTableKeyReleased(evt);
            }
        });
        undefinedTermsjTable.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseReleased(java.awt.event.MouseEvent evt) {
                undefinedTermsjTableMouseReleased(evt);
            }
        });

        undefinedTermsjScrollPane.setViewportView(undefinedTermsjTable);

        undefinedTermsjPanel.add(undefinedTermsjScrollPane, java.awt.BorderLayout.CENTER);

        undefinedTermsRelationjComboBox.setToolTipText(messages.getString("Choose_a_relation_to_display_in_the_table."));
        undefinedTermsRelationjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                undefinedTermsRelationjComboBoxActionPerformed(evt);
            }
        });

        undefinedTermsButtonsjPanel.add(undefinedTermsRelationjComboBox);

        undefinedTermsIterationjComboBox.setToolTipText(messages.getString("Choose_an_iteration_to_display_in_the_table."));
        undefinedTermsIterationjComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                undefinedTermsIterationjComboBoxActionPerformed(evt);
            }
        });

        undefinedTermsButtonsjPanel.add(undefinedTermsIterationjComboBox);

        precedentTermsUndefinedTermsjButton.setMnemonic(messages.getString("Precedent_terms").charAt(0));
        precedentTermsUndefinedTermsjButton.setText(messages.getString("Precedent_terms"));
        precedentTermsUndefinedTermsjButton.setToolTipText(messages.getString("Use_the_precedent_accepted_and_rejected_terms_to_accept_and_reject_new_terms."));
        precedentTermsUndefinedTermsjButton.setPreferredSize(new java.awt.Dimension(150, 26));
        precedentTermsUndefinedTermsjButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                precedentTermsUndefinedTermsjButtonActionPerformed(evt);
            }
        });

        undefinedTermsButtonsjPanel.add(precedentTermsUndefinedTermsjButton);

        undefinedTermsjPanel.add(undefinedTermsButtonsjPanel, java.awt.BorderLayout.SOUTH);

        undefinedTermsjSplitPane.setTopComponent(undefinedTermsjPanel);

        undefinedTermsContextjScrollPane.setViewportView(undefinedTermsContextjTextPane);

        undefinedTermsjSplitPane.setBottomComponent(undefinedTermsContextjScrollPane);

        mainjTabbedPane.addTab(messages.getString("Undefined_terms"), undefinedTermsjSplitPane);

        getContentPane().add(mainjTabbedPane, java.awt.BorderLayout.CENTER);

        statejPanel.setLayout(new java.awt.BorderLayout());

        statejPanel.setMaximumSize(new java.awt.Dimension(63, 26));
        statejLabel.setText(messages.getString("Status_line"));
        statejLabel.setToolTipText(messages.getString("display_informations_on_current_operation"));
        statejLabel.setDoubleBuffered(true);
        statejLabel.setPreferredSize(new java.awt.Dimension(600, 16));
        statejLabel.setOpaque(true);
        statejPanel.add(statejLabel, java.awt.BorderLayout.WEST);

        extractionjProgressBar.setDoubleBuffered(true);
        extractionjProgressBar.setPreferredSize(new java.awt.Dimension(200, 14));
        statejPanel.add(extractionjProgressBar, java.awt.BorderLayout.CENTER);

        getContentPane().add(statejPanel, java.awt.BorderLayout.SOUTH);

        programMenu.setMnemonic(messages.getString("Program").charAt(0));
        programMenu.setText(messages.getString("Program"));
        selectNewInputFilejMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_N, java.awt.event.InputEvent.CTRL_MASK));
        selectNewInputFilejMenuItem.setMnemonic(messages.getString("Select_new_input_file...").charAt(0));
        selectNewInputFilejMenuItem.setText(messages.getString("Select_new_input_file..."));
        selectNewInputFilejMenuItem.setToolTipText(messages.getString("Create_a_new_set_of_input_and_options_files."));
        selectNewInputFilejMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                selectNewInputFilejMenuItemActionPerformed(evt);
            }
        });

        programMenu.add(selectNewInputFilejMenuItem);

        saveResultsjMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_S, java.awt.event.InputEvent.CTRL_MASK));
        saveResultsjMenuItem.setMnemonic(messages.getString("Save_all_results").charAt(0));
        saveResultsjMenuItem.setText(messages.getString("Save_all_results"));
        saveResultsjMenuItem.setToolTipText(messages.getString("save_all_results_in_default_files"));
        saveResultsjMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                saveResultsjMenuItemActionPerformed(evt);
            }
        });

        programMenu.add(saveResultsjMenuItem);

        languagejMenu.setMnemonic(messages.getString("Language").charAt(0));
        languagejMenu.setText(messages.getString("Language"));
        languagejMenu.setToolTipText(messages.getString("change_the_language_of_the_interface"));
        automaticjRadioButtonMenuItem.setMnemonic(messages.getString("Automatic").charAt(0));
        automaticjRadioButtonMenuItem.setSelected(true);
        automaticjRadioButtonMenuItem.setText(messages.getString("Automatic"));
        automaticjRadioButtonMenuItem.setToolTipText(messages.getString("Choose_the_language_according_to_the_language_environment_variable."));
        automaticjRadioButtonMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                automaticjRadioButtonMenuItemActionPerformed(evt);
            }
        });

        languagejMenu.add(automaticjRadioButtonMenuItem);

        englishjRadioButtonMenuItem.setMnemonic(messages.getString("English").charAt(0));
        englishjRadioButtonMenuItem.setText(messages.getString("English"));
        englishjRadioButtonMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                englishjRadioButtonMenuItemActionPerformed(evt);
            }
        });

        languagejMenu.add(englishjRadioButtonMenuItem);

        francaisjRadioButtonMenuItem.setMnemonic(messages.getString("Francais").charAt(0));
        francaisjRadioButtonMenuItem.setText(messages.getString("Francais"));
        francaisjRadioButtonMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                francaisjRadioButtonMenuItemActionPerformed(evt);
            }
        });

        languagejMenu.add(francaisjRadioButtonMenuItem);

        programMenu.add(languagejMenu);

        exitMenuItem.setMnemonic(messages.getString("Exit_program").charAt(0));
        exitMenuItem.setText(messages.getString("Exit_program"));
        exitMenuItem.setToolTipText(messages.getString("exit_from_the_program"));
        exitMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                exitMenuItemActionPerformed(evt);
            }
        });

        programMenu.add(exitMenuItem);

        menuBar.add(programMenu);

        displayjMenu.setMnemonic(messages.getString("Display").charAt(0));
        displayjMenu.setText(messages.getString("Display"));
        displayMeasuresjCheckBoxMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_F3, 0));
        displayMeasuresjCheckBoxMenuItem.setText(messages.getString("Measures"));
        displayMeasuresjCheckBoxMenuItem.setToolTipText(messages.getString("Toggle_display_of_measures_in_the_lists"));
        displayMeasuresjCheckBoxMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                displayMeasuresjCheckBoxMenuItemActionPerformed(evt);
            }
        });

        displayjMenu.add(displayMeasuresjCheckBoxMenuItem);

        displayTagsjCheckBoxMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_F4, 0));
        displayTagsjCheckBoxMenuItem.setText(messages.getString("Tags"));
        displayTagsjCheckBoxMenuItem.setToolTipText(messages.getString("Toggle_display_of_tags_in_the_lists"));
        displayTagsjCheckBoxMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                displayTagsjCheckBoxMenuItemActionPerformed(evt);
            }
        });

        displayjMenu.add(displayTagsjCheckBoxMenuItem);

        displayOccjCheckBoxMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_F5, 0));
        displayOccjCheckBoxMenuItem.setText(messages.getString("Numbers"));
        displayOccjCheckBoxMenuItem.setToolTipText(messages.getString("Toggle_display_of_numbers_in_the_lists"));
        displayOccjCheckBoxMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                displayOccjCheckBoxMenuItemActionPerformed(evt);
            }
        });

        displayjMenu.add(displayOccjCheckBoxMenuItem);

        displayExpertModejCheckBoxMenuItem.setText(messages.getString("Expert_Mode"));
        displayExpertModejCheckBoxMenuItem.setToolTipText(messages.getString("Toggle_display_of_expert_options"));
        displayExpertModejCheckBoxMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                displayExpertModejCheckBoxMenuItemActionPerformed(evt);
            }
        });

        displayjMenu.add(displayExpertModejCheckBoxMenuItem);

        menuBar.add(displayjMenu);

        helpMenu.setMnemonic(messages.getString("Help").charAt(0));
        helpMenu.setText(messages.getString("Help"));
        contexualHelpjMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_F1, 0));
        contexualHelpjMenuItem.setMnemonic(messages.getString("Contextual_help").charAt(0));
        contexualHelpjMenuItem.setText(messages.getString("Contextual_help"));
        contexualHelpjMenuItem.setToolTipText(messages.getString("display_help_corresponding_to_focused_element_of_interface"));
        contexualHelpjMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                contexualHelpjMenuItemActionPerformed(evt);
            }
        });

        helpMenu.add(contexualHelpjMenuItem);

        contentsMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_F1, java.awt.event.InputEvent.CTRL_MASK));
        contentsMenuItem.setMnemonic(messages.getString("Contents").charAt(0));
        contentsMenuItem.setText(messages.getString("Contents"));
        contentsMenuItem.setToolTipText(messages.getString("display_contents_of_help"));
        contentsMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                contentsMenuItemActionPerformed(evt);
            }
        });

        helpMenu.add(contentsMenuItem);

        firstStartjMenuItem.setText(messages.getString("First_start"));
        firstStartjMenuItem.setToolTipText(messages.getString("First_start_guide"));
        firstStartjMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                firstStartjMenuItemActionPerformed(evt);
            }
        });

        helpMenu.add(firstStartjMenuItem);

        helpUnderCursorjCheckBoxMenuItem.setMnemonic(messages.getString("Help_under_cursor").charAt(0));
        helpUnderCursorjCheckBoxMenuItem.setSelected(true);
        helpUnderCursorjCheckBoxMenuItem.setText(messages.getString("Help_under_cursor"));
        helpUnderCursorjCheckBoxMenuItem.setToolTipText(messages.getString("toggle_activation_of_help_under_cursor"));
        helpUnderCursorjCheckBoxMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                helpUnderCursorjCheckBoxMenuItemActionPerformed(evt);
            }
        });

        helpMenu.add(helpUnderCursorjCheckBoxMenuItem);

        aboutMenuItem.setMnemonic(messages.getString("About").charAt(0));
        aboutMenuItem.setText(messages.getString("About"));
        aboutMenuItem.setToolTipText(messages.getString("about_this_program"));
        aboutMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                aboutMenuItemActionPerformed(evt);
            }
        });

        helpMenu.add(aboutMenuItem);

        menuBar.add(helpMenu);

        setJMenuBar(menuBar);

        pack();
    }//GEN-END:initComponents

    /**
     * Just repaints the lists with the new display options.
     */
    private void displayOccjCheckBoxMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_displayOccjCheckBoxMenuItemActionPerformed

        resultsjTable.repaint();
        acceptedTermsjTable.repaint();
        rejectedTermsjTable.repaint();
        undefinedTermsjTable.repaint();
    }//GEN-LAST:event_displayOccjCheckBoxMenuItemActionPerformed

    /**
     * Show/hide some advanced elements of the interface.
     */
    private void displayExpertModejCheckBoxMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_displayExpertModejCheckBoxMenuItemActionPerformed

        if (displayExpertModejCheckBoxMenuItem.isSelected()) {
            precedentTermsjButton.setVisible(true);
            jLabel2.setVisible(true);
            outputFileComboBox.setVisible(true);
            outputFileChooserjButton.setVisible(true);
            savejButton.setVisible(true);
            jLabel3.setVisible(true);
            unitComboBox.setVisible(true);
            jLabel4.setVisible(true);
            sentPhrComboBox.setVisible(true);
            jLabel14.setVisible(true);
            nameOfTextjComboBox.setVisible(true);
            tfidfjCheckBox.setVisible(true);
            occFirstjCheckBox.setVisible(true);
            jLabel12.setVisible(true);
            jLabel8.setVisible(true);
            nbIntegersSpinner.setVisible(true);
            jLabel9.setVisible(true);
            nbDecimalsSpinner.setVisible(true);

            precedentTermsUndefinedTermsjButton.setVisible(true);

            extractButton.setBackground(new Color(204, 204, 204));
            optionsComboBox.setBackground(new Color(204, 204, 204));
            mainjTabbedPane.setTitleAt(0,
               messages.getString("Expressions"));
            if (iteration == 0) {
                mainjTabbedPane.setEnabledAt(1, true);
            }
            mainjTabbedPane.setTitleAt(1,
               messages.getString("Preprocessing"));
            mainjTabbedPane.setTitleAt(2,
               messages.getString("Extraction"));
            mainjTabbedPane.setEnabledAt(3, true);
            mainjTabbedPane.setTitleAt(3,
               messages.getString("Rules"));
            mainjTabbedPane.setTitleAt(4,
               messages.getString("Accepted/Rejected_Terms"));
            mainjTabbedPane.setTitleAt(5,
               messages.getString("Undefined_terms"));

        } else {
            precedentTermsjButton.setVisible(false);
            jLabel2.setVisible(false);
            outputFileComboBox.setVisible(false);
            outputFileChooserjButton.setVisible(false);
            savejButton.setVisible(false);
            jLabel3.setVisible(false);
            unitComboBox.setVisible(false);
            jLabel4.setVisible(false);
            sentPhrComboBox.setVisible(false);
            jLabel14.setVisible(false);
            nameOfTextjComboBox.setVisible(false);
            tfidfjCheckBox.setVisible(false);
            occFirstjCheckBox.setVisible(false);
            jLabel12.setVisible(false);
            jLabel8.setVisible(false);
            nbIntegersSpinner.setVisible(false);
            jLabel9.setVisible(false);
            nbDecimalsSpinner.setVisible(false);

            precedentTermsUndefinedTermsjButton.setVisible(false);

            extractButton.setBackground(new Color(202, 0, 0));
            optionsComboBox.setBackground(new Color(202, 0, 0));
            mainjTabbedPane.setTitleAt(0,
               "1 "+messages.getString("Expressions"));
            mainjTabbedPane.setEnabledAt(1, false);
            mainjTabbedPane.setTitleAt(2,
               "2 "+messages.getString("Extraction"));
            mainjTabbedPane.setEnabledAt(3, false);
            mainjTabbedPane.setTitleAt(4,
               "3 "+messages.getString("Accepted/Rejected_Terms"));
            mainjTabbedPane.setTitleAt(5,
               "4 "+messages.getString("Undefined_terms"));
        }
    }//GEN-LAST:event_displayExpertModejCheckBoxMenuItemActionPerformed

    /**
     * Just repaints the lists with the new display options.
     */
    private void displayTagsjCheckBoxMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_displayTagsjCheckBoxMenuItemActionPerformed

        resultsjTable.repaint();
        acceptedTermsjTable.repaint();
        rejectedTermsjTable.repaint();
        undefinedTermsjTable.repaint();
    }//GEN-LAST:event_displayTagsjCheckBoxMenuItemActionPerformed

    /**
     * Just repaints the lists with the new display options.
     */
    private void displayMeasuresjCheckBoxMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_displayMeasuresjCheckBoxMenuItemActionPerformed

        resultsjTable.repaint();
        acceptedTermsjTable.repaint();
        rejectedTermsjTable.repaint();
        undefinedTermsjTable.repaint();
    }//GEN-LAST:event_displayMeasuresjCheckBoxMenuItemActionPerformed

    /**
     * Display the first start guide.
     */
    private void firstStartjMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_firstStartjMenuItemActionPerformed

        JFrame jf = HTMLWindow("file:"
                   +System.getProperty("user.dir")
                   +System.getProperty("file.separator")
                   +"doc"
                   +System.getProperty("file.separator")
                   +messages.getString("guide.html"),
                   messages.getString("First_time_you_run_this_program."));
    }//GEN-LAST:event_firstStartjMenuItemActionPerformed


    /*-************************
     * Precedent Terms Dialog *
     **************************/

    private void precedentTermsCanceljButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_precedentTermsCanceljButtonActionPerformed
        precedentTermsjDialog.hide();
    }//GEN-LAST:event_precedentTermsCanceljButtonActionPerformed

    private void precedentTermsRejectjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_precedentTermsRejectjButtonActionPerformed
        JTable ut;
        int threshold, status;
        Occurrence occ;
        Component selectedTab;

        selectedTab = mainjTabbedPane.getSelectedComponent();
        if (selectedTab.equals(extractionjSplitPane)) {
            ut = resultsjTable;
        } else if (selectedTab.equals(undefinedTermsjSplitPane)) {
            ut = undefinedTermsjTable;
        } else {
            return;
        }
        status = REJECTED;
        threshold = ((Integer)precedentTermsjSpinner.getValue()).intValue();

        ut.clearSelection();
        for (int row = 0; row < ut.getRowCount(); row++) {
            // this occurrence has already a status
            if ((selectedTab.equals(extractionjSplitPane)
                && rowsStatus[row] != UNDEFINED)
            || !(ut.getModel().getValueAt(row, 0) instanceof Occurrence)) {
                continue;
            }

            occ = (Occurrence)ut.getModel().getValueAt(row, 0);

            if (precedentTermsParameter(occ, threshold) == status) {
                ut.addRowSelectionInterval(row, row);
            }
        }
        precedentTermsjDialog.hide();
    }//GEN-LAST:event_precedentTermsRejectjButtonActionPerformed

    private void precedentTermsAcceptjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_precedentTermsAcceptjButtonActionPerformed
        JTable ut;
        Occurrence occ;
        int threshold, status;
        Component selectedTab;

        selectedTab = mainjTabbedPane.getSelectedComponent();
        if (selectedTab.equals(extractionjSplitPane)) {
            ut = resultsjTable;
        } else if (selectedTab.equals(undefinedTermsjSplitPane)) {
            ut = undefinedTermsjTable;
        } else {
            return;
        }
        status = ACCEPTED;
        threshold = ((Integer)precedentTermsjSpinner.getValue()).intValue();

        ut.clearSelection();
        for (int row = 0; row < ut.getRowCount(); row++) {
            // this occurrence has already a status
            if ((selectedTab.equals(extractionjSplitPane)
                && rowsStatus[row] != UNDEFINED)
            || !(ut.getModel().getValueAt(row, 0) instanceof Occurrence)) {
                continue;
            }

            occ = (Occurrence)ut.getModel().getValueAt(row, 0);

            if (precedentTermsParameter(occ, threshold) == status) {
                ut.addRowSelectionInterval(row, row);
            }
        }
        precedentTermsjDialog.hide();
    }//GEN-LAST:event_precedentTermsAcceptjButtonActionPerformed


    /*-*********************
     * Preprocessing panel *
     ***********************/

    /**
     * Search Occurrences contained in the input file and add to them the
     * lines where they appear.
     *
     * @param terms list of Occurences without lines
     * @return false only and only if there is an error
     */
    private boolean addLinesToOccurrence(LinkedHashSet terms) {


        String lineContent;
        String unit, previousUnit, previous2Unit;
        int line;
        BufferedReader br;
        Matcher matcherUnit;
        LinkedList termsList;
        Occurrence occ;
        int index;

        // open the input file in reading mode
        try {
            br = new BufferedReader(new FileReader(et.getInputFileName()));

        } catch (FileNotFoundException fnfe) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("File_not_found")+"\n"
                +fnfe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return false;
        }

        // list nedeed for the get(int) method
        termsList = new LinkedList(terms);
        unit = null;
        previousUnit = null;
        previous2Unit = null;
        line = 1;

        try {

        lineContent = br.readLine();

        // read the input file line by line only one time
        while (lineContent != null) {
            matcherUnit = et.getPatternUnit().matcher(lineContent);
            previous2Unit = null;
            previousUnit = null;

            // split each line into units
            while(matcherUnit.find()) {
                unit = matcherUnit.group();

                // if a sentence/phrase separator is found skip it
                if (et.getPatternSentPhraSeparator()
                    .matcher(unit).matches()){
                    previousUnit = null;
                    continue;
                }

                // wait to find the second unit before filling
                // occurrences for units pair
                if (previousUnit == null) {
                    previousUnit = unit;
                    continue;
                }

                index = termsList.indexOf(
                        new Occurrence(previousUnit, unit));
                if (index != -1) {
                    occ = (Occurrence)termsList.get(index);
                    // save line of input file for the units pair
                    occ.addLine(line);
                }

                // wait to find the third unit before filling
                // occurrences for units triple
                if (previous2Unit == null) {
                    previous2Unit = previousUnit;
                    previousUnit = unit;
                    continue;
                }

                index = termsList.indexOf(
                        new Occurrence(previous2Unit, previousUnit, unit));
                if (index != -1) {
                    occ = (Occurrence)termsList.get(index);
                    // save line of input file for the units triple
                    occ.addLine(line);
                }

                previous2Unit = previousUnit;
                previousUnit = unit;
            }
            lineContent = br.readLine();
            line++;
        }

        // close the input file
        br.close();

        } catch(IOException ioe){
            JOptionPane.showMessageDialog(frame,
                messages.getString("Error_in_reading_file")+"\n"
                +ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return false;
        }

        return true;
    }

    /**
     * Edit tags of selected terms.
     */
    private void editTagsAlreadyFormedTermsjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_editTagsAlreadyFormedTermsjButtonActionPerformed

        int listIndex;
        DefaultListModel lm;
        Occurrence occ, occ2;
        Iterator it;
        String tag, tag2, tag3;
        String word, word2, word3;
        // contain tag (value) of each word (key)
        Hashtable words;

        lm = (DefaultListModel)alreadyFormedTermsjList.getModel();
        words = new Hashtable();

        for (int i = 0 ;
             i < alreadyFormedTermsjList.getSelectedIndices().length;
             i++) {

            listIndex = (alreadyFormedTermsjList.getSelectedIndices())[i];
            occ = (Occurrence)lm.get(listIndex);

            // ask the first tag
            alreadyFormedTermsjList.ensureIndexIsVisible(listIndex);
            tag = null;
            word = Occurrence.getWord(occ.getUnit1(), et.getSeparator());
            do {
                tag = (String)JOptionPane.showInputDialog(frame,
                    messages.getString(
                        "Please_give_the_tag_associated_to_the_word:")
                    +" "+word,
                    "",
                    JOptionPane.QUESTION_MESSAGE,
                    null, null,
                    Occurrence.getTag(occ.getUnit1(), et.getSeparator()));
            } while (tag != null && tag.trim().length() == 0);
            if (tag == null) { return; }
            words.put(word, tag);

            // ask the second tag
            alreadyFormedTermsjList.ensureIndexIsVisible(listIndex);
            tag2 = null;
            word2 = Occurrence.getWord(occ.getUnit2(), et.getSeparator());
            do {
                tag2 = (String)JOptionPane.showInputDialog(frame,
                    messages.getString(
                        "Please_give_the_tag_associated_to_the_word:")
                    +" "+word2,
                    "",
                    JOptionPane.QUESTION_MESSAGE,
                    null, null,
                    Occurrence.getTag(occ.getUnit2(), et.getSeparator()));
            } while (tag2 != null && tag2.trim().length() == 0);
            if (tag2 == null) { return; }
            words.put(word2, tag2);

            // ask the third tag
            tag3 = null;
            word3 = null;
            if (occ.is3Units()) {
                word3 =
                    Occurrence.getWord(occ.getUnit3(), et.getSeparator());
                alreadyFormedTermsjList.ensureIndexIsVisible(listIndex);
                do {
                    tag3 = (String)JOptionPane.showInputDialog(frame,
                        messages.getString(
                            "Please_give_the_tag_associated_to_the_word:")
                        +" "+word3,
                        "",
                        JOptionPane.QUESTION_MESSAGE,
                        null, null,
                        Occurrence.getTag(occ.getUnit3(),
                                          et.getSeparator()));
                } while (tag3 != null && tag3.trim().length() == 0);
                if (tag3 == null) { return; }
                words.put(word3, tag3);
            }

            if (occ.is2Units()) {
                occ2 = new Occurrence(
                    word+et.getSeparator()+tag+" ",
                    word2+et.getSeparator()+tag2+" ");
                it = occ.getLinesIterator();
                while (it.hasNext()) {
                    occ2.addLine(((Integer)it.next()).intValue());
                }
                // put the new Occurrence with tags and lines
                lm.setElementAt(occ2, listIndex);

            } else {
                occ2 = new Occurrence(
                    word+et.getSeparator()+tag+" ",
                    word2+et.getSeparator()+tag2+" ",
                    word3+et.getSeparator()+tag3+" ");
                it = occ.getLinesIterator();
                while (it.hasNext()) {
                    occ2.addLine(((Integer)it.next()).intValue());
                }
                // put the new Occurrence with tags and lines
                lm.setElementAt(occ2, listIndex);
            }
        }
    }//GEN-LAST:event_editTagsAlreadyFormedTermsjButtonActionPerformed

    private void helpTagsAlreadyFormedTermsjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_helpTagsAlreadyFormedTermsjButtonActionPerformed

        regexHelpTagsjButtonActionPerformed(null);
    }//GEN-LAST:event_helpTagsAlreadyFormedTermsjButtonActionPerformed

    /**
     * Add terms contain in a file to memory tables. Must be executed
     * before the first extraction and after newExtractTerms().
     *
     * @param fileComboBox ComboBox which selected item is a String of the
     * file name that contain the list of terms, one term by line and each
     * unit of term corresponding to the unit definition.
     * @param mode REJECTED or ACCEPTED, memory tables where the terms are
     * added
     * @return true if and only if no errors are found
     */
    private boolean addTermFileToMemoryTables(JComboBox fileComboBox,
                                              int mode) {

        Occurrence occ;
        BufferedReader br;
        String lineContent;
        Matcher matcher, typeMatcher;
        Pattern typePattern;
        String unit1, unit2, unit3;


        try {

        br = new BufferedReader(new FileReader(
                     (String)fileComboBox.getSelectedItem()));

        lineContent = br.readLine();
        typePattern = Pattern.compile("\\[([^\\]]+)\\]");

        while (lineContent != null) {

            // split the line in units
            unit1 = null; unit2 = null; unit3 = null;
            matcher = et.getPatternUnit().matcher(lineContent);
            typeMatcher = typePattern.matcher(lineContent);

            if (matcher.find()) { unit1 = matcher.group(); }
            else { lineContent = br.readLine(); continue; }

            if (matcher.find()) { unit2 = matcher.group(); }
            else { lineContent = br.readLine(); continue; }

            if (matcher.find()) { unit3 = matcher.group(); }

            occ = null;

            if (unit3 != null) {
                occ = new Occurrence(unit1, unit2, unit3);
                if (typeMatcher.find()) {
                    occ.setTypeName(typeMatcher.group(1));
                }

            } else {
                occ = new Occurrence(unit1, unit2);
                if (typeMatcher.find()) {
                    occ.setTypeName(typeMatcher.group(1));
                }
            }

            switch (mode) {

            case ACCEPTED:
                acceptedTerms.add(occ);
                currentAcceptedTerms.add(occ);
                currentAcceptedTermsTreeSet.add(occ.getKey());
                break;

            case REJECTED:
                rejectedTerms.add(occ);
                break;
            }
            currentSkippedTermsTreeSet.add(occ.getKey());

            lineContent = br.readLine();
        }

        } catch (IOException ioe) {
            mainjTabbedPane.setSelectedComponent(preProcessingjPanel);
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage()
                +" ("+(String)fileComboBox.getSelectedItem()+")",
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            fileComboBox.requestFocusInWindow();
            return false;
        }

        if (acceptedTerms.size() > 0
            || rejectedTerms.size() > 0) {
            return true;

        } else {
            mainjTabbedPane.setSelectedComponent(preProcessingjPanel);
            JOptionPane.showMessageDialog(frame,
                messages.getString("No_data_found_in_this_file.")
                +" ("+(String)fileComboBox.getSelectedItem()+")",
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            fileComboBox.requestFocusInWindow();
            return false;
        }
    }

    /**
     * Display a file chooser for selecting the list of terms to reject.
     */
    private void rejectListjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_rejectListjButtonActionPerformed

        JFileChooser fileChooser;
        String file;
        File currentDir;

        currentDir = new File(System.getProperty("user.dir"));
        fileChooser = new JFileChooser(currentDir);
        fileChooser.setDialogTitle(
            messages.getString("list_of_terms_to_reject"));
        fileChooser.addChoosableFileFilter(
            new FileChooserFilterExtension("terms",
                messages.getString("Terms_files")));
        int result = fileChooser.showOpenDialog(frame);

        try {
        if (result == JFileChooser.APPROVE_OPTION) {
            file = fileChooser.getSelectedFile().getCanonicalPath();
            if (file.startsWith(currentDir.getCanonicalPath())) {
                file = file.substring(currentDir.toString().length()+1);
            }
            rejectListjComboBox.setSelectedItem(file);
        }
        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            rejectListjComboBox.requestFocusInWindow();
            return;
        }

    }//GEN-LAST:event_rejectListjButtonActionPerformed

    /**
     * Display a file chooser for selecting the list of terms to accept.
     */
    private void acceptListjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_acceptListjButtonActionPerformed

        JFileChooser fileChooser;
        String file;
        File currentDir;

        currentDir = new File(System.getProperty("user.dir"));
        fileChooser = new JFileChooser(currentDir);
        fileChooser.setDialogTitle(
            messages.getString("list_of_terms_to_accept"));
        fileChooser.addChoosableFileFilter(
            new FileChooserFilterExtension("terms",
                messages.getString("Terms_files")));
        int result = fileChooser.showOpenDialog(frame);

        try {
            if (result == JFileChooser.APPROVE_OPTION) {
                file = fileChooser.getSelectedFile().getCanonicalPath();
                if (file.startsWith(currentDir.getCanonicalPath())) {
                    file = file.substring(currentDir.toString().length()+1);
                }
                acceptListjComboBox.setSelectedItem(file);
            }
        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            acceptListjComboBox.requestFocusInWindow();
            return;
        }

    }//GEN-LAST:event_acceptListjButtonActionPerformed

    /**
     * Remove the selected terms from the already formed terms list.
     */
    private void removeAlreadyFormedTermsjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_removeAlreadyFormedTermsjButtonActionPerformed

        int index, fromIndex, toIndex;
        int [] si = alreadyFormedTermsjList.getSelectedIndices();
        DefaultListModel lm = (DefaultListModel)
            alreadyFormedTermsjList.getModel();

        if (si.length == 0) { return; }

        toIndex = (si)[si.length-1];
        fromIndex = toIndex;
        for (int i = si.length-1; i >= 0; i--) {
            index = (si)[i];
            if (index < fromIndex - 1) {
                // if this is a new group of continue selection
                lm.removeRange(fromIndex, toIndex);
                toIndex = index;
            }
            fromIndex = index;
        }
        // last range to remove
        if (toIndex < lm.getSize() && fromIndex <= toIndex) {
            lm.removeRange(fromIndex, toIndex);
        }
    }//GEN-LAST:event_removeAlreadyFormedTermsjButtonActionPerformed

    /**
     * Find already formed terms and add them to the associated list with
     * their different possible combinaisons of units. Retrieve the tag of
     * each word and if it doesn't exist in text ask the user for the tag.
     * Add lines to Occurrence where terms and separated words occur.
     */
    private void searchAlreadyFormedTermsjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_searchAlreadyFormedTermsjButtonActionPerformed

        String lineContent, unit;
        String word, tag, tag2, tag3;
        BufferedReader br;
        int line, index;
        Occurrence occ, occ2;
        Matcher matcherUnit;
        String [] subUnits;
        DefaultListModel lm;
        // contain tag (value) of each word (key) in text
        Hashtable words;
        boolean badWord;
        Iterator it;
        Object [] occurrences;

        try {

        // get and test all parameters of the interface
        et = newExtractTerms();

        // if there is an error in the provided extraction parameters
        if (et == null) { return; }

        // open input file in reading
        br = new BufferedReader(new FileReader(et.getInputFileName()));

        words = new Hashtable();
        line = 1;
        lineContent = br.readLine();
        lm = (DefaultListModel)alreadyFormedTermsjList.getModel();
        lm.clear();
        while (lineContent != null) { // read line by line
            matcherUnit = et.getPatternUnit().matcher(lineContent);

                while(matcherUnit.find()) { // read unit by unit
                    unit = matcherUnit.group();

                    // if a sentence/phrase separator is found skip it
                    if (et.getPatternSentPhraSeparator()
                        .matcher(unit).matches()) {
                        continue;
                    }

                    // get the word and the tag of the unit
                    index = unit.indexOf(et.getSeparator());
                    word = unit.substring(0, index);
                    tag = unit.substring(index+1).trim();

                    // split the word around '-' separator
                    subUnits = word.split("-");

                    // test that separated words are not empty
                    badWord = false;
                    for (int i = 0; i < subUnits.length; i++) {
                        if (subUnits[i].length() == 0) {
                            badWord = true;
                            break;
                        }
                    }
                    if (badWord) continue;

                    switch (subUnits.length) {
                        case 1:
                            // save each word/tag in a hash table
                            if (!words.containsKey(word)) {
                                words.put(word, tag);
                            }
                            break;
                        case 2:
                            occ2 = new Occurrence(subUnits[0], subUnits[1]);
                            index = lm.indexOf(occ2);
                            if (index == -1) {
                                // if this term is not already in the list
                                occ = occ2;
                            } else {
                                occ = (Occurrence)lm.remove(index);
                            }
                            occ.addLine(line);
                            lm.addElement(occ);
                            break;
                        case 3:
                            occ2 = new Occurrence(subUnits[0], subUnits[1],
                                                  subUnits[2]);
                            index = lm.indexOf(occ2);
                            occ = (index == -1) ? occ2
                                : (Occurrence)lm.remove(index);
                            occ.addLine(line);
                            lm.addElement(occ);

                            // add all possible combinaisons
                            occ2 = new Occurrence(subUnits[0], subUnits[1]);
                            index = lm.indexOf(occ2);
                            occ = (index == -1) ? occ2
                                : (Occurrence)lm.remove(index);
                            occ.addLine(line);
                            lm.addElement(occ);

                            occ2 = new Occurrence(subUnits[1], subUnits[2]);
                            index = lm.indexOf(occ2);
                            occ = (index == -1) ? occ2
                                : (Occurrence)lm.remove(index);
                            occ.addLine(line);
                            lm.addElement(occ);
                            break;
                    }
                }
            lineContent = br.readLine();
            line++;
        }

        // close the input file
        br.close();

        // sort the list
        occurrences = lm.toArray();
        Arrays.sort(occurrences);
        lm.clear();
        for (int i = 0; i < occurrences.length; i++) {
            lm.addElement(occurrences[i]);
        }

        // search words without tag in the table and get the associated tag
        // add the tag to the word in the Occurrence from words table
        // else ask user for tag
        for (int listIndex = 0; listIndex < lm.size(); listIndex++) {
            occ = (Occurrence)lm.getElementAt(listIndex);
            tag = (String)words.get(occ.getUnit1());
            tag2 = (String)words.get(occ.getUnit2());
            tag3 = null;
            if (occ.is3Units()) {
                tag3 = (String)words.get(occ.getUnit3());
            }

            if (tag == null) { tag = ""; }
            if (tag2 == null) { tag2 = ""; }
            if (occ.is3Units() && tag3 == null) { tag3 = ""; }

            if (occ.is2Units()) {
                occ2 = new Occurrence(
                    occ.getUnit1()+et.getSeparator()+tag+" ",
                    occ.getUnit2()+et.getSeparator()+tag2+" ");
                it = occ.getLinesIterator();
                while (it.hasNext()) {
                    occ2.addLine(((Integer)it.next()).intValue());
                }
                // put the new Occurrence with tags and lines
                lm.setElementAt(occ2, listIndex);

            } else {
                occ2 = new Occurrence(
                    occ.getUnit1()+et.getSeparator()+tag+" ",
                    occ.getUnit2()+et.getSeparator()+tag2+" ",
                    occ.getUnit3()+et.getSeparator()+tag3+" ");
                it = occ.getLinesIterator();
                while (it.hasNext()) {
                    occ2.addLine(((Integer)it.next()).intValue());
                }
                // put the new Occurrence with tags and lines
                lm.setElementAt(occ2, listIndex);
            }
        }

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }
    }//GEN-LAST:event_searchAlreadyFormedTermsjButtonActionPerformed


    /*-*************
     * Rules panel *
     ***************/

    /**
     * Read the list of extraction results and for each Occurrence object
     * apply local rules according to the number of units of Occurrence.
     * Some Occurrence will be marked as accepted and other as rejected.
     * Not implemented : Then contextual rules are apply on the remaining
     * undefined Occurrences in the same way as precedent local rules.
     */
    private void applyRulesToExtractionResults() {

        Object obj;
        LocalRule lr;

        // for each cell of the table of results
        for (int row = 0; row < resultsjTable.getRowCount(); row++) {
            obj = resultsjTable.getValueAt(row, 0);

            if (!(obj instanceof Occurrence)
                || rowsStatus[row] != UNDEFINED) {
                continue;
            }

            if (((Occurrence)obj).is2Units()) {
                // for each 2 units rule
                for (int i = 0; i < localRules2Units.size(); i ++) {
                    lr = (LocalRule)localRules2Units.get(i);
                    if (lr.apply((Occurrence)obj)) {
                        rowsStatus[row] = (lr.getStatus() == lr.ACCEPTED) ?
                            ACCEPTED : REJECTED;
                        ((Occurrence)obj).setTypeName(lr.getType());
                        break; // the first rule that applies wins
                    }
                }

            } else {
                // for each 3 units rule
                for (int i = 0; i < localRules3Units.size(); i ++) {
                    lr = (LocalRule)localRules3Units.get(i);
                    if (lr.apply((Occurrence)obj)) {
                        rowsStatus[row] = (lr.getStatus() == lr.ACCEPTED) ?
                            ACCEPTED : REJECTED;
                        ((Occurrence)obj).setTypeName(lr.getType());
                        break; // the first rule that applies wins
                    }
                }
            }
        }
    }

    /**
     * Add a new rule in the list of rules for 2 units terms.
     */
    private void lrAdd2UnitsjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrAdd2UnitsjButtonActionPerformed

        LocalRule lr;
        Vector v;

        if (lrStringjTextField.getText().trim().length() < 1) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("Please_provide_the_characters_string."),
                messages.getString("Error_in_input_parameters"),
                JOptionPane.ERROR_MESSAGE);
            lrStringjTextField.requestFocusInWindow();
            return;
        }

        if (lrTypejComboBox.getSelectedItem() == null
            || ((String)lrTypejComboBox.getSelectedItem())
            .equals("--------------")) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("Please_choose_a_type_of_term."),
                messages.getString("Error_in_input_parameters"),
                JOptionPane.ERROR_MESSAGE);
            lrTypejComboBox.requestFocusInWindow();
            return;
        }

        lr = new LocalRule(LocalRule.NO_OPERATOR, null, null,
            lrUnitjComboBox.getSelectedIndex()+1,
            lrPartjComboBox.getSelectedIndex(),
            lrStringjTextField.getText(),
            separatorjTextField.getText().charAt(0),
            (lrTypejComboBox.getSelectedIndex() < ((DefaultComboBoxModel)
            lrTypejComboBox.getModel()).getIndexOf("--------------")) ?
            LocalRule.ACCEPTED : LocalRule.REJECTED,
            (String)lrTypejComboBox.getSelectedItem());

        // update memory table
        localRules2Units.add(lr);

        updateLocalRulesList(localRules2UnitsjList.getModel().getSize(),
                             -1);
    }//GEN-LAST:event_lrAdd2UnitsjButtonActionPerformed

    /**
     * Add a new rule in the list of rules for 3 units terms.
     */
    private void lrAdd3UnitsjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrAdd3UnitsjButtonActionPerformed

        LocalRule lr;

        if (lrStringjTextField.getText().trim().length() < 1) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("Please_provide_the_characters_string."),
                messages.getString("Error_in_input_parameters"),
                JOptionPane.ERROR_MESSAGE);
            lrStringjTextField.requestFocusInWindow();
            return;
        }

        if (lrTypejComboBox.getSelectedItem() == null
            || ((String)lrTypejComboBox.getSelectedItem())
            .equals("--------------")) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("Please_choose_a_type_of_term."),
                messages.getString("Error_in_input_parameters"),
                JOptionPane.ERROR_MESSAGE);
            lrTypejComboBox.requestFocusInWindow();
            return;
        }

        lr = new LocalRule(LocalRule.NO_OPERATOR, null, null,
            lrUnitjComboBox.getSelectedIndex()+1,
            lrPartjComboBox.getSelectedIndex(),
            lrStringjTextField.getText(),
            separatorjTextField.getText().charAt(0),
            (lrTypejComboBox.getSelectedIndex() < ((DefaultComboBoxModel)
            lrTypejComboBox.getModel()).getIndexOf("--------------")) ?
            LocalRule.ACCEPTED : LocalRule.REJECTED,
            (String)lrTypejComboBox.getSelectedItem());

        // update memory table
        localRules3Units.add(lr);

        updateLocalRulesList(-1,localRules3UnitsjList.getModel().getSize());
    }//GEN-LAST:event_lrAdd3UnitsjButtonActionPerformed

    /**
     * Remove from memory tables and lists the selected local rules of the
     * local rules lists.
     */
    private void lrRemovejButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrRemovejButtonActionPerformed

        int index;

        for (int i = localRules2UnitsjList.getSelectedIndices().length-1;
            i >= 0 ; i--) {
            index = (localRules2UnitsjList.getSelectedIndices())[i];
            localRules2Units.removeElementAt(index);
        }

        for (int i = localRules3UnitsjList.getSelectedIndices().length-1;
            i >= 0 ; i--) {
            index = (localRules3UnitsjList.getSelectedIndices())[i];
            localRules3Units.removeElementAt(index);
        }
        updateLocalRulesList(-1, -1);
    }//GEN-LAST:event_lrRemovejButtonActionPerformed

    /**
     * Decompose the selected rules in their direct subcomponents.
     */
    private void lrDecomposejButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrDecomposejButtonActionPerformed

        int index;
        LocalRule lr;
        boolean noOperatorRule = false;

        for (int i = localRules2UnitsjList.getSelectedIndices().length-1;
            i >= 0 ; i--) {
            index = (localRules2UnitsjList.getSelectedIndices())[i];
            lr = (LocalRule)localRules2Units.get(index);

            if (lr.getOperator() == LocalRule.AND
             || lr.getOperator() == LocalRule.OR) {
                localRules2Units.addElement(lr.getNestedRule1());
                localRules2Units.addElement(lr.getNestedRule2());
                localRules2Units.removeElementAt(index);

            } else if (lr.getOperator() == LocalRule.NOT) {
                localRules2Units.setElementAt(lr.getNestedRule1(), index);

            } else if (!noOperatorRule
                       && lr.getOperator() == LocalRule.NO_OPERATOR) {
                lrTypejComboBox.setSelectedItem(lr.getType());
                lrUnitjComboBox.setSelectedIndex(lr.getUnit()-1);
                lrPartjComboBox.setSelectedIndex(lr.getPart());
                lrStringjTextField.setText(lr.getString());
                localRules2Units.removeElementAt(index);
                // we could only decompose one no operator rule
                noOperatorRule = true;
            }
        }

        for (int i = localRules3UnitsjList.getSelectedIndices().length-1;
            i >= 0 ; i--) {
            index = (localRules3UnitsjList.getSelectedIndices())[i];
            lr = (LocalRule)localRules3Units.get(index);

            if (lr.getOperator() == LocalRule.AND
             || lr.getOperator() == LocalRule.OR) {
                localRules3Units.addElement(lr.getNestedRule1());
                localRules3Units.addElement(lr.getNestedRule2());
                localRules3Units.removeElementAt(index);

            } else if (lr.getOperator() == LocalRule.NOT) {
                localRules3Units.setElementAt(lr.getNestedRule1(), index);

            } else if (!noOperatorRule
                       && lr.getOperator() == LocalRule.NO_OPERATOR) {
                lrTypejComboBox.setSelectedItem(lr.getType());
                lrUnitjComboBox.setSelectedIndex(lr.getUnit()-1);
                lrPartjComboBox.setSelectedIndex(lr.getPart());
                lrStringjTextField.setText(lr.getString());
                localRules3Units.removeElementAt(index);
                // we could only decompose one no operator rule
                noOperatorRule = true;
            }
        }
        updateLocalRulesList(-1, -1);
    }//GEN-LAST:event_lrDecomposejButtonActionPerformed

    /**
     * Shift the selected rules down.
     */
    private void lrDownjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrDownjButtonActionPerformed

        int index;
        int selectedIndex2 = -1;
        int selectedIndex3 = -1;

        for (int i = localRules2UnitsjList.getSelectedIndices().length-1;
            i >= 0 ; i--) {
            index = (localRules2UnitsjList.getSelectedIndices())[i];
            if (index == localRules2UnitsjList.getModel().getSize()-1) {
                continue;
            }
            localRules2Units.insertElementAt(
                localRules2Units.get(index), index+2);
            localRules2Units.removeElementAt(index);
            selectedIndex2 = localRules2UnitsjList.getSelectedIndex()+1;
        }

        for (int i = localRules3UnitsjList.getSelectedIndices().length-1;
            i >= 0 ; i--) {
            index = (localRules3UnitsjList.getSelectedIndices())[i];
            if (index == localRules3UnitsjList.getModel().getSize()-1) {
                continue;
            }
            localRules3Units.insertElementAt(
                localRules3Units.get(index), index+2);
            localRules3Units.removeElementAt(index);
            selectedIndex3 = localRules3UnitsjList.getSelectedIndex()+1;
        }
        updateLocalRulesList(selectedIndex2, selectedIndex3);
    }//GEN-LAST:event_lrDownjButtonActionPerformed

    /**
     * Shift the selected rules up.
     */
    private void lrUpjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrUpjButtonActionPerformed

        int index;
        int selectedIndex2 = -1;
        int selectedIndex3 = -1;

        for (int i = 0;
            i < localRules2UnitsjList.getSelectedIndices().length ; i++) {
            index = (localRules2UnitsjList.getSelectedIndices())[i];
            if (index == 0) { continue; }
            localRules2Units.insertElementAt(
                localRules2Units.get(index), index-1);
            localRules2Units.removeElementAt(index+1);
            selectedIndex2 = localRules2UnitsjList.getSelectedIndex()-1;
        }

        for (int i = 0;
            i < localRules3UnitsjList.getSelectedIndices().length ; i++) {
            index = (localRules3UnitsjList.getSelectedIndices())[i];
            if (index == 0) { continue; }
            localRules3Units.insertElementAt(
                localRules3Units.get(index), index-1);
            localRules3Units.removeElementAt(index+1);
            selectedIndex3 = localRules3UnitsjList.getSelectedIndex()-1;
        }
        updateLocalRulesList(selectedIndex2, selectedIndex3);
    }//GEN-LAST:event_lrUpjButtonActionPerformed

    /**
     * Disjonction the 2 selected rules. The status of the resulting rule
     * is the same as the first rule.
     */
    private void lrOrjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrOrjButtonActionPerformed

        int [] index;
        LocalRule prec1Lr, prec2Lr, newLr;
        int selectedIndex2 = -1;
        int selectedIndex3 = -1;

        if (localRules2UnitsjList.getSelectedIndices().length > 0
         && localRules2UnitsjList.getSelectedIndices().length != 2) {
             JOptionPane.showMessageDialog(frame,
                messages.getString("Please_select_exactly_2_rules."),
                messages.getString("Error_in_input_parameters"),
                JOptionPane.ERROR_MESSAGE);

        } else if (localRules2UnitsjList.getSelectedIndices().length == 2) {
            index = localRules2UnitsjList.getSelectedIndices();
            prec1Lr = (LocalRule)localRules2Units.get(index[0]);
            prec2Lr = (LocalRule)localRules2Units.get(index[1]);
            if (prec1Lr.getStatus() != prec2Lr.getStatus()
                || !prec1Lr.getType().equals(prec2Lr.getType())) {
                 JOptionPane.showMessageDialog(frame, messages.getString(
                    "Type_of_both_rules_must_be_the_same."),
                    messages.getString("Error_in_input_parameters"),
                    JOptionPane.ERROR_MESSAGE);
                 return;
            }
            newLr = new LocalRule(LocalRule.OR, prec1Lr, prec2Lr);
            localRules2Units.addElement(newLr);
            localRules2Units.removeElementAt(index[1]);
            localRules2Units.removeElementAt(index[0]);
            selectedIndex2 = localRules2UnitsjList.getModel().getSize()-1;
        }

        if (localRules3UnitsjList.getSelectedIndices().length > 0
         && localRules3UnitsjList.getSelectedIndices().length != 2) {
             JOptionPane.showMessageDialog(frame,
                messages.getString("Please_select_exactly_2_rules."),
                messages.getString("Error_in_input_parameters"),
                JOptionPane.ERROR_MESSAGE);

        } else if (localRules3UnitsjList.getSelectedIndices().length == 2) {
            index = localRules3UnitsjList.getSelectedIndices();
            prec1Lr = (LocalRule)localRules3Units.get(index[0]);
            prec2Lr = (LocalRule)localRules3Units.get(index[1]);
            if (prec1Lr.getStatus() != prec2Lr.getStatus()
                || !prec1Lr.getType().equals(prec2Lr.getType())) {
                 JOptionPane.showMessageDialog(frame, messages.getString(
                    "Type_of_both_rules_must_be_the_same."),
                    messages.getString("Error_in_input_parameters"),
                    JOptionPane.ERROR_MESSAGE);
                 return;
            }
            newLr = new LocalRule(LocalRule.OR, prec1Lr, prec2Lr);
            localRules3Units.addElement(newLr);
            localRules3Units.removeElementAt(index[1]);
            localRules3Units.removeElementAt(index[0]);
            selectedIndex3 = localRules3UnitsjList.getModel().getSize()-1;
        }

        updateLocalRulesList(selectedIndex2, selectedIndex3);
    }//GEN-LAST:event_lrOrjButtonActionPerformed

    /**
     * Conjonction the 2 selected rules. The status of the resulting rule
     * is the same as the first rule.
     */
    private void lrAndjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrAndjButtonActionPerformed

        int [] index;
        LocalRule prec1Lr, prec2Lr, newLr;
        int selectedIndex2 = -1;
        int selectedIndex3 = -1;

        if (localRules2UnitsjList.getSelectedIndices().length > 0
         && localRules2UnitsjList.getSelectedIndices().length != 2) {
             JOptionPane.showMessageDialog(frame,
                messages.getString("Please_select_exactly_2_rules."),
                messages.getString("Error_in_input_parameters"),
                JOptionPane.ERROR_MESSAGE);

        } else if (localRules2UnitsjList.getSelectedIndices().length == 2) {
            index = localRules2UnitsjList.getSelectedIndices();
            prec1Lr = (LocalRule)localRules2Units.get(index[0]);
            prec2Lr = (LocalRule)localRules2Units.get(index[1]);
            if (prec1Lr.getStatus() != prec2Lr.getStatus()
                || !prec1Lr.getType().equals(prec2Lr.getType())) {
                 JOptionPane.showMessageDialog(frame, messages.getString(
                    "Type_of_both_rules_must_be_the_same."),
                    messages.getString("Error_in_input_parameters"),
                    JOptionPane.ERROR_MESSAGE);
                 return;
            }
            newLr = new LocalRule(LocalRule.AND, prec1Lr, prec2Lr);
            localRules2Units.addElement(newLr);
            localRules2Units.removeElementAt(index[1]);
            localRules2Units.removeElementAt(index[0]);
            selectedIndex2 = localRules2UnitsjList.getModel().getSize()-1;
        }

        if (localRules3UnitsjList.getSelectedIndices().length > 0
         && localRules3UnitsjList.getSelectedIndices().length != 2) {
             JOptionPane.showMessageDialog(frame,
                messages.getString("Please_select_exactly_2_rules."),
                messages.getString("Error_in_input_parameters"),
                JOptionPane.ERROR_MESSAGE);

        } else if (localRules3UnitsjList.getSelectedIndices().length == 2) {
            index = localRules3UnitsjList.getSelectedIndices();
            prec1Lr = (LocalRule)localRules3Units.get(index[0]);
            prec2Lr = (LocalRule)localRules3Units.get(index[1]);
            if (prec1Lr.getStatus() != prec2Lr.getStatus()
                || !prec1Lr.getType().equals(prec2Lr.getType())) {
                 JOptionPane.showMessageDialog(frame, messages.getString(
                    "Type_of_both_rules_must_be_the_same."),
                    messages.getString("Error_in_input_parameters"),
                    JOptionPane.ERROR_MESSAGE);
                 return;
            }
            newLr = new LocalRule(LocalRule.AND, prec1Lr, prec2Lr);
            localRules3Units.addElement(newLr);
            localRules3Units.removeElementAt(index[1]);
            localRules3Units.removeElementAt(index[0]);
            selectedIndex3 = localRules3UnitsjList.getModel().getSize()-1;
        }

        updateLocalRulesList(selectedIndex2, selectedIndex3);
    }//GEN-LAST:event_lrAndjButtonActionPerformed

    /**
     * Negation of the selected rules.
     */
    private void lrNotjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrNotjButtonActionPerformed

        int index;
        LocalRule precLr, newLr;

        for (int i = 0;
            i < localRules2UnitsjList.getSelectedIndices().length ; i++) {
            index = (localRules2UnitsjList.getSelectedIndices())[i];
            precLr = (LocalRule)localRules2Units.get(index);
            newLr = new LocalRule(LocalRule.NOT, precLr, null);
            localRules2Units.setElementAt(newLr, index);
        }

        for (int i = 0;
            i < localRules3UnitsjList.getSelectedIndices().length ; i++) {
            index = (localRules3UnitsjList.getSelectedIndices())[i];
            precLr = (LocalRule)localRules3Units.get(index);
            newLr = new LocalRule(LocalRule.NOT, precLr, null);
            localRules3Units.setElementAt(newLr, index);
        }
        updateLocalRulesList(localRules2UnitsjList.getSelectedIndex(),
                             localRules3UnitsjList.getSelectedIndex());
    }//GEN-LAST:event_lrNotjButtonActionPerformed

    /**
     * Update lists of the rules panel from the local rules memory tables.
     *
     * @param selectedIndex2 index of the cell in the 2 units rule list to
     * select after update ; -1 for no selection
     * @param selectedIndex3 index of the cell in the 3 units rule list to
     * select after update ; -1 for no selection
     */
    private void updateLocalRulesList(int selectedIndex2,
                                      int selectedIndex3) {

        DefaultListModel lm;
        String s;

        lm = (DefaultListModel)localRules2UnitsjList.getModel();
        lm.clear();
        for (int i = 0; i < localRules2Units.size(); i ++) {
            s = ((LocalRule)localRules2Units.get(i)).getType();
            s += " ";
            s += ((LocalRule)localRules2Units.get(i)).toString();
            lm.addElement(s);
        }
        if (selectedIndex2 >= 0 && selectedIndex2 < lm.getSize()) {
            localRules2UnitsjList.setSelectedIndex(selectedIndex2);
        }

        lm = (DefaultListModel)localRules3UnitsjList.getModel();
        lm.clear();
        for (int i = 0; i < localRules3Units.size(); i ++) {
            s = ((LocalRule)localRules3Units.get(i)).getType();
            s += " ";
            s += ((LocalRule)localRules3Units.get(i)).toString();
            lm.addElement(s);
        }
        if (selectedIndex3 >= 0 && selectedIndex3 < lm.getSize()) {
            localRules3UnitsjList.setSelectedIndex(selectedIndex3);
        }
    }

    /** 
     * Deselect "Add 2 units rule" button if the rule apply to the third
     * unit.
     */
    private void lrUnitjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lrUnitjComboBoxActionPerformed
        if (lrUnitjComboBox.getSelectedIndex() < 2) {
            lrAdd2UnitsjButton.setEnabled(true);
        } else {
            lrAdd2UnitsjButton.setEnabled(false);
        }
    }//GEN-LAST:event_lrUnitjComboBoxActionPerformed


    /*-***********************
     * Undefined terms panel *
     *************************/

    /**
     * Display the undefined terms of the selected iteration.
     */
    private void undefinedTermsIterationjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_undefinedTermsIterationjComboBoxActionPerformed
        displayTerms(undefinedTermsIterationjComboBox,
                     undefinedTerms, undefinedTermsjTable,
                     false, messages.getString("Undefined_terms"));
    }//GEN-LAST:event_undefinedTermsIterationjComboBoxActionPerformed

    /**
     * Display the undefined terms of the selected relation.
     */
    private void undefinedTermsRelationjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_undefinedTermsRelationjComboBoxActionPerformed
        displayTerms(undefinedTermsRelationjComboBox,
                     undefinedTerms, undefinedTermsjTable,
                     true, messages.getString("Undefined_terms"));
    }//GEN-LAST:event_undefinedTermsRelationjComboBoxActionPerformed

    /*
     * Compute the status of undefined term (occ) from the already accepted
     * and rejected terms.
     *
     * <p>Algorithm :
     *
     * <p>count of precedent accepted terms that contain the same first unit
     * that the undefined term
     *
     * <br> + the same for the second unit (units pair) or third unit
     * (units triple)
     *
     * <br> - count of precedent rejected terms that contain the same first
     * unit that the undefined term
     *
     * <br> - the same for the second unit (units pair) or third unit
     * (units triple)
     *
     * @param occ occurrence to compare with precedents terms
     * @param threshold minimum strictly positive value to accept ou reject
     * occ
     * @return ACCEPTED, REJECTED or UNDEFINED
     */
    private int precedentTermsParameter(Occurrence occ, int threshold) {

        int valueAccept, valueReject, result;
        Occurrence occ2;
        Iterator it;

        valueAccept = 0;
        it = acceptedTerms.iterator();
        while (it.hasNext()) {
            occ2 = (Occurrence)it.next();
            if (occ.is2Units() != occ2.is2Units()) {
                // don't compare 2 and 3 units terms
                continue;
            }
            if (occ.getUnit1().equals(occ2.getUnit1())) {
                valueAccept++;
            }
            if (occ.is2Units()) {
                if (occ.getUnit2().equals(occ2.getUnit2())) {
                    valueAccept++;
                }
            } else {
                if (occ.getUnit3().equals(occ2.getUnit3())) {
                    valueAccept++;
                }
            }
        }
    
        valueReject = 0;
        it = rejectedTerms.iterator();
        while (it.hasNext()) {
            occ2 = (Occurrence)it.next();
            if (occ.is2Units() != occ2.is2Units()) {
                // don't compare 2 and 3 units terms
                continue;
            }
            if (occ.getUnit1().equals(occ2.getUnit1())) {
                valueReject++;
            }
            if (occ.is2Units()) {
                if (occ.getUnit2().equals(occ2.getUnit2())) {
                    valueReject++;
                }
            } else {
                if (occ.getUnit3().equals(occ2.getUnit3())) {
                    valueReject++;
                }
            }
        }

        result = valueAccept-valueReject;

        if (result >= threshold) {
            return ACCEPTED;
        } else if (result <= -threshold) {
            return REJECTED;
        } else {
            return UNDEFINED;
        }
    }

    private void precedentTermsUndefinedTermsjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_precedentTermsUndefinedTermsjButtonActionPerformed

        if (undefinedTermsjTable.getRowCount() > 0) {
            precedentTermsjDialog.show();
        }
    }//GEN-LAST:event_precedentTermsUndefinedTermsjButtonActionPerformed

    /**
     * Listener for the mouse events on the undefined terms table.
     */
    private void undefinedTermsjTableMouseReleased(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_undefinedTermsjTableMouseReleased

        Occurrence occ;

        // display context
        if (evt.getButton() == java.awt.event.MouseEvent.BUTTON2
            && undefinedTermsjTable.getSelectedRowCount() > 0) {

            // get the selected term Occurrence
            occ = (Occurrence)undefinedTermsjTable.getModel().getValueAt(
                undefinedTermsjTable.getSelectedRow(), 0);

            displayContext(occ, undefinedTermsContextjTextPane);

        // if a selection is active and mouse button 3 is pressed shows the
        // popup menu
        } else if (evt.getButton() == java.awt.event.MouseEvent.BUTTON3
            && undefinedTermsjTable.getSelectedRowCount() > 0) {

            undefinejMenuItem.setVisible(false);
            searchStringjMenuItem.setVisible(false);
            resultsjPopupMenu.show(undefinedTermsjTable,
                evt.getX(), evt.getY());
        }
    }//GEN-LAST:event_undefinedTermsjTableMouseReleased

    /**
     * Listener for the keys events on the undefined terms table.
     */
    private void undefinedTermsjTableKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_undefinedTermsjTableKeyReleased

        Object obj;

        if (undefinedTermsjTable.getRowCount() < 1) { return; }

        // display context
        if (evt.getKeyCode() == java.awt.event.KeyEvent.VK_SPACE
            && undefinedTermsjTable.getSelectedRowCount() > 0) {

            // get the selected term Occurrence
            obj = undefinedTermsjTable.getModel().getValueAt(
                      undefinedTermsjTable.getSelectedRow(),
                      undefinedTermsjTable.getSelectedColumn());

            if (obj instanceof Occurrence) {
                displayContext((Occurrence)obj,
                               undefinedTermsContextjTextPane);
            }
        }

    }//GEN-LAST:event_undefinedTermsjTableKeyReleased

    /**
     * Save the undefined terms table content in a file.
     */
    private void saveUndefinedTerms() {

        PrintWriter out;
        String saveFileName;
        Iterator it;
        int iter;
        Occurrence occ;


        if (et == null) { return; }

        saveFileName = (String)optionsDirjComboBox.getSelectedItem()+"-"
            +(new DecimalFormat("00")).format(
                (new Integer(iteration-1)).doubleValue())
            +".uterms";

        try {
            // open the file in writing mode
            out = new PrintWriter(new BufferedWriter(
                      new FileWriter(saveFileName)));

            // save table content in a file
            out.println(messages.getString("Undefined_terms"));
            out.println();
            iter = -1;
            it = undefinedTerms.iterator();
            while(it.hasNext()) {
                occ = (Occurrence)it.next();

                if (occ.getIteration() > iter) {
                    iter = occ.getIteration();
                    out.println();
                    out.println(messages.getString("Iteration")+" "+iter);
                }
                out.println(occ);
            }

            // close the file
            out.close();

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getMessage()
                +" ("+saveFileName+")",
                messages.getString("Error_in_writing_file"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }
    }

    /**
     * Reject undefined selected terms. Remove these terms from memory table
     * and from undefined terms table. Add these terms to the current
     * accepted terms memory tables.
     */
    private void rejectUndefinedTerms(String typeName) {

        JTable ut = undefinedTermsjTable;
        int row;

        for (int i = 0; i < ut.getSelectedRows().length; i++) {

            row = (ut.getSelectedRows())[i];
            Occurrence occ = (Occurrence)ut.getModel().getValueAt(row, 0);

            // remove selected term from undefined terms
            undefinedTerms.remove(occ);

            // add selected term to rejected terms
            occ.setIteration(iteration);
            occ.setTypeName(typeName);
            rejectedTerms.add(occ);
        }
        updateTermsTables();
    }

    /**
     * Accept undefined selected terms. Remove these terms from memory table
     * and from undefined terms table. Add these terms to the current
     * accepted terms memory tables.
     */
    private void acceptUndefinedTerms(String typeName) {

        JTable ut = undefinedTermsjTable;
        int row;

        for (int i = 0; i < ut.getSelectedRows().length; i++) {

            row = (ut.getSelectedRows())[i];
            Occurrence occ = (Occurrence)ut.getModel().getValueAt(row, 0);

            // remove selected term from undefined terms
            undefinedTerms.remove(occ);

            // add selected term to accepted terms
            occ.setIteration(iteration);
            occ.setTypeName(typeName);
            acceptedTerms.add(occ);
            currentAcceptedTerms.add(occ);
            currentAcceptedTermsTreeSet.add((occ).getKey());
        }
        updateTermsTables();
    }


    /*-*******************************
     * Accepted/Rejected Terms panel *
     *********************************/


    /**
     * Display the terms of the selected relation or iteration.
     *
     * @param jComboBox selected combo box of the relation or iteration
     * @param termsLHS contain all terms which can be displayed
     * @param termsJTable table where terms are displayed
     * @param displayRelation true if displaying terms by relation ; false
     * if displaying terms by iteration
     * @param columnName name of the only column of the table
     */
    private void displayTerms(JComboBox jComboBox,
                              LinkedHashSet termsLHS,
                              JTable termsJTable,
                              boolean displayRelation,
                              String columnName) {

        UneditableDefaultTableModel udtm;
        Iterator it;
        Occurrence occ;
        int row, selectedIteration;
        String relation, selectedRelation, selectedType;
        Vector results;
        int occIteration;

        if (termsLHS.size() == 0) {
            // empty the table
            termsJTable.setModel(new UneditableDefaultTableModel());
            return;
        }

        if (jComboBox.getItemCount() == 0) {
            return;
        }

        results = new Vector();
        termsJTable.setModel(new UneditableDefaultTableModel());
        udtm = (UneditableDefaultTableModel)termsJTable.getModel();
        it = termsLHS.iterator();
        row = 0;
        occIteration = -1;

        if (jComboBox.getSelectedIndex() == 0) {
            // display all terms ordered by iteration
            while (it.hasNext()) {
                occ = (Occurrence)it.next();

                if (occ.getIteration() > occIteration) {
                    occIteration = occ.getIteration();
                    results.add("");
                    results.add(messages.getString("Iteration")
                                +" "+occIteration);
                }
                results.add(occ);
            }

        } else if (displayRelation) {
            // display terms of the selected relation
            selectedRelation =
                (String)jComboBox.getSelectedItem();
            while (it.hasNext()) {
                occ = (Occurrence)it.next();

                if (occ.is2Units()) {
                    relation = (String)
                        twoUnitsComboBox.getItemAt(occ.getRelation());
                } else {
                    relation = (String)
                        threeUnitsComboBox.getItemAt(occ.getRelation());
                }

                if (relation.equals(selectedRelation)) {
                    if (occ.getIteration() > occIteration) {
                        occIteration = occ.getIteration();
                        results.add("");
                        results.add(messages.getString("Iteration")
                                    +" "+occIteration);
                    }
                    results.add(occ);
                }
            }

        } else {
            // display terms of the selected iteration
            results.add("");
            results.add(messages.getString("Iteration")+" "
                   +(jComboBox.getSelectedIndex()-1));
            selectedIteration = ((Integer)jComboBox
                                 .getSelectedItem()).intValue();
            while (it.hasNext()) {
                occ = (Occurrence)it.next();
                if (occ.getIteration() == selectedIteration) {
                    results.add(occ);
                }
            }

        }

        // remove terms that do not match the selected type
        if ((termsLHS.equals(acceptedTerms)
            && acceptedTermsTypesjComboBox.getSelectedIndex() != 0)
        || (termsLHS.equals(rejectedTerms)
            && rejectedTermsTypesjComboBox.getSelectedIndex() != 0)) {

            selectedType = (termsLHS.equals(acceptedTerms)) ?
                    (String)acceptedTermsTypesjComboBox.getSelectedItem() :
                    (String)rejectedTermsTypesjComboBox.getSelectedItem();

            for (int i = results.size()-1; i >= 0; i--) {
                if (results.get(i) instanceof Occurrence
                && (((Occurrence)results.get(i)).getTypeName() == null
                    || !((Occurrence)results.get(i)).getTypeName()
                        .equals(selectedType))) {
                    results.remove(i);
                }
            }
        }

        udtm.addColumn(columnName, results);
        termsJTable.setDefaultRenderer(
            termsJTable.getColumnClass(0), new TermsCellRenderer());

    }

    /**
     * Display only the selected type of rejected terms.
     */
    private void rejectedTermsTypesjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_rejectedTermsTypesjComboBoxActionPerformed

        if (rejectedTermsRelationjComboBox.getSelectedIndex() != 0) {
            rejectedTermsRelationjComboBoxActionPerformed(null);
        } else {
            rejectedTermsIterationjComboBoxActionPerformed(null);
        }
    }//GEN-LAST:event_rejectedTermsTypesjComboBoxActionPerformed

    /**
     * Display only the selected type of accepted terms.
     */
    private void acceptedTermsTypesjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_acceptedTermsTypesjComboBoxActionPerformed

        if (acceptedTermsRelationjComboBox.getSelectedIndex() != 0) {
            acceptedTermsRelationjComboBoxActionPerformed(null);
        } else {
            acceptedTermsIterationjComboBoxActionPerformed(null);
        }
    }//GEN-LAST:event_acceptedTermsTypesjComboBoxActionPerformed

    /**
     * Display the accepted terms of the selected iteration.
     */
    private void acceptedTermsIterationjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_acceptedTermsIterationjComboBoxActionPerformed

        String actionCommand;
        
        // limit the automatic updating
        if (evt != null
        && evt.getActionCommand().equals("Programmatically")) { return; }

        actionCommand = acceptedTermsRelationjComboBox.getActionCommand();
        acceptedTermsRelationjComboBox.setActionCommand("Programmatically");
        acceptedTermsRelationjComboBox
            .setSelectedItem(messages.getString("All"));
        acceptedTermsRelationjComboBox.setActionCommand(actionCommand);

        displayTerms(acceptedTermsIterationjComboBox,
                     acceptedTerms, acceptedTermsjTable,
                     false, messages.getString("Accepted_terms"));
    }//GEN-LAST:event_acceptedTermsIterationjComboBoxActionPerformed

    /**
     * Display the accepted terms of the selected relation.
     */
    private void acceptedTermsRelationjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_acceptedTermsRelationjComboBoxActionPerformed

        String actionCommand;
        
        // limit the automatic updating
        if (evt != null
        && evt.getActionCommand().equals("Programmatically")) { return; }

        actionCommand = acceptedTermsIterationjComboBox.getActionCommand();
        acceptedTermsIterationjComboBox.setActionCommand("Programmatically");
        acceptedTermsIterationjComboBox
            .setSelectedItem(messages.getString("All"));
        acceptedTermsIterationjComboBox.setActionCommand(actionCommand);

        displayTerms(acceptedTermsRelationjComboBox,
                     acceptedTerms, acceptedTermsjTable,
                     true, messages.getString("Accepted_terms"));
    }//GEN-LAST:event_acceptedTermsRelationjComboBoxActionPerformed

    /**
     * Display the rejected terms of the selected iteration.
     */
    private void rejectedTermsIterationjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_rejectedTermsIterationjComboBoxActionPerformed

        String actionCommand;
        
        // limit the automatic updating
        if (evt != null
        && evt.getActionCommand().equals("Programmatically")) { return; }

        actionCommand = rejectedTermsRelationjComboBox.getActionCommand();
        rejectedTermsRelationjComboBox.setActionCommand("Programmatically");
        rejectedTermsRelationjComboBox
            .setSelectedItem(messages.getString("All"));
        rejectedTermsRelationjComboBox.setActionCommand(actionCommand);
        
        displayTerms(rejectedTermsIterationjComboBox,
                     rejectedTerms, rejectedTermsjTable,
                     false, messages.getString("Rejected_terms"));
    }//GEN-LAST:event_rejectedTermsIterationjComboBoxActionPerformed

    /**
     * Display the rejected terms of the selected relation.
     */
    private void rejectedTermsRelationjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_rejectedTermsRelationjComboBoxActionPerformed

        String actionCommand;
        
        // limit the automatic updating
        if (evt != null
        && evt.getActionCommand().equals("Programmatically")) { return; }

        actionCommand = rejectedTermsIterationjComboBox.getActionCommand();
        rejectedTermsIterationjComboBox.setActionCommand("Programmatically");
        rejectedTermsIterationjComboBox
            .setSelectedItem(messages.getString("All"));
        rejectedTermsIterationjComboBox.setActionCommand(actionCommand);
        
        displayTerms(rejectedTermsRelationjComboBox,
                     rejectedTerms, rejectedTermsjTable,
                     true, messages.getString("Rejected_terms"));
    }//GEN-LAST:event_rejectedTermsRelationjComboBoxActionPerformed

    /**
     * Listener for the mouse events on the rejected terms table.
     */
    private void rejectedTermsjTableMouseReleased(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_rejectedTermsjTableMouseReleased

        Occurrence occ;

        // display context
        if (evt.getButton() == java.awt.event.MouseEvent.BUTTON2
            && rejectedTermsjTable.getSelectedRowCount() > 0) {

            // get the selected term Occurrence
            occ = (Occurrence)rejectedTermsjTable.getModel().getValueAt(
                rejectedTermsjTable.getSelectedRow(), 0);

            displayContext(occ, termsContextjTextPane);

        // if a selection is active and mouse button 3 is pressed shows the
        // popup menu
        } else if (evt.getButton() == java.awt.event.MouseEvent.BUTTON3
            && rejectedTermsjTable.getSelectedRowCount() > 0) {

            undefinejMenuItem.setVisible(true);
            searchStringjMenuItem.setVisible(false);
            resultsjPopupMenu.show(rejectedTermsjTable,
                evt.getX(), evt.getY());
        }
    }//GEN-LAST:event_rejectedTermsjTableMouseReleased

    /**
     * Listener for the keys events on the rejected terms table.
     */
    private void rejectedTermsjTableKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_rejectedTermsjTableKeyReleased

        Object obj;

        if (rejectedTermsjTable.getRowCount() < 1) { return; }

        // display context
        if (evt.getKeyCode() == java.awt.event.KeyEvent.VK_SPACE
            && rejectedTermsjTable.getSelectedRowCount() > 0) {

            // get the selected term Occurrence
            obj = rejectedTermsjTable.getModel().getValueAt(
                      rejectedTermsjTable.getSelectedRow(),
                      rejectedTermsjTable.getSelectedColumn());

            if (obj instanceof Occurrence) {
                displayContext((Occurrence)obj, termsContextjTextPane);
            }
        }

    }//GEN-LAST:event_rejectedTermsjTableKeyReleased

    /**
     * Save the rejected terms table content in a file.
     */
    private void saveRejectedTerms() {

        PrintWriter out;
        String saveFileName;
        Iterator it;
        int iter;
        Occurrence occ;


        if (et == null) { return; }

        saveFileName = (String)optionsDirjComboBox.getSelectedItem()+"-"
            +(new DecimalFormat("00")).format(
                (new Integer(iteration-1)).doubleValue())
            +".rterms";

        try {
            // open the file in writing mode
            out = new PrintWriter(new BufferedWriter(
                      new FileWriter(saveFileName)));

            // save table content in a file
            out.println(messages.getString("Rejected_terms"));
            out.println();
            iter = -1;
            it = rejectedTerms.iterator();
            while(it.hasNext()) {
                occ = (Occurrence)it.next();

                if (occ.getIteration() > iter) {
                    iter = occ.getIteration();
                    out.println();
                    out.println(messages.getString("Iteration")+" "+iter);
                }
                out.println(occ);
            }

            // close the file
            out.close();

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getMessage()
                +" ("+saveFileName+")",
                messages.getString("Error_in_writing_file"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }
    }

    /**
     * Update the terms table with the selected filters.
     * Iteration has the priority on relation.
     * FIXME: no more need to handle priority
     */
    private void updateTermsTables() {

        // iteration has the priority
        if (rejectedTermsIterationjComboBox.getSelectedItem()
            .equals(messages.getString("All"))) {
            rejectedTermsRelationjComboBox.setSelectedIndex(
                rejectedTermsRelationjComboBox.getSelectedIndex());
       } else {
            rejectedTermsIterationjComboBox.setSelectedIndex(
                rejectedTermsIterationjComboBox.getSelectedIndex());
        }

        if (acceptedTermsIterationjComboBox.getSelectedItem()
            .equals(messages.getString("All"))) {
            acceptedTermsRelationjComboBox.setSelectedIndex(
                acceptedTermsRelationjComboBox.getSelectedIndex());
       } else {
            acceptedTermsIterationjComboBox.setSelectedIndex(
                acceptedTermsIterationjComboBox.getSelectedIndex());
       }

        if (undefinedTermsIterationjComboBox.getSelectedItem()
            .equals(messages.getString("All"))) {
            undefinedTermsRelationjComboBox.setSelectedIndex(
                undefinedTermsRelationjComboBox.getSelectedIndex());
       } else {
            undefinedTermsIterationjComboBox.setSelectedIndex(
                undefinedTermsIterationjComboBox.getSelectedIndex());
       }
    }

    /**
     * Undefine selected rejected terms. Remove these terms from memory
     * table and from rejected terms table.
     */
    private void undefineRejectedTerms() {

        JTable rt = rejectedTermsjTable;
        int row;

        for (int i = 0; i < rt.getSelectedRows().length; i++) {

            row = (rt.getSelectedRows())[i];
            Occurrence occ = (Occurrence)rt.getModel().getValueAt(row, 0);

            // remove selected term from rejected terms
            rejectedTerms.remove(occ);

            // add selected term to the undefined terms
            occ.setIteration(iteration);
            occ.setTypeName(null);
            undefinedTerms.add(occ);
        }
        updateTermsTables();
    }

    /**
     * Accept selected rejected terms. Remove these terms from memory table
     * and from rejected terms table. Add these terms to the current
     * accepted terms memory tables.
     */
    private void acceptRejectedTerms(String typeName) {

        JTable rt = rejectedTermsjTable;
        int row;

        for (int i = 0; i < rt.getSelectedRows().length; i++) {

            row = (rt.getSelectedRows())[i];
            Occurrence occ = (Occurrence)rt.getModel().getValueAt(row, 0);

            // remove selected term from rejected terms
            rejectedTerms.remove(occ);

            // add selected term to the accepted terms
            occ.setIteration(iteration);
            occ.setTypeName(typeName);
            acceptedTerms.add(occ);
            currentAcceptedTerms.add(occ);
            currentAcceptedTermsTreeSet.add((occ).getKey());
        }
        updateTermsTables();
    }

    /**
     * Change the type of the selected rejected terms.
     */
    private void changeRejectedTermsType(String typeName) {

        JTable rt = rejectedTermsjTable;
        int row;

        for (int i = 0; i < rt.getSelectedRows().length; i++) {

            row = (rt.getSelectedRows())[i];
            Occurrence occ = (Occurrence)rt.getModel().getValueAt(row, 0);
            occ.setTypeName(typeName);
        }
        updateTermsTables();
    }

    /**
     * Listener for the mouse events on the accepted terms table.
     */
    private void acceptedTermsjTableMouseReleased(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_acceptedTermsjTableMouseReleased

        Occurrence occ;

        // display context
        if (evt.getButton() == java.awt.event.MouseEvent.BUTTON2
            && acceptedTermsjTable.getSelectedRowCount() > 0) {

            // get the selected term Occurrence
            occ = (Occurrence)acceptedTermsjTable.getModel().getValueAt(
                      acceptedTermsjTable.getSelectedRow(), 0);

            displayContext(occ, termsContextjTextPane);

        // if a selection is active and mouse button 3 is pressed shows the
        // popup menu
        } else if (evt.getButton() == java.awt.event.MouseEvent.BUTTON3
            && acceptedTermsjTable.getSelectedRowCount() > 0) {

            undefinejMenuItem.setVisible(true);
            searchStringjMenuItem.setVisible(false);
            resultsjPopupMenu.show(acceptedTermsjTable,
                evt.getX(), evt.getY());
        }
    }//GEN-LAST:event_acceptedTermsjTableMouseReleased

    /**
     * Listener for the keys events on the accepted terms table.
     */
    private void acceptedTermsjTableKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_acceptedTermsjTableKeyReleased

        Object obj;

        if (acceptedTermsjTable.getRowCount() < 1) { return; }

        // display context
        if (evt.getKeyCode() == java.awt.event.KeyEvent.VK_SPACE
            && acceptedTermsjTable.getSelectedRowCount() > 0) {

            // get the selected term Occurrence
            obj = acceptedTermsjTable.getModel().getValueAt(
                      acceptedTermsjTable.getSelectedRow(),
                      acceptedTermsjTable.getSelectedColumn());


            if (obj instanceof Occurrence) {
                displayContext((Occurrence)obj, termsContextjTextPane);
            }
        }

    }//GEN-LAST:event_acceptedTermsjTableKeyReleased

    /**
     * Undefine a accepted term. Remove this term from memory table and from
     * accepted terms table. Replace the grouped term with its ungrouped
     * units in the input file.
     */
    private void undefineAcceptedTerms() {

        JTable at = acceptedTermsjTable;
        int row;
        Occurrence occ, occ2;
        Iterator it;
        boolean grouped;
        String groupedUnits;

        for (int i = 0; i < at.getSelectedRows().length; i++) {

            row = (at.getSelectedRows())[i];
            occ = (Occurrence)at.getModel().getValueAt(row, 0);

            // search if the accepted term is already contained in another
            // accepted term of the precedent iterations
            grouped = false;
            it = acceptedTerms.iterator();
            while (it.hasNext()) {
                occ2 = (Occurrence)it.next();
                if (occ.is2Units()) {
                    groupedUnits = occ.getUnit1()+"-"+occ.getUnit2();
                } else {
                    groupedUnits = occ.getUnit1()+"-"+occ.getUnit2()
                        +"-"+occ.getUnit3();
                }
                if (occ2.getUnit1().equals(groupedUnits)
                 || occ2.getUnit2().equals(groupedUnits)
                 || (occ2.is3Units()
                        && occ2.getUnit3().equals(groupedUnits))) {
                    grouped = true; break;
                }
            }

            // return if the term doesn't belong to the last iteration
            if (grouped && occ.getIteration() != iteration-2) {
                JOptionPane.showMessageDialog(frame,
                    messages.getString("The_term_must_belong_to_the_last"
                                       +"_iteration."),
                    messages.getString("Error_in_modifing_term"),
                    JOptionPane.ERROR_MESSAGE);
                return;
            }

            // initialise a new ExtractTerms
            et = newExtractTerms();

            if (et == null) { return; }

            LinkedHashSet lhs = new LinkedHashSet();
            lhs.add(occ);

            // ungroup term
            et.replace(lhs, "-", et.getInputFileName(),
                       et.getInputFileName(), et.UNGROUP_MODE);

            // remove selected term from accepted terms
            acceptedTerms.remove(occ);

            // add selected term to undefined terms
            occ.setIteration(iteration);
            occ.setTypeName(null);
            undefinedTerms.add(occ);
        }
        updateTermsTables();
    }

    /**
     * Reject a accepted term. Remove this term from memory table and from
     * accepted terms table. Replace the grouped term with its ungrouped
     * units in the input file. Add this term to the current accepted terms
     * memory tables.
     */
    private void rejectAcceptedTerms(String typeName) {

        JTable at = acceptedTermsjTable;
        int row;
        Occurrence occ, occ2;
        Iterator it;
        boolean grouped;
        String groupedUnits;

        for (int i = 0; i < at.getSelectedRows().length; i++) {

            row = (at.getSelectedRows())[i];
            occ = (Occurrence)at.getModel().getValueAt(row, 0);

            // search if the accepted term is already contained in another
            // accepted term of the precedent iterations
            grouped = false;
            it = acceptedTerms.iterator();
            while (it.hasNext()) {
                occ2 = (Occurrence)it.next();
                if (occ.is2Units()) {
                    groupedUnits = occ.getUnit1()+"-"+occ.getUnit2();
                } else {
                    groupedUnits = occ.getUnit1()+"-"+occ.getUnit2()
                        +"-"+occ.getUnit3();
                }
                if (occ2.getUnit1().equals(groupedUnits)
                 || occ2.getUnit2().equals(groupedUnits)
                 || (occ2.is3Units()
                        && occ2.getUnit3().equals(groupedUnits))) {
                    grouped = true; break;
                }
            }

            // return if the term doesn't belong to the last iteration
            if (grouped && occ.getIteration() != iteration-2) {
                JOptionPane.showMessageDialog(frame,
                    messages.getString("The_term_must_belong_to_the_last"
                                       +"_iteration."),
                    messages.getString("Error_in_modifing_term"),
                    JOptionPane.ERROR_MESSAGE);
                return;
            }

            // initialise a new ExtractTerms
            et = newExtractTerms();

            if (et == null) { return; }

            LinkedHashSet lhs = new LinkedHashSet();
            lhs.add(occ);
        
            // ungroup term
            et.replace(lhs, "-", et.getInputFileName(),
                       et.getInputFileName(), et.UNGROUP_MODE);

            // remove selected term from accepted terms
            acceptedTerms.remove(occ);

            // add selected term to rejected terms
            occ.setIteration(iteration);
            occ.setTypeName(typeName);
            rejectedTerms.add(occ);
        }
        updateTermsTables();
    }

    /**
     * Change the type of the selected accepted terms.
     */
    private void changeAcceptedTermsType(String typeName) {

        JTable at = acceptedTermsjTable;
        int row;

        for (int i = 0; i < at.getSelectedRows().length; i++) {

            row = (at.getSelectedRows())[i];
            Occurrence occ = (Occurrence)at.getModel().getValueAt(row, 0);
            occ.setTypeName(typeName);
        }
        updateTermsTables();
    }

    /**
     * Save the accepted terms table content in a file.
     */
    private void saveAcceptedTerms() {

        PrintWriter out;
        String saveFileName;
        Iterator it;
        int iter;
        Occurrence occ;


        if (et == null) { return; }

        saveFileName = (String)optionsDirjComboBox.getSelectedItem()+"-"
            +(new DecimalFormat("00")).format(
                (new Integer(iteration-1)).doubleValue())
            +".aterms";

        try {
            // open the file in writing mode
            out = new PrintWriter(new BufferedWriter(
                      new FileWriter(saveFileName)));

            // save table content in a file
            out.println(messages.getString("Accepted_terms"));
            out.println();
            iter = -1;
            it = acceptedTerms.iterator();
            while(it.hasNext()) {
                occ = (Occurrence)it.next();

                if (occ.getIteration() > iter) {
                    iter = occ.getIteration();
                    out.println();
                    out.println(messages.getString("Iteration")+" "+iter);
                }
                out.println(occ);
            }

            // close the file
            out.close();

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getMessage()
                +" ("+saveFileName+")",
                messages.getString("Error_in_writing_file"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }
    }

    /**
     * Update the memory tables containing the terms according to the status
     * of cells in the table of results.
     */
    private void updateMemoryTermsTables() {

        // for each row in the results table
        for (int row = 0; row < resultsjTable.getRowCount(); row++) {

            Object o = resultsjTable.getModel().getValueAt(row, 0);

            if (!(o instanceof Occurrence)) {
                continue;
            }

            switch (rowsStatus[row]) {

            case ACCEPTED:
                acceptedTerms.add((Occurrence)o);
                currentAcceptedTerms.add((Occurrence)o);
                currentAcceptedTermsTreeSet.add(((Occurrence)o).getKey());
                break;

            case REJECTED:
                rejectedTerms.add((Occurrence)o);
                break;

            case UNDEFINED:
                undefinedTerms.add((Occurrence)o);
                break;
            }

            currentSkippedTermsTreeSet.add(((Occurrence)o).getKey());
        }
    }


    /*-*******************
     * Expressions panel *
     *********************/

    /**
     * Update all lists using elements of the expressions panel, that is to
     * say regex names, formulas names and type names.
     */
    private void updateExpressions() {
        TableModel tm;
        int selectedIndex;
        String line;

        tm = regexjTable.getModel();

        // unit definition name
        selectedIndex = unitComboBox.getSelectedIndex();
        unitComboBox.removeAllItems();
        for (int row = 0; row < 10; row++) {
            if (tm.getValueAt(row, 0) != null
                && !((String)tm.getValueAt(row, 0)).trim().equals("")) {
                unitComboBox.addItem(tm.getValueAt(row, 0));
            }
        }
        if (unitComboBox.getItemCount() > selectedIndex) {
            unitComboBox.setSelectedIndex(selectedIndex);
        }

        // sentence/phrase separators definition name
        selectedIndex = sentPhrComboBox.getSelectedIndex();
        sentPhrComboBox.removeAllItems();
        for (int row = 0; row < 10; row++) {
            if (tm.getValueAt(row, 2) != null
                && !((String)tm.getValueAt(row, 2)).trim().equals("")) {
                sentPhrComboBox.addItem(tm.getValueAt(row, 2));
            }
        }
        if (sentPhrComboBox.getItemCount() > selectedIndex) {
            sentPhrComboBox.setSelectedIndex(selectedIndex);
        }

        // name of text selector name
        selectedIndex = nameOfTextjComboBox.getSelectedIndex();
        nameOfTextjComboBox.removeAllItems();
        for (int row = 0; row < 10; row++) {
            if (tm.getValueAt(row, 4) != null
                && !((String)tm.getValueAt(row, 4)).trim().equals("")) {
                nameOfTextjComboBox.addItem(tm.getValueAt(row, 4));
            }
        }
        if (nameOfTextjComboBox.getItemCount() > selectedIndex) {
            nameOfTextjComboBox.setSelectedIndex(selectedIndex);
        }

        tm = regex2jTable.getModel();

        // update terms relation list
        acceptedTermsRelationjComboBox.removeAllItems();
        rejectedTermsRelationjComboBox.removeAllItems();
        undefinedTermsRelationjComboBox.removeAllItems();
        acceptedTermsRelationjComboBox.addItem(messages.getString("All"));
        rejectedTermsRelationjComboBox.addItem(messages.getString("All"));
        undefinedTermsRelationjComboBox.addItem(messages.getString("All"));

        // units pair selector name
        selectedIndex = twoUnitsComboBox.getSelectedIndex();
        twoUnitsComboBox.removeAllItems();
        for (int row = 0; row < 100; row++) {
            if (tm.getValueAt(row, 0) != null
                && !((String)tm.getValueAt(row, 0)).trim().equals("")) {
                twoUnitsComboBox.addItem(tm.getValueAt(row, 0));
                acceptedTermsRelationjComboBox
                    .addItem(tm.getValueAt(row, 0));
                rejectedTermsRelationjComboBox
                    .addItem(tm.getValueAt(row, 0));
                undefinedTermsRelationjComboBox
                    .addItem(tm.getValueAt(row, 0));
            }
        }
        if (twoUnitsComboBox.getItemCount() > selectedIndex) {
            twoUnitsComboBox.setSelectedIndex(selectedIndex);
        }

        // units triple selector name
        selectedIndex = threeUnitsComboBox.getSelectedIndex();
        threeUnitsComboBox.removeAllItems();
        for (int row = 0; row < 100; row++) {
            if (tm.getValueAt(row, 4) != null
                && !((String)tm.getValueAt(row, 4)).trim().equals("")) {
                threeUnitsComboBox.addItem(tm.getValueAt(row, 4));
                acceptedTermsRelationjComboBox.addItem(tm.getValueAt(row, 4));
                rejectedTermsRelationjComboBox.addItem(tm.getValueAt(row, 4));
                undefinedTermsRelationjComboBox.addItem(tm.getValueAt(row, 4));
            }
        }
        if (threeUnitsComboBox.getItemCount() > selectedIndex) {
            threeUnitsComboBox.setSelectedIndex(selectedIndex);
        }

        tm = formulasjTable.getModel();

        // formula name
        selectedIndex = measuresjComboBox.getSelectedIndex();
        measuresjComboBox.removeAllItems();
        for (int row = 0; row < 10; row++) {
            if (tm.getValueAt(row, 0) != null
                && !((String)tm.getValueAt(row, 0)).trim().equals("")) {
                measuresjComboBox.addItem(tm.getValueAt(row, 0));
            }
        }
        if (measuresjComboBox.getItemCount() > selectedIndex) {
            measuresjComboBox.setSelectedIndex(selectedIndex);
        }

        // update the type of accepted and rejected terms menu items
        // and combo boxes
        tm = termsTypesjTable.getModel();
        lrTypejComboBox.removeAllItems();
        acceptedTermsTypesjComboBox.removeAllItems();
        acceptedTermsTypesjComboBox.addItem(messages.getString("All"));
        for (int row = 0; row < 50; row++) {
            line = (String)tm.getValueAt(row, 0);
            if (line != null && !line.trim().equals("")) {
                lrTypejComboBox.addItem(line);
                acceptedTermsTypesjComboBox.addItem(line);
                acceptjMenu.getItem(row).setText(line);
                acceptjMenu.getItem(row).setVisible(true);
            } else {
                acceptjMenu.getItem(row).setVisible(false);
            }
        }
        lrTypejComboBox.addItem("--------------");
        rejectedTermsTypesjComboBox.removeAllItems();
        rejectedTermsTypesjComboBox.addItem(messages.getString("All"));
        for (int row = 0; row < 50; row++) {
            line = (String)tm.getValueAt(row, 1);
            if (line != null && !line.trim().equals("")) {
                lrTypejComboBox.addItem(line);
                rejectedTermsTypesjComboBox.addItem(line);
                rejectjMenu.getItem(row).setText(line);
                rejectjMenu.getItem(row).setVisible(true);
            } else {
                rejectjMenu.getItem(row).setVisible(false);
            }
        }
    }

    /**
     * Disable filled cells in tables of the Expressions panel.
     */
    private void disableExpressions() {

        ExpressionsTableModel etm;

        etm = (ExpressionsTableModel)regexjTable.getModel();
        for (int col = 0; col < etm.getColumnCount(); col++) {
            for (int row = 0; row < etm.getRowCount(); row++) {
                if (etm.getValueAt(row, col) != null
                && ((String)etm.getValueAt(row, col)).trim().length() > 0) {
                    etm.setRowMaxDisabled(col, row);
                }
            }
        }
        etm = (ExpressionsTableModel)regex2jTable.getModel();
        for (int col = 0; col < etm.getColumnCount(); col++) {
            for (int row = 0; row < etm.getRowCount(); row++) {
                if (etm.getValueAt(row, col) != null
                && (etm.getValueAt(row, col) instanceof String
                  && ((String)etm.getValueAt(row, col)).trim().length() > 0)
                || (etm.getValueAt(row, col) instanceof Integer
                   && (((Integer)etm.getValueAt(row, col)).intValue() > 0)))
                {
                    etm.setRowMaxDisabled(col, row);
                }
            }
        }
        etm = (ExpressionsTableModel)formulasjTable.getModel();
        for (int col = 0; col < etm.getColumnCount(); col++) {
            for (int row = 0; row < etm.getRowCount(); row++) {
                if (etm.getValueAt(row, col) != null
                && ((String)etm.getValueAt(row, col)).trim().length() > 0) {
                    etm.setRowMaxDisabled(col, row);
                }
            }
        }
        etm = (ExpressionsTableModel)termsTypesjTable.getModel();
        for (int col = 0; col < etm.getColumnCount(); col++) {
            for (int row = 0; row < etm.getRowCount(); row++) {
                if (etm.getValueAt(row, col) != null
                && ((String)etm.getValueAt(row, col)).trim().length() > 0) {
                    etm.setRowMaxDisabled(col, row);
                }
            }
        }
        separatorjTextField.setEnabled(false);
    }

    /**
     * Enable all cells in tables of the Expressions panel.
     */
    private void enableExpressions() {

        ExpressionsTableModel etm;

        etm = (ExpressionsTableModel)regexjTable.getModel();
        for (int col = 0; col < etm.getColumnCount(); col++) {
            etm.setRowMaxDisabled(col, -1);
        }
        etm = (ExpressionsTableModel)regex2jTable.getModel();
        for (int col = 0; col < etm.getColumnCount(); col++) {
            etm.setRowMaxDisabled(col, -1);
        }
        etm = (ExpressionsTableModel)formulasjTable.getModel();
        for (int col = 0; col < etm.getColumnCount(); col++) {
            etm.setRowMaxDisabled(col, -1);
        }
        etm = (ExpressionsTableModel)termsTypesjTable.getModel();
        for (int col = 0; col < etm.getColumnCount(); col++) {
            etm.setRowMaxDisabled(col, -1);
        }
        separatorjTextField.setEnabled(true);
    }

    /**
     * Display help on tags.
     */
    private void regexHelpTagsjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_regexHelpTagsjButtonActionPerformed

        HTMLWindow("file:"
                   +System.getProperty("user.dir")
                   +System.getProperty("file.separator")
                   +"doc"
                   +System.getProperty("file.separator")
                   +"tags.html",
                   messages.getString("Help_on_tags"));
    }//GEN-LAST:event_regexHelpTagsjButtonActionPerformed

    /**
     * Display help on regular expressions.
     */
    private void regexHelpjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_regexHelpjButtonActionPerformed

        HTMLWindow("file:"
                   +System.getProperty("user.dir")
                   +System.getProperty("file.separator")
                   +"doc"
                   +System.getProperty("file.separator")
                   +"pattern.html",
                   messages.getString("Help_on_regular_expressions"));

    }//GEN-LAST:event_regexHelpjButtonActionPerformed

    /**
     * Add or remove a FocusListener on regex tables.
     */
    private void regexTestjToggleButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_regexTestjToggleButtonActionPerformed

        if (regexTestjToggleButton.isSelected()) {
            regexjTable.addFocusListener(new java.awt.event.FocusAdapter() {
                public void focusGained(java.awt.event.FocusEvent evt) {
                    testRegexFocusGained(evt);
                }
            });
            regex2jTable.addFocusListener(new java.awt.event.FocusAdapter(){
                public void focusGained(java.awt.event.FocusEvent evt) {
                    testRegexFocusGained(evt);
                }
            });

            statejLabel.setText(
                messages.getString("Select_a_cell_that_contains_a_regex."));

        } else {
            // remove the last focus listener added
            // it suppose that no other method that this method will add one
            regexjTable.removeFocusListener(
                regexjTable.getFocusListeners()
                    [regexjTable.getFocusListeners().length-1]);
            regex2jTable.removeFocusListener(
                regex2jTable.getFocusListeners()
                    [regex2jTable.getFocusListeners().length-1]);
        }
    }//GEN-LAST:event_regexTestjToggleButtonActionPerformed

    /**
     * Test selected regular expression on the first 10 matches of input
     * file.
     */
    private void testRegexFocusGained(java.awt.event.FocusEvent evt) {

        BufferedReader br;
        String lineContent, match;
        String regex;
        String inputFileName;
        Matcher matcher;
        Pattern pattern;
        Vector matches;
        int i, nbOfMatch = 10;
        int[] regexColumns = {1, 3, 5};
        int[] regex2Columns = {0, 1, 2, 4, 5, 6, 7};
        Iterator it;
        Component sc;

        regex = "";

        // remove the focus to avoid infinite loop on this method
        regexFormulasjPanel.requestFocusInWindow();

        // selected regex cell belong to the first regex table
        if (regexjTable.equals(evt.getComponent())
            && Arrays.binarySearch(regexColumns,
                regexjTable.getSelectedColumn()) >= 0) {
            regex = (String)regexjTable
                .getValueAt(regexjTable.getSelectedRow(),
                            regexjTable.getSelectedColumn());

        // selected regex cell belong to the second regex table
        } else if (regex2jTable.equals(evt.getComponent())
                   && Arrays.binarySearch(regex2Columns,
                   regex2jTable.getSelectedColumn()) >= 0) {

            // concat the 2 regex of the units pair selector
            if (regex2jTable.getSelectedColumn() == 0) {
                regex = (String)regex2jTable
                    .getValueAt(regex2jTable.getSelectedRow(), 1)
                    +(String)regex2jTable
                    .getValueAt(regex2jTable.getSelectedRow(), 2);

            // concat the 3 regex of the units triple selector
            } else if (regex2jTable.getSelectedColumn() == 4) {
                regex = (String)regex2jTable
                    .getValueAt(regex2jTable.getSelectedRow(), 5)
                    +(String)regex2jTable
                    .getValueAt(regex2jTable.getSelectedRow(), 6)
                    +(String)regex2jTable
                    .getValueAt(regex2jTable.getSelectedRow(), 7);

            } else {
                regex = (String)regex2jTable
                    .getValueAt(regex2jTable.getSelectedRow(),
                                regex2jTable.getSelectedColumn());
            }
        }

        // selected column isn't a regex column
        if (regex == null || regex.trim().equals("")) {
            return;
        }

        // test the validity of regular expressions
        try {
            pattern = Pattern.compile(regex);

        } catch (PatternSyntaxException pse) {
            JOptionPane.showMessageDialog(frame,
                pse.getLocalizedMessage()+"\n"
                +messages.getString("Bad_syntax_in_regular_expression."),
                messages.getString("Bad_parameters"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }

        try {
            // open the input file in reading mode
            inputFileName = inputFileNameFromInterface();
            if (inputFileName == null) { return; }
            br = new BufferedReader(new FileReader(inputFileName));

            // search matching of the regex in the input file
            lineContent = br.readLine();
            matches = new Vector();
            i = 0;
            while (lineContent != null && i < nbOfMatch) {
                matcher = pattern.matcher(lineContent);

                while(matcher.find() && i < nbOfMatch) {
                    if (matcher.groupCount() == 1) {
                        matches.add(matcher.group(1));
                    } else {
                        matches.add(matcher.group());
                    }
                    i++;
                }
                lineContent = br.readLine();
            }

            // close the input file
            br.close();

        } catch (FileNotFoundException fnfe) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("File_not_found")+"\n"
                +fnfe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return;

        } catch(IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("Error_in_reading_file")+"\n"
                +ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return;
        }

        // format the results
        match = "";
        it = matches.iterator();
        while (it.hasNext()) {
            match += "["+(String)it.next()+"]\n";
        }

        if (match.equals("")) {
            match = messages.getString("No_result.");
        }

        JOptionPane.showMessageDialog(frame,
            messages.getString(
                "test_selected_regular_expression_on_the_first_10")
            +"\n\n"+match,
            messages.getString("Test_regular_expression"),
            JOptionPane.INFORMATION_MESSAGE);
    }


    /*-******************
     * Extraction panel *
     ********************/

    private void twoUnitsComboBoxFocusGained(java.awt.event.FocusEvent evt) {//GEN-FIRST:event_twoUnitsComboBoxFocusGained

        twoUnitsComboBox.setForeground(Color.RED);
        threeUnitsComboBox.setForeground(Color.BLACK);
    }//GEN-LAST:event_twoUnitsComboBoxFocusGained

    private void threeUnitsComboBoxFocusGained(java.awt.event.FocusEvent evt) {//GEN-FIRST:event_threeUnitsComboBoxFocusGained

        threeUnitsComboBox.setForeground(Color.RED);
        twoUnitsComboBox.setForeground(Color.BLACK);
    }//GEN-LAST:event_threeUnitsComboBoxFocusGained

    /*
     * Update optionsComboxBox when optionsDirjCombobox is changed by user
     */
    private void optionsDirjComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_optionsDirjComboBoxActionPerformed

        File[] optionsFiles;
        File currentDir;
        File currentOptionsDir;
        String selectedItem, actionCommand;

        // limit the automatic updating of file lists
        if (evt.getActionCommand().equals("Programmatically")) { return; }

        if (optionsDirjComboBox.getSelectedItem() != null
        && ((String)optionsDirjComboBox.getSelectedItem())
            .trim().length() > 0) {
            updateFilesList();

        } else {
            JOptionPane.showMessageDialog(frame,
                messages.getString("You_must_provide_a_directory_name."),
                messages.getString("Error_in_loading_options"),
                JOptionPane.ERROR_MESSAGE);
            optionsDirjComboBox.requestFocusInWindow();
            return;
        }
    }//GEN-LAST:event_optionsDirjComboBoxActionPerformed

    /**
     * Set the status of the selected cells of table to rejected.
     */
    private void rejectTypejMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_rejectTypejMenuItemActionPerformed

        KeyboardFocusManager focusManager =
            KeyboardFocusManager.getCurrentKeyboardFocusManager();

        // selected component before the pop-menu got focus.
        Component sc = focusManager.getPermanentFocusOwner();

        if (resultsjTable.equals(sc)) {
            for (int row = 0; row < resultsjTable.getSelectedRowCount(); row++){
                rowsStatus[(resultsjTable.getSelectedRows())[row]] = REJECTED;
                ((Occurrence)resultsjTable.getModel()
                    .getValueAt((resultsjTable.getSelectedRows())[row], 0))
                    .setTypeName(((JMenuItem)evt.getSource()).getText());
            }
            resultsjTable.clearSelection();

        } else if (acceptedTermsjTable.equals(sc)) {
            rejectAcceptedTerms(((JMenuItem)evt.getSource()).getText());

        } else if (rejectedTermsjTable.equals(sc)) {
            changeRejectedTermsType(((JMenuItem)evt.getSource()).getText());

        } else if (undefinedTermsjTable.equals(sc)) {
            rejectUndefinedTerms(((JMenuItem)evt.getSource()).getText());
        }
    }//GEN-LAST:event_rejectTypejMenuItemActionPerformed

    /**
     * Set the status of the selected cells of table to accepted.
     */
    private void acceptTypejMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_acceptTypejMenuItemActionPerformed

        KeyboardFocusManager focusManager =
            KeyboardFocusManager.getCurrentKeyboardFocusManager();

        // selected component before the pop-up menu got focus.
        Component sc = focusManager.getPermanentFocusOwner();

        if (resultsjTable.equals(sc)) {
        // for each selected row in the table
        for (int row = 0; row < resultsjTable.getSelectedRowCount(); row++){
            rowsStatus[(resultsjTable.getSelectedRows())[row]] = ACCEPTED;
            ((Occurrence)resultsjTable.getModel()
                .getValueAt((resultsjTable.getSelectedRows())[row], 0))
                .setTypeName(((JMenuItem)evt.getSource()).getText());
        }
        resultsjTable.clearSelection();

        } else if (acceptedTermsjTable.equals(sc)) {
            changeAcceptedTermsType(((JMenuItem)evt.getSource()).getText());

        } else if (rejectedTermsjTable.equals(sc)) {
            acceptRejectedTerms(((JMenuItem)evt.getSource()).getText());

        } else if (undefinedTermsjTable.equals(sc)) {
            acceptUndefinedTerms(((JMenuItem)evt.getSource()).getText());
        }
    }//GEN-LAST:event_acceptTypejMenuItemActionPerformed

    /**
     * Set the status of the selected cells of table to undefined.
     */
    private void undefinejMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_undefinejMenuItemActionPerformed

        KeyboardFocusManager focusManager =
            KeyboardFocusManager.getCurrentKeyboardFocusManager();

        // selected component before the pop-up menu got focus.
        Component sc = focusManager.getPermanentFocusOwner();

        if (resultsjTable.equals(sc)) {
        // for each selected row in the table
        for (int row = 0; row < resultsjTable.getSelectedRowCount(); row++){
            rowsStatus[(resultsjTable.getSelectedRows())[row]] = UNDEFINED;
            ((Occurrence)resultsjTable.getModel()
                .getValueAt((resultsjTable.getSelectedRows())[row], 0))
                .setTypeName(null);
        }
        resultsjTable.clearSelection();

        } else if (acceptedTermsjTable.equals(sc)) {
            undefineAcceptedTerms();

        } else if (rejectedTermsjTable.equals(sc)) {
            undefineRejectedTerms();
        }
    }//GEN-LAST:event_undefinejMenuItemActionPerformed

    /**
     * Search any string in the input file.
     */
    private void searchStringjMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_searchStringjMenuItemActionPerformed
        searchStringInInputFile();
    }//GEN-LAST:event_searchStringjMenuItemActionPerformed

    /**
     * Search a string provided by the user in the input file and display
     * its occurrences in the context JTextPane.
     */
    private void searchStringInInputFile() {
        
        String lineContent, initialValue;
        BufferedReader br;
        String[] lineContents;
        Document doc;
        int line;
        Occurrence occ;
        String inputFileName;

        try {

        inputFileName = inputFileNameFromInterface();
        if (inputFileName == null) { return; }

        // get the selected term in the results table
        if (resultsjTable.getSelectedRowCount() > 0) {
            initialValue = ((Occurrence)resultsjTable.getModel().getValueAt(
                      resultsjTable.getSelectedRow(), 0)).getKey();
        } else {
            initialValue = "";
        }

        String searchedString = (String)
            JOptionPane.showInputDialog(frame,
                messages.getString("Give_the_string_to_search."),
                messages.getString("search_a_string_in_the_input_file"),
                JOptionPane.QUESTION_MESSAGE,
                null, null, initialValue);

        if (searchedString == null) { return; }

        if (searchedString.trim().equals("")
            || searchedString.length() < 3) {

            JOptionPane.showMessageDialog(frame,
                messages.getString(
                    "The_minimum_allowed_length_of_the_string_is_3."),
                messages.getString("Error_in_input_value"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }

        // open input file in reading
        br = new BufferedReader(new FileReader(inputFileName));

        // replace '\' with '\\' to escape the escape character
        searchedString = searchedString.replaceAll("\\\\", "\\\\\\\\");
        SimpleAttributeSet redAttribute = new SimpleAttributeSet();
        StyleConstants.setForeground(redAttribute, Color.RED);
        doc = contextjTextPane.getDocument();
        contextjTextPane.setText("");
        line = 1;
        lineContent = br.readLine();

        // search line by line the searched string
        while (lineContent != null) {
            lineContents = lineContent.split("\\Q"+searchedString+"\\E");

            // if the string is found in the current line
            if (lineContents.length > 1) {
                doc.insertString(doc.getLength(), "["+line+"] ", null);
                // for each occurrence of the term in the line
                for (int i = 0; i < lineContents.length; i++) {

                    // insert text around the term
                    doc.insertString(doc.getLength(),
                                     lineContents[i], null);

                    // insert searched string
                    if (i < lineContents.length - 1) {
                        doc.insertString(doc.getLength(),
                                         searchedString, redAttribute);
                    }
                }
                // insert an end of line
                doc.insertString(doc.getLength(), "\n", null);
            }
            lineContent = br.readLine();
            line++;
        }

        // close the input file
        br.close();

        if (doc.getLength() == 0) {
            contextjTextPane.setText(messages.getString("No_result."));
        }

        } catch (FileNotFoundException fnfe) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("File_not_found")
                +"\n"+fnfe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return;

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return;

        } catch (javax.swing.text.BadLocationException ble) {
            assert false;
        }
    }
    
    /**
     * Diplay the lines where the Occurrence term occurs into the
     * contextjTextPane. If Occurrence is in the line, color it in red.
     *
     * @param occ term to diplay in context
     * @param contextjTextPane element where to display the lines of context
     */
    private void displayContext(Occurrence occ,
                                JTextPane contextjTextPane) {

        int line, lineOcc;
        Iterator it;
        String lineContent, term, regex;
        String[] lineContents;
        String inputFileName, separator;
        BufferedReader br;
        Document doc;
        int termCol;
        Matcher matcher;

        doc = contextjTextPane.getDocument();
        contextjTextPane.setText("");
        // separator between word and tag
        separator = separatorjTextField.getText();

        if (occ == null) { return; }

        try {

        if (separator.trim().equals("")) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("This_parameter_is_not_provided.")
                +"\n\n"+messages.getString("Separator"),
                messages.getString("Bad_parameters"),
                JOptionPane.ERROR_MESSAGE);
            mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
            separatorjTextField.requestFocusInWindow();
            return;
        }

        if (!occ.hasLines()) {
            doc.insertString(doc.getLength(), messages.getString(
            "This_occurrence_doesn't_appear_in_the_input_text."),
            null);
            return;
        }

        // open input file in reading
        inputFileName = inputFileNameFromInterface();
        if (inputFileName == null) { return; }
        br = new BufferedReader(new FileReader(inputFileName));

        SimpleAttributeSet redAttribute = new SimpleAttributeSet();
        StyleConstants.setForeground(redAttribute, Color.RED);
        line = 0;
        term = occ.getKey();
        // replace '\' with '\\' to escape the escape character
        term = term.replaceAll("\\\\", "\\\\\\\\");

        // for each line (in ascending order) where term occurs
        it = occ.getLinesIterator();
        while(it.hasNext()) {
            lineOcc = ((Integer)it.next()).intValue();
            // skip the lines that don't contain the term
            while (line < lineOcc-1) {
                br.readLine();
                line++;
            }

            lineContent = br.readLine();
            line++;

            // get the text around the term in lineContents array
            lineContents = lineContent.split("\\Q"+term+"\\E");

            // if units are not found because they have been grouped
            if (lineContents.length == 1) {
                // construct a regex for finding the grouped units
                regex = Occurrence.getWord(occ.getUnit1(), separator);
                regex += "-";
                regex += Occurrence.getWord(occ.getUnit2(), separator);
                if (occ.is3Units()) {
                    regex += "-";
                    regex += Occurrence.getWord(occ.getUnit3(), separator);
                }
                // replace '\' with '\\' to escape the escape character
                regex = regex.replaceAll("\\\\", "\\\\\\\\");
                regex = "\\Q"+regex+"\\S+"+"\\E";
                lineContents = lineContent.split(regex);
                // get the term
                matcher = Pattern.compile(regex).matcher(lineContent);
                if (matcher.find()) {
                    term = matcher.group();
                } else {
                    // no match found
                    term ="";
                }
            }

            // insert number of line
            doc.insertString(doc.getLength(), "["+line+"] ", null);

            // for each occurrence of the term in the line
            for (int i = 0; i < lineContents.length; i++) {

                // insert text around the term
                doc.insertString(doc.getLength(),
                                 lineContents[i], null);

                // insert term
                if (i < lineContents.length - 1) {
                    doc.insertString(doc.getLength(),
                                     term, redAttribute);
                }
            }
            // insert an end of line
            doc.insertString(doc.getLength(), "\n", null);
        }

        // close the input file
        br.close();

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_context"),
                JOptionPane.ERROR_MESSAGE);
            return;

        } catch (BadLocationException ble) {
            assert false : ble.getLocalizedMessage();
            return;
        }
    }

    /**
     * Listener for key events on the results table.
     * <p>
     * [Space] key display the lines where the selected term occurs.
     * <br>[num pad 1 to 9] key accept the selected terms.
     * <br>[ctrl + num pad 1 to 9] key reject the selected terms.
     * <br>[Delete] key undefined the selected terms
     * <br>[S] key search a string in the input file
     */
    private void resultsjTableKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_resultsjTableKeyReleased

        Occurrence occ;

        if (resultsjTable.getRowCount() < 1) { return; }

        // display context
        if (evt.getKeyCode() == java.awt.event.KeyEvent.VK_SPACE
            && resultsjTable.getSelectedRowCount() > 0) {

            // get the selected term Occurrence
            occ = (Occurrence)resultsjTable.getModel().getValueAt(
                      resultsjTable.getSelectedRow(), 0);

            displayContext(occ, contextjTextPane);

        // accept selected terms
        } else if (!evt.isControlDown()
            && evt.getKeyCode() >= java.awt.event.KeyEvent.VK_NUMPAD1
            && evt.getKeyCode() <= java.awt.event.KeyEvent.VK_NUMPAD9
            && (evt.getKeyCode() - java.awt.event.KeyEvent.VK_NUMPAD0)
                < acceptedTermsTypesjComboBox.getItemCount()
            && resultsjTable.getSelectedRowCount() > 0) {

            // for each selected row in the table
            for (int row = 0; row < resultsjTable.getSelectedRowCount();
                 row++){
                rowsStatus[(resultsjTable.getSelectedRows())[row]] =
                    ACCEPTED;
                // set the type of the selected term
                ((Occurrence)resultsjTable.getModel()
                    .getValueAt((resultsjTable.getSelectedRows())[row], 0))
                    .setTypeName(((JMenuItem)acceptjMenu.getItem(
                       evt.getKeyCode()-java.awt.event.KeyEvent.VK_NUMPAD1))
                        .getText());
            }
            resultsjTable.clearSelection();

        // reject selected terms
        } else if (evt.isControlDown()
            && evt.getKeyCode() >= java.awt.event.KeyEvent.VK_NUMPAD1
            && evt.getKeyCode() <= java.awt.event.KeyEvent.VK_NUMPAD9
            && (evt.getKeyCode() - java.awt.event.KeyEvent.VK_NUMPAD0)
                < rejectedTermsTypesjComboBox.getItemCount()
            && resultsjTable.getSelectedRowCount() > 0) {

            // for each selected row in the table
            for (int row = 0; row < resultsjTable.getSelectedRowCount();
                 row++){
                rowsStatus[(resultsjTable.getSelectedRows())[row]] =
                    REJECTED;
                // set the type of the selected term
                ((Occurrence)resultsjTable.getModel()
                    .getValueAt((resultsjTable.getSelectedRows())[row], 0))
                    .setTypeName(((JMenuItem)rejectjMenu.getItem(
                       evt.getKeyCode()-java.awt.event.KeyEvent.VK_NUMPAD1))
                        .getText());
            }
            resultsjTable.clearSelection();

        // undefined selected terms
        } else if (evt.getKeyCode() == java.awt.event.KeyEvent.VK_DELETE
            && resultsjTable.getSelectedRowCount() > 0) {

            // for each selected row in the table
            for (int row = 0; row < resultsjTable.getSelectedRowCount();
                 row++){
                rowsStatus[(resultsjTable.getSelectedRows())[row]] =
                    UNDEFINED;
                ((Occurrence)resultsjTable.getModel()
                    .getValueAt((resultsjTable.getSelectedRows())[row], 0))
                    .setTypeName(null);
            }
            resultsjTable.clearSelection();

        } else if (evt.getKeyCode() == java.awt.event.KeyEvent.VK_S) {
            searchStringInInputFile();
        }
    }//GEN-LAST:event_resultsjTableKeyReleased

    /**
     * Right mouse click shows a contextual menu for accepting or rejecting
     * the selected terms. Middle mouse click display the lines where the
     * selected term occurs.
     */
    private void resultsjTableMouseReleased(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_resultsjTableMouseReleased

        Occurrence occ;

        // display context
        if (evt.getButton() == java.awt.event.MouseEvent.BUTTON2
            && resultsjTable.getSelectedRowCount() > 0) {

            // get the selected term Occurrence
            occ = (Occurrence)resultsjTable.getModel().getValueAt(
                      resultsjTable.getSelectedRow(), 0);

            displayContext(occ, contextjTextPane);

        // if a selection is active and mouse button 3 is pressed shows the
        // popup menu
        } else if (evt.getButton() == java.awt.event.MouseEvent.BUTTON3
            && resultsjTable.getSelectedRowCount() > 0) {

            undefinejMenuItem.setVisible(true);
            searchStringjMenuItem.setVisible(true);
            resultsjPopupMenu.show(resultsjTable, evt.getX(), evt.getY());
        }
    }//GEN-LAST:event_resultsjTableMouseReleased

    private void outputFileChooserjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_outputFileChooserjButtonActionPerformed
        
        JFileChooser fileChooser;
        String file;
        File currentDir;

        currentDir = new File(System.getProperty("user.dir"));
        fileChooser = new JFileChooser(currentDir);
        fileChooser.setDialogTitle(
            messages.getString("save_the_results_in_file"));
        fileChooser.setFileFilter(
            new FileChooserFilterExtension(".results",
                messages.getString("Results_files")));
        fileChooser.setDialogType(fileChooser.SAVE_DIALOG);
        int result = fileChooser.showOpenDialog(frame);

        try {
            if (result == JFileChooser.APPROVE_OPTION) {
                file = fileChooser.getSelectedFile().getCanonicalPath();
                if (file.startsWith(currentDir.getCanonicalPath())) {
                    file = file.substring(currentDir.toString().length()+1);
                }
                outputFileComboBox.setSelectedItem(file);
            }
        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }

    }//GEN-LAST:event_outputFileChooserjButtonActionPerformed
        
    private void optionsComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_optionsComboBoxActionPerformed

        // limit the automatic loading of options to user actions
        if (evt.getActionCommand().equals("Programmatically")) { return; }

        if (optionsComboBox.getSelectedItem() == null
        || ((String)optionsComboBox.getSelectedItem())
            .trim().length() == 0) {
            return;
        }

        loadOptions();
    }//GEN-LAST:event_optionsComboBoxActionPerformed

    /**
     * Start a new extraction from the beginning. Clear all tables of terms.
     */
    private void startAgainjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_startAgainjButtonActionPerformed

        String newFileName;

        // show a warning only if at least one extraction have been done
        if (iteration > 0) {
            int result = JOptionPane.showConfirmDialog(frame,
                messages.getString("All_data_will_be_reset_!"),
                messages.getString(
                    "Start_the_extraction_again_form_the_beginning."),
                JOptionPane.OK_CANCEL_OPTION,
                JOptionPane.QUESTION_MESSAGE);

            if (result == JOptionPane.CANCEL_OPTION) { return; }
        }

        iteration = 0;

        // clear all memory tables because new input file is chosen
        acceptedTerms.clear();
        currentAcceptedTerms.clear();
        rejectedTerms.clear();
        undefinedTerms.clear();
        currentAcceptedTermsTreeSet.clear();
        currentSkippedTermsTreeSet.clear();

        // clear all selections in the results table
        resultsjTable.clearSelection();

        // clear the status of the results table cells
        Arrays.fill(rowsStatus, UNDEFINED);

        // enable all cells in the Expressions panel tables
        enableExpressions();

        // clear tables and lists content
        resultsjTable.setModel(new DefaultTableModel());
        acceptedTermsjTable.setModel(new UneditableDefaultTableModel());
        rejectedTermsjTable.setModel(new UneditableDefaultTableModel());
        undefinedTermsjTable.setModel(new UneditableDefaultTableModel());

        // clear terms filter lists
        acceptedTermsRelationjComboBox.removeAllItems();
        rejectedTermsRelationjComboBox.removeAllItems();
        undefinedTermsRelationjComboBox.removeAllItems();
        acceptedTermsIterationjComboBox.removeAllItems();
        rejectedTermsIterationjComboBox.removeAllItems();
        undefinedTermsIterationjComboBox.removeAllItems();
        acceptedTermsTypesjComboBox.removeAllItems();
        rejectedTermsTypesjComboBox.removeAllItems();

        // clear selected output file
        outputFileComboBox.setSelectedItem(null);

        // clear terms type list in rules panel
        lrTypejComboBox.removeAllItems();

        // enable again the elements of the interface
        optionsDirjComboBox.setEnabled(true);
        optionsComboBox.setEnabled(true);
        mainjTabbedPane.setEnabledAt(1, true);

        updateFilesList();
    }//GEN-LAST:event_startAgainjButtonActionPerformed

    /**
     * Call the precedent terms Dialog.
     */
    private void precedentTermsjButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_precedentTermsjButtonActionPerformed

        if (resultsjTable.getRowCount() > 0) {
            precedentTermsjDialog.show();
        } else {
            JOptionPane.showMessageDialog(frame,
                messages.getString("No_result."),
                "",
                JOptionPane.ERROR_MESSAGE);
        }
    }//GEN-LAST:event_precedentTermsjButtonActionPerformed

    /**
     * Create a new ExtractTerms from parameters of the user interface.
     * Test all parameters and display error dialog if needed.
     *
     * @return The new ExtractTerms object ; null if an error occurs.
     */
    private ExtractTerms newExtractTerms() {

        TableModel tm;
        String inputFileName;
        int pruning2Units, pruning3Units;
        int formula;
        boolean tfidf;
        boolean occFirst;
        String regexUnit;
        String regexNameOfText;
        String regexSentPhraSeparator;
        String regex2Units1, regex2Units2;
        int tagKept2Units;
        String regex3Units1, regex3Units2, regex3Units3;
        int tagKept3Units;
        int maxResults2Units, maxResults3Units;
        String separator;
        int relation2Units;
        int relation3Units;
        boolean acceptedType;
        boolean rejectedType;

        // initialise objects
        inputFileName = "";
        formula = -1;
        regexUnit = "";
        regexSentPhraSeparator = "";
        regexNameOfText = "";
        regex2Units1 = "";
        regex2Units2 = "";
        tagKept2Units = -1;
        regex3Units1 = "";
        regex3Units2 = "";
        regex3Units3 = "";
        tagKept3Units = -1;
        maxResults2Units = -1;
        maxResults3Units = -1;
        separator = "";
        acceptedType = false;
        rejectedType = false;

        if (twoUnitsComboBox.getItemCount() > 0
        && twoUnitsComboBox.getForeground() != Color.RED
        && threeUnitsComboBox.getForeground() != Color.RED) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("Please_select_a_relation_to_extract."),
                messages.getString("Bad_parameters"),
                JOptionPane.ERROR_MESSAGE);
            twoUnitsComboBox.requestFocusInWindow();
            return null;
        }

        // relation to extract
        relation2Units = (twoUnitsComboBox.getForeground() == Color.RED) ?
            twoUnitsComboBox.getSelectedIndex() : -1;
        relation3Units = (threeUnitsComboBox.getForeground() == Color.RED) ?
            threeUnitsComboBox.getSelectedIndex() : -1;

        tm = regexjTable.getModel();

        // unit definition regex
        for (int row = 0; row < 10; row++) {
            if (tm.getValueAt(row, 0) != null
                && tm.getValueAt(row, 1) != null
                && unitComboBox.getSelectedItem() != null
                && ((String)unitComboBox.getSelectedItem())
                .equals((String)tm.getValueAt(row, 0))) {
                regexUnit = (String)tm.getValueAt(row, 1);
            }
        }

        // sentence/phrase separators
        for (int row = 0; row < 10; row++) {
            if (tm.getValueAt(row, 2) != null
                && tm.getValueAt(row, 3) != null
                && sentPhrComboBox.getSelectedItem() != null
                && ((String)sentPhrComboBox.getSelectedItem())
                .equals((String)tm.getValueAt(row, 2))) {
                regexSentPhraSeparator = (String)tm.getValueAt(row, 3);
            }
        }

        // name of text selector
        for (int row = 0; row < 10; row++) {
            if (tm.getValueAt(row, 4) != null
                && tm.getValueAt(row, 5) != null
                && nameOfTextjComboBox.getSelectedItem() != null
                && ((String)nameOfTextjComboBox.getSelectedItem())
                .equals((String)tm.getValueAt(row, 4))) {
                regexNameOfText = (String)tm.getValueAt(row, 5);
            }
        }

        tm = regex2jTable.getModel();

        // units pair selector
        for (int row = 0; row < 100; row++) {
            if (tm.getValueAt(row, 0) != null
                && tm.getValueAt(row, 1) != null
                && tm.getValueAt(row, 2) != null
                && tm.getValueAt(row, 3) != null
                && twoUnitsComboBox.getSelectedItem() != null
                && ((String)twoUnitsComboBox.getSelectedItem())
                .equals((String)tm.getValueAt(row, 0))) {
                regex2Units1 = (String)tm.getValueAt(row, 1);
                regex2Units2 = (String)tm.getValueAt(row, 2);
                tagKept2Units = ((Integer)tm.getValueAt(row, 3)).intValue();
            }
        }

        // units triple selector
        for (int row = 0; row < 100; row++) {
            if (tm.getValueAt(row, 4) != null
                && tm.getValueAt(row, 5) != null
                && tm.getValueAt(row, 6) != null
                && tm.getValueAt(row, 7) != null
                && tm.getValueAt(row, 8) != null
                && threeUnitsComboBox.getSelectedItem() != null
                && ((String)threeUnitsComboBox.getSelectedItem())
                .equals((String)tm.getValueAt(row, 4))) {
                regex3Units1 = (String)tm.getValueAt(row, 5);
                regex3Units2 = (String)tm.getValueAt(row, 6);
                regex3Units3 = (String)tm.getValueAt(row, 7);
                tagKept3Units = ((Integer)tm.getValueAt(row, 8)).intValue();
            }
        }

        // formula expression
        tm = formulasjTable.getModel();
        for (int row = 0; row < 10; row++) {
            if (tm.getValueAt(row, 0) != null
                && tm.getValueAt(row, 1) != null
                && measuresjComboBox.getSelectedItem() != null
                && ((String)measuresjComboBox.getSelectedItem())
                .equals((String)tm.getValueAt(row, 0))) {
                formula = row;
            }
        }

        // accepted and rejected terms types
        tm = termsTypesjTable.getModel();
        for (int row = 0; row < 50; row++) {
            if (tm.getValueAt(row, 0) != null) { acceptedType = true; }
            if (tm.getValueAt(row, 1) != null) { rejectedType = true; }
        }

        // tf*idf
        tfidf = tfidfjCheckBox.isSelected();

        // occurrence first
        occFirst = occFirstjCheckBox.isSelected();

        // separator between word and tag
        separator = separatorjTextField.getText().trim();

        // no test needed for these elements because values are constrained
        pruning2Units = ((Integer)twoUnitsSpinner.getValue()).intValue();
        pruning3Units = ((Integer)threeUnitsSpinner.getValue()).intValue();
        maxResults2Units =
            ((Integer)twoUnitsResultsSpinner.getValue()).intValue();
        maxResults3Units =
            ((Integer)threeUnitsResultsSpinner.getValue()).intValue();

        inputFileName = inputFileNameFromInterface();
        if (inputFileName == null) { return null; }

        // verify that each regex is provided
        // verify that each parameters is provided
        // display an error message if needed and select the element of
        // interface that is the last cause of the  error
        String message = "";
        if (regexUnit.equals("")) {
            message = messages.getString("Unit_definition");
            if (unitComboBox.getItemCount() == 0
                || unitComboBox.getSelectedItem() != null) {
                mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
                regexjTable.requestFocusInWindow();
                regexjTable.editCellAt(0, 0);
            } else {
                mainjTabbedPane.setSelectedIndex(2);
                unitComboBox.requestFocusInWindow();
            }
        } else if (regexSentPhraSeparator.equals("")) {
            message = messages
               .getString("Sentence_phrase_separators_definition");
            if (sentPhrComboBox.getItemCount() == 0
                || sentPhrComboBox.getSelectedItem() != null) {
                mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
                regexjTable.requestFocusInWindow();
                regexjTable.editCellAt(0, 2);
            } else {
                mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
                sentPhrComboBox.requestFocusInWindow();
            }
        } else if (regexNameOfText.equals("")) {
            message = messages.getString("Name_of_text_selector");
            if (nameOfTextjComboBox.getItemCount() == 0
                || nameOfTextjComboBox.getSelectedItem() != null) {
                mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
                regexjTable.requestFocusInWindow();
                regexjTable.editCellAt(0, 4);
            } else {
                mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
                nameOfTextjComboBox.requestFocusInWindow();
            }
        } else if (regex2Units1.equals("")) {
            message = messages
                .getString("Units_pair_selector");
            if (twoUnitsComboBox.getItemCount() == 0
                || twoUnitsComboBox.getSelectedItem() != null) {
                mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
                regex2jTable.requestFocusInWindow();
                regex2jTable.editCellAt(0, 1);
            } else {
                mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
                twoUnitsComboBox.requestFocusInWindow();
            }
        } else if (tagKept2Units == -1) {
            message = messages.getString("Tag_kept")+" ";
            message += messages.getString("Units_pair");
                mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
            regex2jTable.requestFocusInWindow();
            regex2jTable.editCellAt(0, 3);
        } else if (regex3Units1.equals("")) {
            message = messages
                .getString("Units_triple_selector");
            if (threeUnitsComboBox.getItemCount() == 0
                || threeUnitsComboBox.getSelectedItem() != null) {
                mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
                regex2jTable.requestFocusInWindow();
                regex2jTable.editCellAt(0, 5);
            } else {
                mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
                threeUnitsComboBox.requestFocusInWindow();
            }
        } else if (tagKept3Units == -1) {
            message = messages.getString("Tag_kept")+" ";
            message += messages.getString("Units_triple");
                mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
            regex2jTable.requestFocusInWindow();
            regex2jTable.editCellAt(0, 8);
        } else  if (formula == -1) {
            message = messages.getString("Relevancy_measure");
            if (measuresjComboBox.getItemCount() == 0
                || measuresjComboBox.getSelectedItem() != null) {
                mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
                formulasjTable.requestFocusInWindow();
                formulasjTable.editCellAt(0, 1);
            } else {
                mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
                measuresjComboBox.requestFocusInWindow();
            }
        } else if (!acceptedType) {
            message += messages.getString("Accepted_types");
            mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
            termsTypesjTable.requestFocusInWindow();
            termsTypesjTable.editCellAt(0, 0);
        } else if (!rejectedType) {
            message +=messages.getString("Rejected_types") ;
            mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
            termsTypesjTable.requestFocusInWindow();
            termsTypesjTable.editCellAt(0, 1);
        } else if (separator.length() != 1) {
            message += messages.getString("Separator");
            mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
            separatorjTextField.requestFocusInWindow();
        }

        if (!message.equals("")) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("This_parameter_is_not_provided.")
                +"\n\n"+message,
                messages.getString("Bad_parameters"),
                JOptionPane.ERROR_MESSAGE);
            return null;
        }

        // test the validity of regular expressions
        try {
            Pattern.compile(regexUnit);
            Pattern.compile(regexNameOfText);
            Pattern.compile(regexSentPhraSeparator);
            Pattern.compile(regex2Units1);
            Pattern.compile(regex2Units2);
            Pattern.compile(regex3Units1);
            Pattern.compile(regex3Units2);
            Pattern.compile(regex3Units3);

        } catch (PatternSyntaxException pse) {
                mainjTabbedPane.setSelectedComponent(regexFormulasjPanel);
            JOptionPane.showMessageDialog(frame,
                pse.getLocalizedMessage()+"\n"
                +messages.getString("Bad_syntax_in_regular_expression."),
                messages.getString("Bad_parameters"),
                JOptionPane.ERROR_MESSAGE);
            return null;
        }

        // create new ExtractTerms object
        return new ExtractTerms(inputFileName,
                                regexUnit,
                                regexNameOfText,
                                regexSentPhraSeparator,
                                regex2Units1, regex2Units2,
                                tagKept2Units,
                                regex3Units1, regex3Units2, regex3Units3,
                                tagKept3Units,
                                formula,
                                tfidf,
                                pruning2Units,
                                pruning3Units,
                                maxResults2Units,
                                maxResults3Units,
                                separator,
                                relation2Units,
                                relation3Units,
                                occFirst);

    }

    /**
     * Call replace then extract methods from ExtractTerms class and write
     * results in the results table. Save the options, memory tables,
     * iterated input file and extraction history of the precedent
     * extraction.
     */
    private void extractButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_extractButtonActionPerformed

        TableModel tm;
        String result, fileName, dirName, line;
        int numFile;
        Iterator it, it2;
        DefaultListModel lm;
        Occurrence occ;
        String newFileName, actionCommand;
        DefaultComboBoxModel cbm2, cbm3;
        Object o;
        boolean allValuesAreSet;
        RandomAccessFile raf, raf2;
        FileChannel srcChannel, dstChannel;
        BufferedReader br;
        String [] history;
        int numLine;


        if (iteration >= 99) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("It's_your_100_iteration."),
                "", JOptionPane.ERROR_MESSAGE);
            return;
        }

        et = newExtractTerms();

        // if there is an error in the provided extraction parameters
        if (et == null) { return; }

        // if this is the first iteration
        // load the reject and accept list if provided
        // load already formed terms list if provided
        // update terms comboboxes
        if (iteration == 0) {

            statejLabel.setText(messages.getString("Preprocessing_is_on."));
            statejLabel.update((statejLabel).getGraphics());

            // forbid the change of the extraction parameters
            optionsDirjComboBox.setEnabled(false);
            optionsComboBox.setEnabled(false);
            mainjTabbedPane.setEnabledAt(1, false);

            // add the accept list terms to the accepted terms
            if (acceptListjComboBox.getSelectedItem() != null
              && !((String)acceptListjComboBox.getSelectedItem())
                .trim().equals("")
              && !addTermFileToMemoryTables(acceptListjComboBox, ACCEPTED)){
                return; // an error has occured
            }

            // add the reject list terms to the rejected terms
            if (rejectListjComboBox.getSelectedItem() != null
              && !((String)rejectListjComboBox.getSelectedItem())
                .trim().equals("")) {

                if (((String)rejectListjComboBox.getSelectedItem()).equals(
                     (String)acceptListjComboBox.getSelectedItem())) {
                    mainjTabbedPane
                        .setSelectedComponent(preProcessingjPanel);
                    JOptionPane.showMessageDialog(frame,
                        messages.getString("Please_choose_two_different_"
                       +"files_for_the_acceptation_and_reject_lists."),
                        "", JOptionPane.ERROR_MESSAGE);
                    rejectListjComboBox.requestFocusInWindow();
                    return;

                } else if (!addTermFileToMemoryTables(rejectListjComboBox,
                                                      REJECTED)) {
                    return; // an error has occured
                }
            }

            // add the already formed terms to accepted terms list
            lm = (DefaultListModel)alreadyFormedTermsjList.getModel();
            for (int i = 0; i < lm.size(); i++) {
                occ = (Occurrence)lm.getElementAt(i);
                acceptedTerms.add(occ);
                currentAcceptedTerms.add(occ);
                currentAcceptedTermsTreeSet.add(occ.getKey());
                currentSkippedTermsTreeSet.add(occ.getKey());
            }

            addLinesToOccurrence(acceptedTerms);
            addLinesToOccurrence(rejectedTerms);

            // set the iteration for all terms
            it = acceptedTerms.iterator();
            while (it.hasNext()) {
                ((Occurrence)it.next()).setIteration(iteration);
            }
            it = rejectedTerms.iterator();
            while (it.hasNext()) {
                ((Occurrence)it.next()).setIteration(iteration);
            }

            // update terms iteration lists
            acceptedTermsIterationjComboBox.addItem(
                messages.getString("All"));
            rejectedTermsIterationjComboBox.addItem(
                messages.getString("All"));
            undefinedTermsIterationjComboBox.addItem(
                messages.getString("All"));
            acceptedTermsIterationjComboBox.addItem(new Integer(iteration));
            rejectedTermsIterationjComboBox.addItem(new Integer(iteration));
           undefinedTermsIterationjComboBox.addItem(new Integer(iteration));

            // increment the number of the extraction iteration
            iteration++;
        }

        updateMemoryTermsTables();

        fileName = "";
        dirName = "";
        numFile = 1;

        try {
            // get the last number for the file naming
            dirName = (String)optionsDirjComboBox.getSelectedItem();
            fileName = dirName+System.getProperty("file.separator")+".info";
            raf = new RandomAccessFile(fileName, "rw");
            // read the first line
            numFile = raf.readInt() + 1;
            // erase the content of the file
            raf.setLength(0);
            // replace with the new number for the naming
            raf.writeInt(numFile);
            raf.close();

            // copy precedent history file in the new history file
            fileName = dirName+System.getProperty("file.separator")
                +(String)optionsComboBox.getSelectedItem()+".hist";
            srcChannel = new FileInputStream(fileName).getChannel();
            fileName = dirName+System.getProperty("file.separator")
                +numFile+".hist";
            dstChannel = new FileOutputStream(fileName).getChannel();
            dstChannel.transferFrom(srcChannel, 0, srcChannel.size());
            srcChannel.close(); dstChannel.close();

            // save new extraction history line at the end of the file
            raf = new RandomAccessFile(fileName, "rw");
            // locate the pointer at the end of the file
            raf.seek(raf.length());
            if (twoUnitsComboBox.getForeground() == Color.RED) {
                raf.writeBytes((String)twoUnitsComboBox.getSelectedItem()
                              +System.getProperty("line.separator"));
            } else {
                raf.writeBytes((String)threeUnitsComboBox.getSelectedItem()
                              +System.getProperty("line.separator"));
            }
            raf.close();

        } catch  (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage()
                +" ("+fileName+")",
                messages.getString("Error_in_writing_file"),
                JOptionPane.ERROR_MESSAGE);
        }

        // change the extension to iter
        newFileName =
            dirName+System.getProperty("file.separator")+numFile+".iter";

        // group accepted terms of the last iteration
        if (currentAcceptedTerms.size() > 0) {

            statejLabel.setText(messages.getString("Grouping_is_on..."));
            statejLabel.update((statejLabel).getGraphics());

            // group accepted terms
            et.replace(currentAcceptedTerms, "-", et.getInputFileName(),
                       newFileName, et.GROUP_MODE);

            updateFilesList();

        } else {
            try{
                // just copy .iter file for the new iteration
                srcChannel =
                    new FileInputStream(et.getInputFileName()).getChannel();
                dstChannel = new FileOutputStream(newFileName).getChannel();
                dstChannel.transferFrom(srcChannel, 0, srcChannel.size());
                srcChannel.close(); dstChannel.close();

            } catch (IOException ioe) {
                JOptionPane.showMessageDialog(frame,
                    ioe.getLocalizedMessage(),
                    messages.getString("Error_in_reading_or_writing_file."),
                    JOptionPane.ERROR_MESSAGE);
            }
        }

        // no more need of accepted terms because they are now grouped
        // in the input file
        currentSkippedTermsTreeSet
            .removeAll(currentAcceptedTermsTreeSet);
        currentAcceptedTermsTreeSet.clear();
        currentAcceptedTerms.clear();

        // temporary disable triggering of events on optionsComboBox
        actionCommand = optionsComboBox.getActionCommand();
        optionsComboBox.setActionCommand("Programmatically");
        addItemUnique(optionsComboBox, String.valueOf(numFile));
        optionsComboBox.setSelectedItem(String.valueOf(numFile));
        // events can already occur on optionsComboBox
        optionsComboBox.setActionCommand(actionCommand);
        et.setInputFileName(newFileName);

        // save options in a file
        saveOptions();

        updateTermsTables();

        statejLabel.setText(messages.getString("Extraction_is_on."));
        statejLabel.update((statejLabel).getGraphics());

        try {
            // extract terms
            if (inputFileLines == 0) {
                inputFileLines = countLines(newFileName);
            }
            extractionjProgressBar.setVisible(true);
            extractionjProgressBar.setMaximum(inputFileLines);

            if (!et.extract(currentSkippedTermsTreeSet, iteration,
                            extractionjProgressBar)) {
                extractionjProgressBar.setVisible(false);
                return; // an error has occur in et.extract() method

            } else {
                extractionjProgressBar.setVisible(false);
                statejLabel.setText(
                    messages.getString("Extraction_is_over."));
            }

        } catch (java.lang.OutOfMemoryError ome) {
            JOptionPane.showMessageDialog(frame, messages.getString(
                "Java_virtual_machine_doesn't_have_enough_memory."), "",
                JOptionPane.ERROR_MESSAGE);
            extractionjProgressBar.setVisible(false);
        }

        int rowCount = 5+5+et.getResults2Units().size()+2
            +5+et.getResults3Units().size()+100;

        // clear all selections in the results table
        resultsjTable.clearSelection();

        // clear the status of the results table cells
        rowsStatus = new int[rowCount];
        Arrays.fill(rowsStatus, UNDEFINED);

        // disable filled cells in the Expressions panel tables
        disableExpressions();

        // create a new TableModel
        Object[] columnNames = {messages.getString("Results")};
        resultsjTable.setModel(new DefaultTableModel(columnNames, rowCount)
            {
                public boolean isCellEditable(int rowIndex, int columnIndex)
                { return false; }
            });
        resultsjTable.setDefaultRenderer(resultsjTable.getColumnClass(0),
                                         new ResultsCellRenderer());

        // write the results in the table
        tm = resultsjTable.getModel();
        int row = 0;
        tm.setValueAt(
            messages.getString("Iteration")+" = "+iteration, row++, 0);
        tm.setValueAt(messages.getString("Relevancy_measure")+": "
            +(String)measuresjComboBox.getSelectedItem()
            +((tfidfjCheckBox.isSelected())?" (tf*idf)":"")
            +((occFirstjCheckBox.isSelected())?" ("
                +messages.getString("Occurrence_first")+")":""), row++, 0);
        tm.setValueAt(messages.getString("Unique_units")+" = "
            +et.getTotalUnitsWithoutDouble(), row++, 0);
        tm.setValueAt(messages.getString("Number_of_documents")+" = "
            +et.getTotalDocuments(), row++, 0);
        tm.setValueAt(null, row++, 0);

        // results for units pair
        if (twoUnitsComboBox.getForeground() != Color.RED) {
            // do nothing
        } else if (et.getTotalResults2Units() == 0) {
            tm.setValueAt(messages.getString("No_results_for_units_pair."),
                          row++, 0);
        } else {
        tm.setValueAt(messages.getString(
            "Relevancy_results_for_units_pair")+": "
            +(String)twoUnitsComboBox.getSelectedItem(), row++, 0);
        tm.setValueAt(
          messages.getString("Units_pair/unique/redondancy")+" = "
          +et.getTotalUnitsPair()+"/"
          +et.getTotalUnitsPairWithoutDouble()+"/"
          +(et.getTotalUnitsPair()-et.getTotalUnitsPairWithoutDouble())*100
          /et.getTotalUnitsPair()+"%", row++, 0);
        tm.setValueAt(
           messages.getString(
           "Units_pair_after_pruning/pourcent/threshold")+" = "
           +et.getTotalResults2Units()+"/"
           +(et.getTotalUnitsPairWithoutDouble()
             -et.getTotalResults2Units())*100
           /et.getTotalUnitsPairWithoutDouble()
           +"%/"+et.getPruning2Units(), row++, 0);
        tm.setValueAt(null, row++, 0);
        tm.setValueAt(messages.getString(
        "measure_unit1_unit2_(nb_of_occurrence,_nb_of_text_where_present)"),
                      row++, 0);
        it = et.getResults2Units().iterator();
        while (it.hasNext()) {
            tm.setValueAt(it.next(), row++, 0);
        }
        }

        // results for units triple
        if (threeUnitsComboBox.getForeground() != Color.RED) {
            // do nothing
        } else if (et.getTotalResults3Units() == 0) {
            tm.setValueAt(messages.getString(
                "No_results_for_units_triple."), row++, 0);
        } else {

        tm.setValueAt(messages.getString(
            "Relevancy_results_for_units_triple")+": "
            +(String)threeUnitsComboBox.getSelectedItem(), row++, 0);
        tm.setValueAt(
          messages.getString("Units_triple/unique/redondancy")+" = "
          +et.getTotalUnitsTriple()+"/"
          +et.getTotalUnitsTripleWithoutDouble()+"/"
          +(et.getTotalUnitsTriple()
            -et.getTotalUnitsTripleWithoutDouble())*100
          /et.getTotalUnitsTriple()+"%", row++, 0);
        tm.setValueAt(messages.getString(
            "Units_triple_after_pruning/pourcent/threshold")+" = "
          +et.getTotalResults3Units()+"/"
          +(et.getTotalUnitsTripleWithoutDouble()
            -et.getTotalResults3Units())*100
          /et.getTotalUnitsTripleWithoutDouble()
                      +"%/"+et.getPruning3Units(), row++, 0);
        tm.setValueAt(null, row++, 0);
        tm.setValueAt(messages.getString(
           "measure_unit1_unit2_unit3_(nb_of_occurrence,_nb_of_text_where"
           +"_present)"), row++, 0);
        it = et.getResults3Units().iterator();
        while (it.hasNext()) {
            tm.setValueAt(it.next(), row++, 0);
        }
        }

        // add history of the extracted relations to the end of results
        history = new String [99];
        line = ""; numLine = 0;
        fileName = (String)optionsDirjComboBox.getSelectedItem()
            +System.getProperty("file.separator")
            +(String)optionsComboBox.getSelectedItem()+".hist";

        try {
            br = new BufferedReader(new FileReader(fileName));
            // if the .hist file is empty
            if (br.ready()) {
                // read all lines of the .hist file
                line = br.readLine();
                while (line != null && numLine < history.length) {
                    history[numLine++] = line;
                    line = br.readLine();
                }
            }
            tm.setValueAt(null, row++, 0);
            tm.setValueAt(messages.getString(
                "History_of_previously_extracted_relations."), row++, 0);
            for (int i = 0; i < numLine; i++) {
                tm.setValueAt(history[i], row++, 0);
            }

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }

        if (et.getTotalResults2Units() == 0
            && ((Integer)twoUnitsSpinner.getValue()).intValue() != 1
            && twoUnitsComboBox.getForeground() == Color.RED) {
            JOptionPane.showMessageDialog(frame,
                messages.getString(
                "You_should_decrease_the_pruning_threshold."),
                "", JOptionPane.INFORMATION_MESSAGE);
            twoUnitsSpinner.requestFocusInWindow();
            return;
        }

        if (et.getTotalResults3Units() == 0
            && ((Integer)threeUnitsSpinner.getValue()).intValue() != 1
            && threeUnitsComboBox.getForeground() == Color.RED) {
            JOptionPane.showMessageDialog(frame,
                messages.getString(
                "You_should_decrease_the_pruning_threshold."),
                "", JOptionPane.INFORMATION_MESSAGE);
            threeUnitsSpinner.requestFocusInWindow();
            return;
        }

        if (twoUnitsComboBox.getForeground() == Color.RED
            && et.getTotalResults2Units() == 0
            && ((Integer)twoUnitsSpinner.getValue()).intValue() == 1) {

            JOptionPane.showMessageDialog(frame,
                messages.getString(
                    "You_should_change_units_pair_or_triple_selector."),
                "", JOptionPane.INFORMATION_MESSAGE);
            twoUnitsComboBox.requestFocusInWindow();
            return;
        }

        if (threeUnitsComboBox.getForeground() == Color.RED
            && et.getTotalResults3Units() == 0
            && ((Integer)threeUnitsSpinner.getValue()).intValue() == 1) {

            JOptionPane.showMessageDialog(frame,
                messages.getString(
                    "You_should_change_units_pair_or_triple_selector."),
                "", JOptionPane.INFORMATION_MESSAGE);
            threeUnitsComboBox.requestFocusInWindow();
            return;
        }

        // apply rules
        applyRulesToExtractionResults();

        // update iteration lists
        acceptedTermsIterationjComboBox.addItem(new Integer(iteration));
        rejectedTermsIterationjComboBox.addItem(new Integer(iteration));
        undefinedTermsIterationjComboBox.addItem(new Integer(iteration));

        // change the name of results file
        outputFileComboBox.setSelectedItem(
            (String)optionsDirjComboBox.getSelectedItem()+"-"
            +(new DecimalFormat("00")).format(
                (new Integer(iteration)).doubleValue())
            +".results");

        // increment the number of the extraction iteration
        iteration++;
    }//GEN-LAST:event_extractButtonActionPerformed

    /**
     * Save the extraction results in a file.
     */
    private void savejButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_savejButtonActionPerformed

        PrintWriter out;
        TableModel tm;

        if (resultsjTable.getRowCount() == 0) { return; }

        try {

            // test if name file is provided
            if (outputFileComboBox.getSelectedItem() == null
                || ((String)outputFileComboBox.getSelectedItem())
                .trim().length() == 0) {
                throw (new IOException(messages.getString(
                          "You_must_provide_a_file_name.")));
            }

            // open the file in writing mode
            out = new PrintWriter(new BufferedWriter(new FileWriter(
                            (String)outputFileComboBox.getSelectedItem())));

            // save results in file
            tm = resultsjTable.getModel();
            for (int row = 0; row < tm.getRowCount(); row++) {
                if (tm.getValueAt(row, 0) != null) {
                    out.println(tm.getValueAt(row, 0));
                } else {
                    out.println("");
                }
            }

            // close the file
            out.close();

            statejLabel.setText(
                messages.getString("Results_saved_in_file.")
                +" ("+(String)outputFileComboBox.getSelectedItem()+")");

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_writing_results"),
                JOptionPane.ERROR_MESSAGE);
            outputFileComboBox.requestFocusInWindow();
            return;
        }

    }//GEN-LAST:event_savejButtonActionPerformed

    /**
     * Save the options in an options file.
     */
    private void saveOptions() {

        String line, fileName, rulesFileName, inputFileName, version;
        PrintWriter out;
        File optionFile;
        BufferedReader br;
        TableModel tm, tm2;
        ObjectOutputStream oos;

        try {

            inputFileName = inputFileNameFromInterface();
            if (inputFileName == null) { return; }
            optionFile = new File(changeExtension(inputFileName, "options"));

            if (optionFile.exists()) {
                br = new BufferedReader(new FileReader(optionFile));
                line = br.readLine();
                br.close();

                // detect several versions of options format
                version = "unknown";
                if (line != null) {
                    if (line.equals("Exit Options Format 0.2")) {
                        version = "0.2";
                    } else if (line.equals("Exit Options Format 0.3")) {
                        version = "0.3";
                    }
                }

                if (version.equals("unknown")) {
                    Object [] options = { messages.getString("Write_on"),
                                          messages.getString("Cancel") };

                    int result = JOptionPane.showOptionDialog(frame,
                        messages.getString("This_file_already_exist_and_is_"
                        +"not_an_options_file.")+"\n("+optionFile+")\n\n"
                        +messages.getString(
                            "Do_you_want_to_write_on_this_file_?"),
                        messages.getString("Saving_options"),
                        JOptionPane.DEFAULT_OPTION,
                        JOptionPane.QUESTION_MESSAGE,
                        null, options, options[1]);

                    if (result == 1) {
                        optionsComboBox.requestFocusInWindow();
                        return;
                    }
                }
            }

            out = new PrintWriter(new BufferedWriter(
                      new FileWriter(optionFile)));

            // save options in file

            // format header
            out.println("Exit Options Format 0.3");

            // input file
            out.println(inputFileName);

            // output file
            out.println((String)outputFileComboBox.getSelectedItem());

            tm = regexjTable.getModel();

            // unit definition name
            for (int row = 0; row < 10; row++) {
                if (tm.getValueAt(row, 0) != null) {
                    out.println((String)tm.getValueAt(row, 0));
                } else {
                    out.println("");
                }
            }
            // unit definition regex
            for (int row = 0; row < 10; row++) {
                if (tm.getValueAt(row, 1) != null) {
                    out.println((String)tm.getValueAt(row, 1));
                } else {
                    out.println("");
                }
            }
            out.println(unitComboBox.getSelectedIndex());

            // sentence/phrase separators definition name
            for (int row = 0; row < 10; row++) {
                if (tm.getValueAt(row, 2) != null) {
                    out.println((String)tm.getValueAt(row, 2));
                } else {
                    out.println("");
                }
            }
            // sentence/phrase separators definition regex
            for (int row = 0; row < 10; row++) {
                if (tm.getValueAt(row, 3) != null) {
                    out.println((String)tm.getValueAt(row, 3));
                } else {
                    out.println("");
                }
            }
            out.println(sentPhrComboBox.getSelectedIndex());

            // name of text selector name
            for (int row = 0; row < 10; row++) {
                if (tm.getValueAt(row, 4) != null) {
                    out.println((String)tm.getValueAt(row, 4));
                } else {
                    out.println("");
                }
            }
            // name of text selector regex
            for (int row = 0; row < 10; row++) {
                if (tm.getValueAt(row, 5) != null) {
                    out.println((String)tm.getValueAt(row, 5));
                } else {
                    out.println("");
                }
            }
            out.println(nameOfTextjComboBox.getSelectedIndex());
            
            tm = regex2jTable.getModel();

            // units pair selector name
            for (int row = 0; row < 100; row++) {
                if (tm.getValueAt(row, 0) != null) {
                    out.println((String)tm.getValueAt(row, 0));
                } else {
                    out.println("");
                }
            }
            // units pair selector regex 1
            for (int row = 0; row < 100; row++) {
                if (tm.getValueAt(row, 1) != null) {
                    out.println((String)tm.getValueAt(row, 1));
                } else {
                    out.println("");
                }
            }
            // units pair selector regex 2
            for (int row = 0; row < 100; row++) {
                if (tm.getValueAt(row, 2) != null) {
                    out.println((String)tm.getValueAt(row, 2));
                } else {
                    out.println("");
                }
            }
            // units pair tag kept
            for (int row = 0; row < 100; row++) {
                if (tm.getValueAt(row, 3) != null) {
                    out.println(((Integer)tm.getValueAt(row, 3))
                                .toString());
                } else {
                    out.println("");
                }
            }
            out.println(twoUnitsComboBox.getSelectedIndex());

            // units triple selector name
            for (int row = 0; row < 100; row++) {
                if (tm.getValueAt(row, 4) != null) {
                    out.println((String)tm.getValueAt(row, 4));
                } else {
                    out.println("");
                }
            }
            // units triple selector regex 1
            for (int row = 0; row < 100; row++) {
                if (tm.getValueAt(row, 5) != null) {
                    out.println((String)tm.getValueAt(row, 5));
                } else {
                    out.println("");
                }
            }
            // units triple selector regex 2
            for (int row = 0; row < 100; row++) {
                if (tm.getValueAt(row, 6) != null) {
                    out.println((String)tm.getValueAt(row, 6));
                } else {
                    out.println("");
                }
            }
            // units triple selector regex 3
            for (int row = 0; row < 100; row++) {
                if (tm.getValueAt(row, 7) != null) {
                    out.println((String)tm.getValueAt(row, 7));
                } else {
                    out.println("");
                }
            }
            // units triple tag kept
            for (int row = 0; row < 100; row++) {
                if (tm.getValueAt(row, 8) != null) {
                    out.println(((Integer)tm.getValueAt(row, 8))
                                .toString());
                } else {
                    out.println("");
                }
            }
            out.println(threeUnitsComboBox.getSelectedIndex());

            if (twoUnitsComboBox.getForeground() == Color.RED) {
                out.println("0");
            } else {
                out.println("1");
            }

            tm = formulasjTable.getModel();

            // formulas names
            for (int row = 0; row < 10; row++) {
                if (tm.getValueAt(row, 0) != null) {
                    out.println((String)tm.getValueAt(row, 0));
                } else {
                    out.println("");
                }
            }
            // formulas expressions
            for (int row = 0; row < 10; row++) {
                if (tm.getValueAt(row, 1) != null) {
                    out.println((String)tm.getValueAt(row, 1));
                } else {
                    out.println("");
                }
            }
            out.println(measuresjComboBox.getSelectedIndex());

            tm = termsTypesjTable.getModel();

            // name of accepted terms type
            for (int row = 0; row < 50; row++) {
                if (tm.getValueAt(row, 0) != null) {
                    out.println((String)tm.getValueAt(row, 0));
                } else {
                    out.println("");
                }
            }
            // name of rejected terms type
            for (int row = 0; row < 50; row++) {
                if (tm.getValueAt(row, 1) != null) {
                    out.println((String)tm.getValueAt(row, 1));
                } else {
                    out.println("");
                }
            }

            // tf*idf
            if (tfidfjCheckBox.isSelected()) {
                out.println("true");
            } else {
                out.println("false");
            }

            // occurrence first
            if (occFirstjCheckBox.isSelected()) {
                out.println("true");
            } else {
                out.println("false");
            }

            // separator between word and tag
            out.println(separatorjTextField.getText());
            // results format number of integers
            out.println(((Integer)nbIntegersSpinner.getValue()).toString());
            // results format number of decimals
            out.println(((Integer)nbDecimalsSpinner.getValue()).toString());
            // max results units pair
            out.println(((Integer)twoUnitsResultsSpinner
                         .getValue()).toString());
            // max results units triple
            out.println(((Integer)threeUnitsResultsSpinner
                         .getValue()).toString());
            // pruning threshold units pair
            out.println(((Integer)twoUnitsSpinner.getValue()).toString());
            // pruning threshold units triple
            out.println(((Integer)threeUnitsSpinner.getValue()).toString());

            // liste d'acceptation
            out.println(acceptListjComboBox.getSelectedItem());
            // liste de rejet
            out.println(rejectListjComboBox.getSelectedItem());
            // affichage de l'aide sous le curseur
            if (ToolTipManager.sharedInstance().isEnabled()) {
                out.println("true");
            } else {
                out.println("false");
            }
            // language of the interface
            if (automaticjRadioButtonMenuItem.isSelected()){
                out.println("automatic");
            } else if (englishjRadioButtonMenuItem.isSelected()) {
                out.println("english");
            } else if (francaisjRadioButtonMenuItem.isSelected()) {
                out.println("francais");
            }

            out.close();

             // change the file  extension to rules
            rulesFileName = changeExtension(inputFileName, "rules");

            // save rules in a file
            oos = new ObjectOutputStream(
                new FileOutputStream(rulesFileName));
            oos.writeObject(localRules2Units);
            oos.writeObject(localRules3Units);
            oos.close();

            // save memory tables in a file
            // change the extension to memtbl
            fileName = changeExtension(inputFileName, "memtbl");
            oos = new ObjectOutputStream(new FileOutputStream(fileName));
            oos.writeObject(acceptedTerms);
            oos.writeObject(rejectedTerms);
            oos.writeObject(undefinedTerms);
            oos.writeObject(currentSkippedTermsTreeSet);
            oos.writeObject(new Integer(iteration));
            oos.close();

            statejLabel.setText(
                messages.getString("Options_saved_in_file.")
                +" ("+optionFile+")");

        } catch (IOException ioe) {
            mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_saving_options"),
                JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return;

        } catch (IllegalArgumentException iae) {
            mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
            JOptionPane.showMessageDialog(frame,
                iae.getLocalizedMessage(),
                messages.getString("Error_in_saving_options"),
                JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return;
        }
    }

    /**
     * Load the options from an options file.
     */
    private void loadOptions() {

        String line, saveFileName, rulesFileName, inputFileName, version;
        BufferedReader br;
        TableModel tm;
        TableModel tm2;
        UneditableDefaultTableModel udtm;
        LinkedHashSet lhs;
        Object o;
        ObjectInputStream ois;
        String currentOptionsDir;
        int selectedIndexes[];
    
        try {
            // table containing selected elements in the drop down lists
            // of extraction options
            selectedIndexes = new int [6];

            inputFileName = inputFileNameFromInterface();
            if (inputFileName == null) { return; }

            currentOptionsDir =
                (String)optionsDirjComboBox.getSelectedItem();

            // open options file in reading mode
            br = new BufferedReader(new FileReader(
                    changeExtension(inputFileName, "options")));

            // load options from file

            // detect several versions of options format
            line = br.readLine();
            version = "unknown";
            if (line != null) {
                if (line.equals("Exit Options Format 0.2")) {
                    version = "0.2";
                } else if (line.equals("Exit Options Format 0.3")) {
                    version = "0.3";
                }
            }

            // format header
            if (version.equals("unknown")) {
                throw (new IOException(messages.getString(
                          "Bad_format_for_options_file.")));
            }

            // input file
            // this information is no more used now
            line = br.readLine();

            // output file
            // this information is no more used now
            line = br.readLine();

            tm = regexjTable.getModel();

            // unit definition name
            for (int row = 0; row < 10; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 0);
                } else {
                    tm.setValueAt(null, row, 0);
                }
            }
            // unit definition regex
            for (int row = 0; row < 10; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 1);
                } else {
                    tm.setValueAt(null, row, 1);
                }
            }
            selectedIndexes[0] = Integer.parseInt(br.readLine());

            // sentence/phrase separators definition name
            for (int row = 0; row < 10; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 2);
                } else {
                    tm.setValueAt(null, row, 2);
                }
            }
            // sentence/phrase separators definition regex
            for (int row = 0; row < 10; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 3);
                } else {
                    tm.setValueAt(null, row, 3);
                }
            }
            selectedIndexes[1] = Integer.parseInt(br.readLine());

            // name of text selector name
            for (int row = 0; row < 10; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 4);
                } else {
                    tm.setValueAt(null, row, 4);
                }
            }
            // name of text selector regex
            for (int row = 0; row < 10; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 5);
                } else {
                    tm.setValueAt(null, row, 5);
                }
            }
            selectedIndexes[2] = Integer.parseInt(br.readLine());
            
            tm = regex2jTable.getModel();

            // units pair selector name
            // and terms relation combo boxes
            for (int row = 0; row < 100; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 0);
                } else {
                    tm.setValueAt(null, row, 0);
                }
            }

            // units pair selector regex 1
            for (int row = 0; row < 100; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 1);
                } else {
                    tm.setValueAt(null, row, 1);
                }
            }
            // units pair selector regex 2
            for (int row = 0; row < 100; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 2);
                } else {
                    tm.setValueAt(null, row, 2);
                }
            }
            // units pair tag kept
            for (int row = 0; row < 100; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(Integer.valueOf(line), row, 3);
                } else {
                    tm.setValueAt(null, row, 3);
                }
            }
            selectedIndexes[3] = Integer.parseInt(br.readLine());

            // units triple selector name
            // and terms relation combo boxes
            for (int row = 0; row < 100; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 4);
                } else {
                    tm.setValueAt(null, row, 4);
                }
            }
            // units triple selector regex 1
            for (int row = 0; row < 100; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 5);
                } else {
                    tm.setValueAt(null, row, 5);
                }
            }
            // units triple selector regex 2
            for (int row = 0; row < 100; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 6);
                } else {
                    tm.setValueAt(null, row, 6);
                }
            }
            // units triple selector regex 3
            for (int row = 0; row < 100; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 7);
                } else {
                    tm.setValueAt(null, row, 7);
                }
            }
            // units triple tag kept
            for (int row = 0; row < 100; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(Integer.valueOf(line), row, 8);
                } else {
                    tm.setValueAt(null, row, 8);
                }
            }
            selectedIndexes[4] = Integer.parseInt(br.readLine());

            // the type of units extracted
            if (Integer.parseInt(br.readLine()) == 0) {
                twoUnitsComboBoxFocusGained(null);
            } else {
                threeUnitsComboBoxFocusGained(null);
            }

            tm = formulasjTable.getModel();

            // formulas names
            for (int row = 0; row < 10; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 0);
                } else {
                    tm.setValueAt(null, row, 0);
                }
            }
            // formulas expressions
            for (int row = 0; row < 10; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 1);
                } else {
                    tm.setValueAt(null, row, 1);
                }
            }
            selectedIndexes[5] = Integer.parseInt(br.readLine());

            tm = termsTypesjTable.getModel();

            // name of accepted terms type
            for (int row = 0; row < 50; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 0);
                } else {
                    tm.setValueAt(null, row, 0);
                }
            }
            // name of rejected terms type
            for (int row = 0; row < 50; row++) {
                line = br.readLine();
                if (!line.trim().equals("")) {
                    tm.setValueAt(line, row, 1);
                } else {
                    tm.setValueAt(null, row, 1);
                }
            }

            // tf*idf
            if (br.readLine().equals("true")) {
                tfidfjCheckBox.setSelected(true);
            } else {
                tfidfjCheckBox.setSelected(false);
            }

            // occurrence first
            if (version.equals("0.3")) {
                if (br.readLine().equals("true")) {
                    occFirstjCheckBox.setSelected(true);
                } else {
                    occFirstjCheckBox.setSelected(false);
                }
            }

            // separator between word and tag
            separatorjTextField.setText(br.readLine());
            // results format number of integers
            nbIntegersSpinner.setValue(Integer.valueOf(br.readLine()));
            // results format number of decimals
            nbDecimalsSpinner.setValue(Integer.valueOf(br.readLine()));
            // max results units pair
            twoUnitsResultsSpinner.setValue(Integer.valueOf(br.readLine()));
            // max results units triple
            threeUnitsResultsSpinner.setValue(
                Integer.valueOf(br.readLine()));
            // pruning threshold units pair
            twoUnitsSpinner.setValue(Integer.valueOf(br.readLine()));
            // pruning threshold units triple
            threeUnitsSpinner.setValue(Integer.valueOf(br.readLine()));

            // acceptation list
            line = br.readLine();
            addItemUnique(acceptListjComboBox, line);
            acceptListjComboBox.setSelectedItem(line);
            // reject list
            line = br.readLine();
            addItemUnique(rejectListjComboBox, line);
            rejectListjComboBox.setSelectedItem(line);

            // help under cursor
            // this information is no more used now
            line = br.readLine();
            // language of the interface
            // this information is no more used now
            line = br.readLine();

            br.close();

            updateExpressions();
            unitComboBox.setSelectedIndex(selectedIndexes[0]);
            sentPhrComboBox.setSelectedIndex(selectedIndexes[1]);
            nameOfTextjComboBox.setSelectedIndex(selectedIndexes[2]);
            twoUnitsComboBox.setSelectedIndex(selectedIndexes[3]);
            threeUnitsComboBox.setSelectedIndex(selectedIndexes[4]);
            measuresjComboBox.setSelectedIndex(selectedIndexes[5]);

            iteration = 0;

            // number of lines of the input file
            inputFileLines = 0;

            // clear all memory tables because new input file is chosen
            acceptedTerms.clear();
            currentAcceptedTerms.clear();
            rejectedTerms.clear();
            undefinedTerms.clear();
            currentAcceptedTermsTreeSet.clear();
            currentSkippedTermsTreeSet.clear();
            localRules2Units.clear();
            localRules3Units.clear();

            // clear all selections in the results table
            resultsjTable.clearSelection();

            // clear the status of the results table cells
            Arrays.fill(rowsStatus, UNDEFINED);

            // clear tables and lists content
            resultsjTable.setModel(new DefaultTableModel());
            acceptedTermsjTable.setModel(new UneditableDefaultTableModel());
            rejectedTermsjTable.setModel(new UneditableDefaultTableModel());
            undefinedTermsjTable.setModel(
                new UneditableDefaultTableModel());
            ((DefaultListModel)alreadyFormedTermsjList.getModel()).clear();
            ((DefaultListModel)localRules2UnitsjList.getModel()).clear();
            ((DefaultListModel)localRules3UnitsjList.getModel()).clear();

            // clear iteration terms filter lists
            acceptedTermsIterationjComboBox.removeAllItems();
            rejectedTermsIterationjComboBox.removeAllItems();
            undefinedTermsIterationjComboBox.removeAllItems();

            // forbid the change of the extraction parameters
            optionsDirjComboBox.setEnabled(false);
            optionsComboBox.setEnabled(false);

            rulesFileName = changeExtension(inputFileName, "rules");

            // if a rules file exist
            if ((new File(rulesFileName)).exists()) {

                // load rules from a file
                ois = new ObjectInputStream(
                    new FileInputStream(rulesFileName));
                localRules2Units = (Vector)ois.readObject();
                localRules3Units = (Vector)ois.readObject();
                ois.close();
                updateLocalRulesList(-1, -1);
            }

            // change the extension to memtbl
            saveFileName = changeExtension(inputFileName, "memtbl");

            // if a backup exists and it should always exists
            // but the 0.memtbl
            if ((new File(saveFileName)).exists()) {

                // load memory tables from a file
                ois = new ObjectInputStream(
                    new FileInputStream(saveFileName));
                acceptedTerms = (LinkedHashSet)ois.readObject();
                rejectedTerms = (LinkedHashSet)ois.readObject();
                undefinedTerms = (LinkedHashSet)ois.readObject();
                currentSkippedTermsTreeSet = (TreeSet)ois.readObject();
                iteration = ((Integer)ois.readObject()).intValue();
                ois.close();

                // update terms iteration lists
                acceptedTermsIterationjComboBox.addItem(
                    messages.getString("All"));
                rejectedTermsIterationjComboBox.addItem(
                    messages.getString("All"));
                undefinedTermsIterationjComboBox.addItem(
                    messages.getString("All"));
                for (int i = 0; i < iteration; i++) {
                    acceptedTermsIterationjComboBox.addItem(
                        new Integer(i));
                    rejectedTermsIterationjComboBox.addItem(
                        new Integer(i));
                    undefinedTermsIterationjComboBox.addItem(
                        new Integer(i));
                }
            }

            if (iteration == 0) {
                mainjTabbedPane.setEnabledAt(1, true);
                enableExpressions();
            } else {
                mainjTabbedPane.setEnabledAt(1, false);
                // disable filled cells in the Expressions panel tables
                disableExpressions();
            }

            statejLabel.setText(
                messages.getString("Options_loaded_from_file.")
                +" ("+changeExtension(inputFileName, "options")+")");

        } catch (FileNotFoundException fnfe) {
            mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
            JOptionPane.showMessageDialog(frame,
                messages.getString("File_not_found")
                +"\n"+fnfe.getLocalizedMessage(),
                messages.getString("Error_in_loading_options"),
                JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return;

        } catch (IOException ioe) {
            mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_loading_options"),
                JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return;

        } catch (ClassNotFoundException cnfe) {
            mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
            JOptionPane.showMessageDialog(frame,
                messages.getString("Bad_format_for_options_file.")
                +"\n"+cnfe.getLocalizedMessage(),
                messages.getString("Error_in_loading_options"),
                JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return;

        } catch (IllegalArgumentException iae) {
            mainjTabbedPane.setSelectedComponent(extractionjSplitPane);
            JOptionPane.showMessageDialog(frame,
                messages.getString("Bad_format_for_options_file.")
                +"\n"+iae.getLocalizedMessage(),
                messages.getString("Error_in_loading_options"),
                JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return;
        }
    }


    /*-************
     * Menu items *
     **************/

    /**
     * Shows a dialog box with information about program.
     */
    private void aboutMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_aboutMenuItemActionPerformed

        Object[] options = { messages.getString("Close"),
                             messages.getString("Log") };
        int result = JOptionPane.showOptionDialog(
            frame,
            messages.getString(
            "Terms_extractor_Grapical_User_Interface"),
            messages.getString("About_this_program"),
            JOptionPane.DEFAULT_OPTION,            
            JOptionPane.INFORMATION_MESSAGE,
            null, options, options[0]);

        if (result == 1) {
            String url = "file:"
                +System.getProperty("user.dir")
                +System.getProperty("file.separator")
                +"doc"
                +System.getProperty("file.separator")
                +"historique.html";
            HTMLWindow(url, messages.getString("Log"));
        }


    }//GEN-LAST:event_aboutMenuItemActionPerformed

    /**
     * Diplay help of the program.
     */
    private void contentsMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_contentsMenuItemActionPerformed

        String url;

        try {

        url = "file:"
            +System.getProperty("user.dir")
            +System.getProperty("file.separator")
            +"doc"
            +System.getProperty("file.separator")
            +"aide.html#0";

        if (helpFrame == null) {
            helpFrame =
                HTMLWindow(url, messages.getString("Help_contents"));

        } else {
            helpFrame.setVisible(true);
            helpFrame.toFront();

            // it seems a little complicated !!
            ((JEditorPane)
             ((JScrollPane)
              helpFrame.getRootPane().getContentPane().getComponent(0))
             .getViewport().getView())
                .setPage(url);
        }

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(
                frame,
                messages.getString("Attempted_to_read_a_bad_URL.")
                +"\n"+ioe.getLocalizedMessage(),
                messages.getString("URL_not_found"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }
    }//GEN-LAST:event_contentsMenuItemActionPerformed

    private void helpUnderCursorjCheckBoxMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_helpUnderCursorjCheckBoxMenuItemActionPerformed

        javax.swing.ToolTipManager toolTipManager;

        toolTipManager = ToolTipManager.sharedInstance();

            if (helpUnderCursorjCheckBoxMenuItem.isSelected()) {
                toolTipManager.setEnabled(true);
 
            } else {
                toolTipManager.setEnabled(false);
            }
    }//GEN-LAST:event_helpUnderCursorjCheckBoxMenuItemActionPerformed

    /**
     * Display help according to the selected element of the interface.
     */
    private void contexualHelpjMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_contexualHelpjMenuItemActionPerformed

        String anchor = "";
        String url;
        JMenu menu;
        Component selectedTab = mainjTabbedPane.getSelectedComponent();

        KeyboardFocusManager focusManager =
            KeyboardFocusManager.getCurrentKeyboardFocusManager();

        // selected component.
        Component sct = focusManager.getFocusOwner();

        // selected component before the menu bar got focus.
        Component sc = focusManager.getPermanentFocusOwner();

        // if the selected component is temporary, for example a menu
        if (!sc.equals(sct)) {
            menu = menuBar.getMenu(
                menuBar.getSelectionModel().getSelectedIndex());
            if (menu.getText().equals(messages.getString("Program"))) {
                anchor = "#5";
            } else if (menu.getText().equals(messages.getString("Help"))) {
                anchor = "#2.7";
            }

        // set anchor url according to selected index of the tabbed pane
        // and the selected component
        } else if (selectedTab.equals(regexFormulasjPanel)) {
            if (regexjTable.equals(sc) || regex2jTable.equals(sc)) {
                anchor = "#3.1";
            } else if (formulasjTable.equals(sc)) {
                anchor = "#3.2";
            } else if (termsTypesjTable.equals(sc)) {
                anchor = "#3.3";
            } else if (separatorjTextField.equals(sc)) {
                anchor = "#3.1";
            } else {
                anchor = "#3";
            }

        } else if (selectedTab.equals(preProcessingjPanel)) {
            if (alreadyFormedTermsjList.equals(sc)) {
                anchor = "#4.1";
            } else if (acceptListjComboBox.getEditor()
                        .getEditorComponent().equals(sc)
                    || rejectListjComboBox.getEditor()
                        .getEditorComponent().equals(sc)) {
                anchor = "#4.2";
            } else {
                anchor = "#4";
            }

        } else if (selectedTab.equals(extractionjSplitPane)) {
            if (precedentTermsjButton.equals(sc)) {
                anchor = "#7";
            }  else if (optionsComboBox.getEditor()
                        .getEditorComponent().equals(sc)
            || optionsDirjComboBox.getEditor()
                        .getEditorComponent().equals(sc)
            || outputFileComboBox.getEditor()
                .getEditorComponent().equals(sc)) {
                anchor = "#5.1";
            } else if (unitComboBox.equals(sc)
                       || sentPhrComboBox.equals(sc)
                       || nameOfTextjComboBox.equals(sc)
                       || twoUnitsComboBox.equals(sc)
                       || threeUnitsComboBox.equals(sc)) {
                anchor = "#5.2";
            } else if (measuresjComboBox.equals(sc)
                       || tfidfjCheckBox.equals(sc)
                       || occFirstjCheckBox.equals(sc)) {
                anchor = "#5.3";
            } else if (((javax.swing.JSpinner.DefaultEditor)twoUnitsSpinner
                        .getEditor()).getTextField().equals(sc)
            || ((javax.swing.JSpinner.DefaultEditor)twoUnitsResultsSpinner
                  .getEditor()).getTextField().equals(sc)
            || ((javax.swing.JSpinner.DefaultEditor)threeUnitsSpinner
                  .getEditor()).getTextField().equals(sc)
            || ((javax.swing.JSpinner.DefaultEditor)threeUnitsResultsSpinner
                .getEditor()).getTextField().equals(sc)
            || ((javax.swing.JSpinner.DefaultEditor)nbIntegersSpinner
                .getEditor()).getTextField().equals(sc)
            || ((javax.swing.JSpinner.DefaultEditor)nbDecimalsSpinner
                .getEditor()).getTextField().equals(sc)) {
                anchor = "#5.4";
            } else if (resultsjTable.equals(sc)) {
                anchor = "#5.5";
            } else if (contextjTextPane.equals(sc)) {
                anchor = "#5.7";
            } else {
                anchor = "#5";
            }

        } else if (selectedTab.equals(rulesjPanel)) {
            anchor = "#6";

        } else if (selectedTab.equals(termsjSplitPane)) {
            if (termsContextjTextPane.equals(sc)) {
                anchor = "#5.7";
            } else {
                anchor = "#5.6";
            }

        } else if (selectedTab.equals(undefinedTermsjSplitPane)) {
            if (precedentTermsUndefinedTermsjButton.equals(sc)) {
                anchor = "#7";
            } else if (undefinedTermsContextjTextPane.equals(sc)) {
                anchor = "#5.7";
            }  else {
                anchor = "#5.6";
            }
        }

    try {
        url = "file:"
            +System.getProperty("user.dir")
            +System.getProperty("file.separator")
            +"doc"
            +System.getProperty("file.separator")
            +"aide.html"+anchor;

        if (helpFrame == null) {
            helpFrame =
                HTMLWindow(url, messages.getString("Help_contents"));

        } else {
            helpFrame.setVisible(true);
            helpFrame.toFront();

            ((JEditorPane)
             ((JScrollPane)
              helpFrame.getRootPane().getContentPane().getComponent(0))
             .getViewport().getView())
                .setPage(url);
        }

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(
                frame,
                messages.getString("Attempted_to_read_a_bad_URL.")
                +"\n"+ioe.getLocalizedMessage(),
                messages.getString("URL_not_found"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }

    }//GEN-LAST:event_contexualHelpjMenuItemActionPerformed


    /**
     * Menu item for selecting new input file.
     * The user must choose an input file. It is copied in a new directory
     * named by the base name of the input file with the name 0.iter.
     * Optionaly the user can choose an options file and a rule file
     * otherwise there are empty.
     * An empty 0.hist and 0.memtbl is created.
     * A .info with the int value 0 is created.
     */
    private void selectNewInputFilejMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_selectNewInputFilejMenuItemActionPerformed

        JFileChooser fileChooser;
        String inputFileName, directoryName, optionsFileName, rulesFileName;
        String fileName;
        File currentDir;
        int result;
        RandomAccessFile raf;
        FileChannel srcChannel, dstChannel;

        try {

        currentDir = new File(System.getProperty("user.dir"));
        fileChooser = new JFileChooser(currentDir);
        fileChooser.setDialogTitle(messages.getString(
            "name_of_input_file_containing_text_to_analyse"));
        fileChooser.addChoosableFileFilter(
            new FileChooserFilterExtension(".iter",
                messages.getString("Iterated_files")));
        fileChooser.addChoosableFileFilter(
            new FileChooserFilterExtension(".corpus",
                messages.getString("Corpus_files")));
        result = fileChooser.showOpenDialog(frame);

        if (result != JFileChooser.APPROVE_OPTION) { return; }

        inputFileName = fileChooser.getSelectedFile().getCanonicalPath();
        directoryName = fileChooser.getSelectedFile().getName();
        directoryName = removeFileExtension(directoryName);

        // test if the directory already exists
        while  ((new File(directoryName)).exists()
             && (new File(directoryName)).isDirectory()) {
            directoryName = (String)
            JOptionPane.showInputDialog(frame,
                messages.getString("Such_an_input_file_is_already_analysed."
                +"_Please_rename_your_new_input_file."),
                messages.getString("Error_in_input_parameters"),
                JOptionPane.ERROR_MESSAGE,
                null, null, directoryName);
            if (directoryName == null) { return; }
        }

        fileChooser = new JFileChooser(currentDir);
        fileChooser.setDialogTitle(messages.getString(
            "name_of_options_file"));
        fileChooser.addChoosableFileFilter(
            new FileChooserFilterExtension(".options",
                messages.getString("Options_files")));
        result = fileChooser.showOpenDialog(frame);

        if (result != JFileChooser.APPROVE_OPTION) { return; }

        optionsFileName = fileChooser.getSelectedFile().getCanonicalPath();

        result = JOptionPane.showConfirmDialog(frame,
            messages.getString("Do_you_want_to_import_a_rules_file_?"),
            "",
            JOptionPane.YES_NO_OPTION,
            JOptionPane.QUESTION_MESSAGE);

        if (result == JOptionPane.NO_OPTION) {
            rulesFileName = null;
        } else {
            fileChooser = new JFileChooser(currentDir);
            fileChooser.setDialogTitle(messages.getString(
                "Name_of_rules_file."));
            fileChooser.addChoosableFileFilter(
                new FileChooserFilterExtension(".rules",
                    messages.getString("Rules_files")));
            result = fileChooser.showOpenDialog(frame);
            if (result != JFileChooser.APPROVE_OPTION) {
                rulesFileName = null;
            } else {
                rulesFileName =
                    fileChooser.getSelectedFile().getCanonicalPath();
            }
        }

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }

        
        iteration = 0;

        // clear all memory tables because new input file is chosen
        acceptedTerms.clear();
        currentAcceptedTerms.clear();
        rejectedTerms.clear();
        undefinedTerms.clear();
        currentAcceptedTermsTreeSet.clear();
        currentSkippedTermsTreeSet.clear();
        localRules2Units.clear();
        localRules3Units.clear();

        // clear all selections in the results table
        resultsjTable.clearSelection();

        // clear the status of the results table cells
        Arrays.fill(rowsStatus, UNDEFINED);

        // clear tables and lists content
        resultsjTable.setModel(new DefaultTableModel());
        acceptedTermsjTable.setModel(new UneditableDefaultTableModel());
        rejectedTermsjTable.setModel(new UneditableDefaultTableModel());
        undefinedTermsjTable.setModel(
            new UneditableDefaultTableModel());
        ((DefaultListModel)alreadyFormedTermsjList.getModel()).clear();
        ((DefaultListModel)localRules2UnitsjList.getModel()).clear();
        ((DefaultListModel)localRules3UnitsjList.getModel()).clear();

        // clear selected output file
        outputFileComboBox.setSelectedItem(null);

        // clear terms filter lists
        acceptedTermsRelationjComboBox.removeAllItems();
        rejectedTermsRelationjComboBox.removeAllItems();
        undefinedTermsRelationjComboBox.removeAllItems();
        acceptedTermsIterationjComboBox.removeAllItems();
        rejectedTermsIterationjComboBox.removeAllItems();
        undefinedTermsIterationjComboBox.removeAllItems();

        // enabled again the elements of the interface
        mainjTabbedPane.setEnabledAt(1, true);


        // create a new directory which name is the name of the input file
        (new File(directoryName)).mkdir();

        try {
            // copy input file to 0.iter in the new directory
            // Create channel on the source
            srcChannel = new FileInputStream(inputFileName).getChannel();
            // Create channel on the destination
            dstChannel = new FileOutputStream(directoryName
                +System.getProperty("file.separator")+"0.iter")
                .getChannel();
            // Copy file contents from source to destination
            dstChannel.transferFrom(srcChannel, 0, srcChannel.size());
            // Close the channels
            srcChannel.close(); dstChannel.close();

            // copy options file to 0.options in the new directory
            srcChannel = new FileInputStream(optionsFileName).getChannel();
            dstChannel = new FileOutputStream(directoryName
                +System.getProperty("file.separator")+"0.options")
                .getChannel();
            dstChannel.transferFrom(srcChannel, 0, srcChannel.size());
            srcChannel.close(); dstChannel.close();

            if (rulesFileName != null
            && !(new File(directoryName+".rules")).exists()) {
                // copy rules file in the "user.dir" directory
                srcChannel = new FileInputStream(rulesFileName).getChannel();
                dstChannel = new FileOutputStream(directoryName
                    +System.getProperty("file.separator")+"0.rules")
                    .getChannel();
                dstChannel.transferFrom(srcChannel, 0, srcChannel.size());
                srcChannel.close(); dstChannel.close();
            }

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_or_writing_file."),
                JOptionPane.ERROR_MESSAGE);
            return;
        }

        try {
            // initialise number for the file naming in .info
            fileName =
                directoryName+System.getProperty("file.separator")+".info";
            raf = new RandomAccessFile(fileName, "rw");
            raf.writeInt(0);
            raf.close();

            // initialise extraction history file 0.hist
            fileName =
                directoryName+System.getProperty("file.separator")+"0.hist";
            if (!(new File(fileName)).createNewFile()) {
                throw (new IOException(fileName));
            }

        } catch  (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_writing_file"),
                JOptionPane.ERROR_MESSAGE);
            return;
        }

        updateFilesList();

        // select the directory and iteration in the interface
        // loading of options is automatic
        optionsDirjComboBox.setSelectedItem(directoryName);
        optionsComboBox.setSelectedItem("0");

        // enable all cells in the Expressions panel tables
        enableExpressions();
    }//GEN-LAST:event_selectNewInputFilejMenuItemActionPerformed

    /**
     * Menu item shortcut for saving the current extraction results and the
     * 3 lists of terms.
     */
    private void saveResultsjMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_saveResultsjMenuItemActionPerformed

        FileChannel srcChannel, dstChannel;
        String iterFileName;

        if (iteration < 1) { return; }

        saveAcceptedTerms();
        saveRejectedTerms();
        saveUndefinedTerms();
        savejButtonActionPerformed(null);

        iterFileName = (String)optionsDirjComboBox.getSelectedItem()+"-"
            +(new DecimalFormat("00")).format(
                (new Integer(iteration-1)).doubleValue())
            +".iter";

        try {
        // copy the current .iter file in the root directory
        srcChannel =
            new FileInputStream(inputFileNameFromInterface()).getChannel();
        dstChannel = new FileOutputStream(iterFileName).getChannel();
        dstChannel.transferFrom(srcChannel, 0, srcChannel.size());
        srcChannel.close(); dstChannel.close();

        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_or_writing_file."),
                JOptionPane.ERROR_MESSAGE);
            return;
        }
    }//GEN-LAST:event_saveResultsjMenuItemActionPerformed

    /**
     * Find language from language environment variable to change local
     * language and thus the language of user interface.
     */
    private void automaticjRadioButtonMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_automaticjRadioButtonMenuItemActionPerformed
        // Add your handling code here:
    }//GEN-LAST:event_automaticjRadioButtonMenuItemActionPerformed

    /**
     * Change local language and thus the language of user interface.
     */
    private void francaisjRadioButtonMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_francaisjRadioButtonMenuItemActionPerformed

        if (Locale.getDefault().equals(new Locale("fr", "FR"))) { return; }

        int result = JOptionPane.showConfirmDialog(frame,
                         messages.getString("All_data_will_be_reset_!"),
                         messages.getString("Changing_language"),
                         JOptionPane.OK_CANCEL_OPTION,
                         JOptionPane.QUESTION_MESSAGE);

        if (result == JOptionPane.CANCEL_OPTION) { return; }

        Locale.setDefault(new Locale("fr", "FR"));
        messages = ResourceBundle.getBundle("messages");
        saveUserPreferences();
        frame.dispose();
        frame = new ExtractTermsGUI();
        frame.show();
    }//GEN-LAST:event_francaisjRadioButtonMenuItemActionPerformed

    /**
     * Change local language and thus the language of user interface.
     */
    private void englishjRadioButtonMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_englishjRadioButtonMenuItemActionPerformed

        if (Locale.getDefault().equals(new Locale("en", "EN"))) { return; }

        int result = JOptionPane.showConfirmDialog(frame,
                         messages.getString("All_data_will_be_reset_!"),
                         messages.getString("Changing_language"),
                         JOptionPane.OK_CANCEL_OPTION,
                         JOptionPane.QUESTION_MESSAGE);

        if (result == JOptionPane.CANCEL_OPTION) { return; }

        Locale.setDefault(new Locale("en", "EN"));
        messages = ResourceBundle.getBundle("messages");
        saveUserPreferences();
        frame.dispose();
        frame = new ExtractTermsGUI();
        frame.show();
    }//GEN-LAST:event_englishjRadioButtonMenuItemActionPerformed

    /**
     * Save program preferences into the user preferences file.
     */
    private void saveUserPreferences() {

        if (automaticjRadioButtonMenuItem.isSelected()){
            prefs.put("language", "automatic");
        } else if (englishjRadioButtonMenuItem.isSelected()) {
            prefs.put("language", "english");
        } else if (francaisjRadioButtonMenuItem.isSelected()) {
            prefs.put("language", "francais");
        }
        prefs.putBoolean("measures",
            displayMeasuresjCheckBoxMenuItem.isSelected());
        prefs.putBoolean("tags", displayTagsjCheckBoxMenuItem.isSelected());
        prefs.putBoolean("occ", displayOccjCheckBoxMenuItem.isSelected());
        prefs.putBoolean("expertMode",
            displayExpertModejCheckBoxMenuItem.isSelected());
        prefs.putBoolean("helpUnderCursor",
            helpUnderCursorjCheckBoxMenuItem.isSelected());

//        try {
//            prefs.exportSubtree(new FileOutputStream(
//                      System.getProperty("user.dir")
//                     +System.getProperty("file.separator")
//                     +"exit.conf"));
//
//        } catch (BackingStoreException bse) {
//            JOptionPane.showMessageDialog(frame,
//                bse.getLocalizedMessage(),
//                messages.getString("Invalid_preferences_format.")
//                +" (exit.conf)",
//                JOptionPane.ERROR_MESSAGE);
//
//        } catch(IOException ioe){
//            JOptionPane.showMessageDialog(frame,
//                ioe.getLocalizedMessage(),
//                messages.getString("Error_in_writing_file")
//                +" (exit.conf)",
//                JOptionPane.ERROR_MESSAGE);
//        }
    }

    /**
     * Menu item for exiting the application. Alias towards exitForm.
     */
    private void exitMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exitMenuItemActionPerformed

        exitForm(null);
    }//GEN-LAST:event_exitMenuItemActionPerformed

    /**
     * Exit the application.
     */
    private void exitForm(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_exitForm

        saveUserPreferences();
        System.exit(0);
    }//GEN-LAST:event_exitForm


    /*-*******************
     * Utilities methods *
     *********************/

    /**
     * Some things to do the first time main window is show.
     */
    private void formWindowOpened(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowOpened

//        prefs = Preferences.userNodeForPackage(ExtractTermsGUI.class);

        // load program preferences from the use preferences file
//        if ((new File(System.getProperty("user.dir")
//                     +System.getProperty("file.separator")
//                     +"exit.conf")).exists()) {
//
//            try {
//                prefs.importPreferences(new FileInputStream(
//                          System.getProperty("user.dir")
//                         +System.getProperty("file.separator")
//                         +"exit.conf"));
//
//            } catch (InvalidPreferencesFormatException ipfe) {
//                JOptionPane.showMessageDialog(frame,
//                    ipfe.getLocalizedMessage(),
//                    messages.getString("Invalid_preferences_format.")
//                    +" (exit.conf)",
//                    JOptionPane.ERROR_MESSAGE);
//
//            } catch(IOException ioe){
//                JOptionPane.showMessageDialog(frame,
//                    ioe.getLocalizedMessage(),
//                    messages.getString("Error_in_reading_file")
//                    +" (exit.conf)",
//                    JOptionPane.ERROR_MESSAGE);
//            }

            if (prefs.get("language", "automatic").equals("automatic")){
                automaticjRadioButtonMenuItem.setSelected(true);
            } else if (prefs.get("language", "english").equals("english")) {
                englishjRadioButtonMenuItemActionPerformed(null);
                englishjRadioButtonMenuItem.setSelected(true);
            } else if (prefs.get("language", "francais").equals("francais")) {
                francaisjRadioButtonMenuItemActionPerformed(null);
                francaisjRadioButtonMenuItem.setSelected(true);
            }
            displayMeasuresjCheckBoxMenuItem.setSelected(
                prefs.getBoolean("measures", false));
            displayTagsjCheckBoxMenuItem.setSelected(
                prefs.getBoolean("tags", false));
            displayOccjCheckBoxMenuItem.setSelected(
                prefs.getBoolean("occ", false));
            displayExpertModejCheckBoxMenuItem.setSelected(
                prefs.getBoolean("expertMode", false));
            displayExpertModejCheckBoxMenuItemActionPerformed(null);
            helpUnderCursorjCheckBoxMenuItem.setSelected(
                prefs.getBoolean("helpUnderCursor", true));
            helpUnderCursorjCheckBoxMenuItemActionPerformed(null);
//        }

        // if there is no iteration directories
        if (optionsDirjComboBox.getItemCount() == 0) {
            
            JOptionPane.showMessageDialog(frame,
                messages.getString(
                    "There_is_no_iteration_file_already_created."
                    +"_Please_select_a_new_input_and_options_file."),
                messages.getString("First_time_you_run_this_program."),
                JOptionPane.INFORMATION_MESSAGE);

            java.util.Timer timer = new java.util.Timer();
            timer.schedule(new FirstTime(timer), 5000);

            selectNewInputFilejMenuItemActionPerformed(null);
        }
    }//GEN-LAST:event_formWindowOpened

    /**
     * Some things to do the first time main window is show.
     */
    private class FirstTime extends java.util.TimerTask {
        java.util.Timer timer;
        
        public FirstTime(java.util.Timer _timer) { timer = _timer; }

        public void run() {
            JFrame jf = HTMLWindow("file:"
                   +System.getProperty("user.dir")
                   +System.getProperty("file.separator")
                   +"doc"
                   +System.getProperty("file.separator")
                   +messages.getString("guide.html"),
                   messages.getString("First_time_you_run_this_program."));

            jf.toFront();
            timer.cancel();
        }
    }

    /**
     * Update panel.
     */
    private void mainjTabbedPaneStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_mainjTabbedPaneStateChanged

        if (mainjTabbedPane.getSelectedComponent()
            .equals(extractionjSplitPane)) {
            updateExpressions();
        }
    }//GEN-LAST:event_mainjTabbedPaneStateChanged

    /**
     * Add a String to the comboBox only if it doesn't already exist.
     */
    private void addItemUnique(JComboBox comboBox, String item) {

        if (item == null) { return; }

        for (int i = 0; i < comboBox.getItemCount(); i++) {
            if (((String)comboBox.getItemAt(i)).equals(item)) {
                return;
            }
        }
        comboBox.addItem(item);
    }

    /**
     * Count lines in a file.
     * @param file name of file to read
     * @return number of line contain in the file ; -1 if an error occurs
     */
    private int countLines(String file) {

        BufferedReader br;
        String lineContent;
        int line;

        try {
            // open the input file in reading mode
            br = new BufferedReader(new FileReader(file));

        } catch (FileNotFoundException fnfe) {
            JOptionPane.showMessageDialog(frame,
                messages.getString("File_not_found")+"\n"
                +fnfe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return -1;
        }

        try {
            lineContent = br.readLine();
            line = 1;
            // read the input file line by line only one time and
            while (lineContent != null) {
                lineContent = br.readLine();
                line++;
            }
            // close the input file
            br.close();

        } catch(IOException ioe){
            JOptionPane.showMessageDialog(frame,
                ioe.getLocalizedMessage(),
                messages.getString("Error_in_reading_file"),
                JOptionPane.ERROR_MESSAGE);
            return -1;
        }

        return line;
    }

    /**
     * Shows an HTML file in a JEditorPane Component included in an external
     * JFrame.
     *
     * @param _url url of HTML file to display
     * @param windowTitle title of the window
     * @return contents frame
     */
    private JFrame HTMLWindow (String _url, String windowTitle) {

        JFrame contentsFrame;

        // build the window
        JEditorPane editorPane = new JEditorPane();
        editorPane.setEditable(false);

        // enable the links activation
        editorPane.addHyperlinkListener(
            new javax.swing.event.HyperlinkListener () {
                public void hyperlinkUpdate(HyperlinkEvent e) {
                    if (e.getEventType()
                        == HyperlinkEvent.EventType.ACTIVATED) {
                        JEditorPane pane = (JEditorPane) e.getSource();
                        if (e instanceof javax.swing.text.html
                            .HTMLFrameHyperlinkEvent) {
                         javax.swing.text.html.HTMLFrameHyperlinkEvent evt =
                           (javax.swing.text.html.HTMLFrameHyperlinkEvent)e;
                         javax.swing.text.html.HTMLDocument doc
                             = (javax.swing.text.html.HTMLDocument)
                             pane.getDocument();
                         doc.processHTMLFrameHyperlinkEvent(evt);
                        } else {
                            try {
                                pane.setPage(e.getURL());
                            } catch (Throwable t) {
                                JOptionPane.showMessageDialog(
                                    pane,
                                    messages.getString(
                                        "Attempted_to_read_a_bad_URL.")
                                    +"\n"+e.getURL(),
                                    messages.getString("URL_not_found"),
                                    JOptionPane.ERROR_MESSAGE);
                                return;
                            }
                        }
                    }
                }
            });

        JScrollPane editorScrollPane = new JScrollPane(editorPane);
        editorScrollPane.setVerticalScrollBarPolicy(
                                     JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        editorScrollPane.setPreferredSize(new Dimension(250, 145));

        contentsFrame = new JFrame(windowTitle);
        contentsFrame.setBounds(50,50, 600, 700);
        contentsFrame.getContentPane().add(editorScrollPane);

        // set the file to be shown in the editorPane
        try {
            editorPane.setPage(_url);

        } catch (IOException e) {
               JOptionPane.showMessageDialog(
               frame,
               messages.getString("Attempted_to_read_a_bad_URL.")
               +e.getLocalizedMessage(),
               messages.getString("URL_not_found"),
               JOptionPane.ERROR_MESSAGE);
               return null;
        }

        // display the window
        contentsFrame.setVisible(true);

        return contentsFrame;
    }

    /**
     * File filter on extension for use with File.list method.
     */
    private class FileFilterExtension implements FilenameFilter {

        private String endWith;

        public FileFilterExtension(String _endWith) {
            endWith = _endWith;
        }

        public boolean accept(File dir, String name) {
            return (name.endsWith(endWith));
        }
    }

    /**
     * Returns the new nameOfFile with extension changed. If nameOfFile
     * doesn't have an extension add an new extension.
     *
     * @param nameOfFile String of the name of file
     * @param extension the new extension
     * @return the new name of file with extension changed ; null if
     * empty string
     */
    private String changeExtension(String nameOfFile, String extension) {

        if (nameOfFile.trim().length() == 0) { return null; }

        if (nameOfFile.lastIndexOf('.') != -1) {
            nameOfFile =
                nameOfFile.substring(0, nameOfFile.lastIndexOf('.'));
        }
        nameOfFile += "."+extension;

        return nameOfFile;
    }

    /**
     * Returns the name of the file without the extension.
     *
     * @param nameOfFile String of the name of file
     * @return the name of file without the extension ; null if
     * empty string
     */
    private String removeFileExtension(String nameOfFile) {

        if (nameOfFile.trim().length() == 0) { return null; }

        if (nameOfFile.lastIndexOf('.') != -1) {
            return nameOfFile.substring(0, nameOfFile.lastIndexOf('.'));
        } else {
            return nameOfFile;
        }
    }

    /**
     * Returns the name of input file from the informations contain in the
     * interface. Display error message and returns null value if directory
     * or file don't exist.
     *
     * @return String of the name of input file ; null if an error occurs
     */
    private String inputFileNameFromInterface() {
        String inputFileName;

        // test String directory
        if (optionsDirjComboBox.getSelectedItem() == null
        || ((String)optionsDirjComboBox.getSelectedItem())
            .trim().length() == 0) {
            JOptionPane.showMessageDialog(frame,
            messages.getString("You_must_provide_a_directory_name."),
            messages.getString("Error_in_input_parameters"),
            JOptionPane.ERROR_MESSAGE);
            optionsDirjComboBox.requestFocusInWindow();
            return null;
        }

        // test String input file
        if (optionsComboBox.getSelectedItem() == null
        || ((String)optionsComboBox.getSelectedItem())
            .trim().length() == 0) {
            JOptionPane.showMessageDialog(frame,
            messages.getString("You_must_provide_a_file_name."),
            messages.getString("Error_in_input_parameters"),
            JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return null;
        }

        inputFileName = (String)optionsDirjComboBox.getSelectedItem()
            +System.getProperty("file.separator")
            +(String)optionsComboBox.getSelectedItem()+".iter";

        // test existence of input file
        if (!new File(inputFileName).exists()) {
            JOptionPane.showMessageDialog(frame,
            messages.getString("You_must_provide_a_file_name."),
            messages.getString("Error_in_input_parameters"),
            JOptionPane.ERROR_MESSAGE);
            optionsComboBox.requestFocusInWindow();
            return null;
        }

        return inputFileName;
    }

    /**
     * File chooser filter on extension for use with JFileChooser.
     */
    private class FileChooserFilterExtension
        extends javax.swing.filechooser.FileFilter {

        private String endWith;
        private String description;

        public FileChooserFilterExtension(String _endWith,
                                          String _description) {
            endWith = _endWith;
            description = _description;
        }

        public boolean accept(File f) {
            return (f.toString().endsWith(endWith) || f.isDirectory());
        }

        public String getDescription() {
            return description;
        }
    }

    /**
     * Read the current directory and synchronise it with the files
     * ComboBoxes.
     */
    private void updateFilesList() {

        File[] iterationDir;
        File[] iterationFiles;
        File[] resultsFiles;
        File[] termsFiles;
        File currentDir;
        File currentOptionsDir;
        String selectedItem, selectedItem2, actionCommand, actionCommand2;

        // temporary disable triggering of events on optionsComboBox
        actionCommand = optionsComboBox.getActionCommand();
        optionsComboBox.setActionCommand("Programmatically");
        actionCommand2 = optionsDirjComboBox.getActionCommand();
        optionsDirjComboBox.setActionCommand("Programmatically");

        // search iteration directories in the execution directory
        currentDir = new File(System.getProperty("user.dir"));
        iterationDir = currentDir.listFiles();
        // remove the directories that are not iteration directories
        for (int i = iterationDir.length-1; i >= 0; i--) {
            if (!iterationDir[i].isDirectory()
              || iterationDir[i].getName().equals("java")
              || iterationDir[i].getName().equals("doc")
              || iterationDir[i].getName().equals("class")
              || !(new File (iterationDir[i]
                +System.getProperty("file.separator")+".info")).exists()) {
                iterationDir[i] = null;
            }
        }
        // add iteration directories to interface
        selectedItem = (String)optionsDirjComboBox.getSelectedItem();
        optionsDirjComboBox.removeAllItems();
        if (iterationDir != null) {
            for (int i = 0; i < iterationDir.length; i++) {
                if (iterationDir[i] != null) {
                    optionsDirjComboBox.addItem(iterationDir[i].getName());
                }
            }
        }
        optionsDirjComboBox.setSelectedItem(selectedItem);

        // search files in the execution directory and options directory
        iterationFiles = null;
        if (optionsDirjComboBox.getSelectedItem() != null
        && ((String)optionsDirjComboBox.getSelectedItem())
            .trim().length() > 0) {
            currentOptionsDir = new File(currentDir
                +System.getProperty("file.separator")
                +(String)optionsDirjComboBox.getSelectedItem());
            iterationFiles = currentOptionsDir
                .listFiles(new FileFilterExtension(".iter"));
        }
        resultsFiles =
            currentDir.listFiles(new FileFilterExtension(".results"));
        termsFiles =
            currentDir.listFiles(new FileFilterExtension("terms"));

        // add options files to interface
        selectedItem = (String)optionsComboBox.getSelectedItem();
        optionsComboBox.removeAllItems();
        if (iterationFiles != null) {
            for (int i = 0; i < iterationFiles.length; i++) {
                optionsComboBox.addItem(
                    removeFileExtension(iterationFiles[i].getName()));
            }
        }
        optionsComboBox.setSelectedItem(selectedItem);

        // add results files to interface
        selectedItem = (String)outputFileComboBox.getSelectedItem();
        outputFileComboBox.removeAllItems();
        if (resultsFiles != null) {
            for (int i = 0; i < resultsFiles.length; i++) {
                outputFileComboBox.addItem(resultsFiles[i].getName());
            }
        }
        outputFileComboBox.setSelectedItem(selectedItem);

        // add terms files to preprocessing panel
        selectedItem = (String)acceptListjComboBox.getSelectedItem();
        selectedItem2 = (String)rejectListjComboBox.getSelectedItem();
        acceptListjComboBox.removeAllItems();
        rejectListjComboBox.removeAllItems();
        if (termsFiles != null) {
            for (int i = 0; i < termsFiles.length; i++) {
                acceptListjComboBox.addItem(termsFiles[i].getName());
                rejectListjComboBox.addItem(termsFiles[i].getName());
            }
        }
        acceptListjComboBox.setSelectedItem(selectedItem);
        rejectListjComboBox.setSelectedItem(selectedItem2);

        // events can already occur on optionsComboBox
        optionsComboBox.setActionCommand(actionCommand);
        optionsDirjComboBox.setActionCommand(actionCommand2);
    }

    /**
     * Same class as DefaultTableModel but with all cells uneditable.
     *
     * @see javax.swing.table.DefaultTableModel
     */
    private class UneditableDefaultTableModel extends DefaultTableModel {

        /**
         * Return always false.
         *
         * @return false.
         */
        public boolean isCellEditable(int rowIndex, int columnIndex) {
            return false;
        }
    }

    /**
     * Listener of the changes in selection of results table.
     * Prevent the user to select rows where no Occurrence are present.
     */
    private class ListSelectionListenerResultsTable
        implements ListSelectionListener {

        private ListSelectionModel lsm;
        private JTable table;

        public ListSelectionListenerResultsTable(JTable _table) {
            table = _table;
            lsm = _table.getSelectionModel();
        }

        /**
         * Unselect all cells that don't contain an Occurrence object.
         */
        public void valueChanged(ListSelectionEvent e) {
            int row = 0;

            if (table.getRowCount() <= 0 || table.getColumnCount() <= 0) {
                return;
            }

            // for each selected row in the table
            for (int i = 0; i < table.getSelectedRowCount(); i++) {

                row = (table.getSelectedRows())[i];
                if ((table.getModel().getValueAt(row, 0) == null)
                    || !(table.getModel().getValueAt(row, 0)
                         instanceof Occurrence)) {
                    lsm.removeSelectionInterval(row, row);
                }
            }
        }
    }

    /**
     * Cell renderer used for results table.
     * <p>
     * The same as the DefaultTableCellRenderer but moreover change the
     * background color according to the status of the cell. Moreover
     * it displays the Occurrence according to the display options
     * checked in the display menu.
     */
    private class ResultsCellRenderer extends DefaultTableCellRenderer {

        String separator;
        DecimalFormat form;

        public Component getTableCellRendererComponent(JTable table,
                                                       Object value,
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int row,
                                                       int column) {

            // get the result of the standard implementation of this method
            Component cr =
                super.getTableCellRendererComponent(table, value,
                                                    isSelected,
                                                    hasFocus, row, column);

            // change the background color of the cell
            // according to the status of the cell
            if (isSelected) {
                // keep the default behavior

            } else if (rowsStatus[row] == ACCEPTED) {
                ((DefaultTableCellRenderer)cr).setBackground(Color.GREEN);

            } else if (rowsStatus[row] == REJECTED) {
                ((DefaultTableCellRenderer)cr).setBackground(Color.RED);

            } else if (rowsStatus[row] == UNDEFINED) {
                ((DefaultTableCellRenderer)cr).setBackground(Color.WHITE);

            } else {
                ((DefaultTableCellRenderer)cr).setBackground(Color.WHITE);
            }

            return cr;
        }

        protected void setValue(Object value) {

            Occurrence occ;
            String s;
            StringBuffer strb;
            int formatNbInt, formatNbDec;

            if (separator == null) {
                separator = separatorjTextField.getText();
            }
            if (form == null) {
                formatNbInt =
                    ((Integer)nbIntegersSpinner.getValue()).intValue();
                formatNbDec =
                    ((Integer)nbDecimalsSpinner.getValue()).intValue();
                strb = new StringBuffer(formatNbInt+1+formatNbDec);
                for (int i = 0; i < formatNbInt; i++) { strb.append("0"); }
                strb.append(".");
                for (int i = 0; i < formatNbDec; i++) { strb.append("0"); }
                form = new DecimalFormat(strb.toString());
            }

            if (value == null) {
                setText("");

            } else if (value instanceof Occurrence) {
                occ = (Occurrence)value;
                s = "";

                if (occ.getMeasure() != Double.NaN
                && displayMeasuresjCheckBoxMenuItem.isSelected()) {
                    s += form.format(occ.getMeasure())+" ";
                }
                if (displayTagsjCheckBoxMenuItem.isSelected()) {
                    s += (occ.is2Units()) ?
                        occ.getUnit1()+occ.getUnit2()
                      : occ.getUnit1()+occ.getUnit2()+occ.getUnit3();
                } else {
                    s += (occ.is2Units()) ?
                        occ.getWord(occ.getUnit1(), separator)+" "
                       +occ.getWord(occ.getUnit2(), separator)
                      : occ.getWord(occ.getUnit1(), separator)+" "+
                        occ.getWord(occ.getUnit2(), separator)+" "+
                        occ.getWord(occ.getUnit3(), separator);
                }
                if (occ.getNamesOfText().size() > 0
                && displayOccjCheckBoxMenuItem.isSelected()) {
                    s += " ("+occ.getOcc()+", "+
                        occ.getNamesOfText().size()+")";
                }
                if (occ.getTypeName() != null) {
                    s += " ["+occ.getTypeName()+"]";
                }
                setText(s);

            } else {
                setText(value.toString());
            }
        }
    }

    /**
     * Cell renderer used for terms tables.
     * <p>
     * The same as the DefaultTableCellRenderer but moreover
     * it displays the Occurrence according to the display options
     * checked in the display menu.
     */
    private class TermsCellRenderer extends DefaultTableCellRenderer {

        String separator;
        DecimalFormat form;

        protected void setValue(Object value) {

            Occurrence occ;
            String s;
            StringBuffer strb;
            int formatNbInt, formatNbDec;

            if (separator == null) {
                separator = separatorjTextField.getText();
            }
            if (form == null) {
                formatNbInt =
                    ((Integer)nbIntegersSpinner.getValue()).intValue();
                formatNbDec =
                    ((Integer)nbDecimalsSpinner.getValue()).intValue();
                strb = new StringBuffer(formatNbInt+1+formatNbDec);
                for (int i = 0; i < formatNbInt; i++) { strb.append("0"); }
                strb.append(".");
                for (int i = 0; i < formatNbDec; i++) { strb.append("0"); }
                form = new DecimalFormat(strb.toString());
            }

            if (value == null) {
                setText("");

            } else if (value instanceof Occurrence) {
                occ = (Occurrence)value;
                s = "";

                if (occ.getMeasure() != Double.NaN
                && displayMeasuresjCheckBoxMenuItem.isSelected()) {
                    s += form.format(occ.getMeasure())+" ";
                }
                if (displayTagsjCheckBoxMenuItem.isSelected()) {
                    s += (occ.is2Units()) ?
                        occ.getUnit1()+occ.getUnit2()
                      : occ.getUnit1()+occ.getUnit2()+occ.getUnit3();
                } else {
                    s += (occ.is2Units()) ?
                        occ.getWord(occ.getUnit1(), separator)+" "
                       +occ.getWord(occ.getUnit2(), separator)
                      : occ.getWord(occ.getUnit1(), separator)+" "+
                        occ.getWord(occ.getUnit2(), separator)+" "+
                        occ.getWord(occ.getUnit3(), separator);
                }
                if (occ.getNamesOfText().size() > 0
                && displayOccjCheckBoxMenuItem.isSelected()) {
                    s += " ("+occ.getOcc()+", "+
                        occ.getNamesOfText().size()+")";
                }
                if (occ.getTypeName() != null) {
                    s += " ["+occ.getTypeName()+"]";
                }
                setText(s);

            } else {
                setText(value.toString());
            }
        }
    }

    /**
     * Table model used for tables of the Expressions panel.
     * <p>
     * The same as the DefaultTableModel but moreover disabled the
     * rows that are inferior to the numbers contain in rowsMaxDisabled.
     */
    private class ExpressionsTableModel extends DefaultTableModel {

        /** Contain the last lines of each elements that are disabled
         * in the expressions panel tables */
        private int [] rowsMaxDisabled;

        public ExpressionsTableModel(int rowCount, int columnCount) {
            super(rowCount, columnCount);
            rowsMaxDisabled = new int [this.getColumnCount()];
            // clear the status of the expressions panel tables cells
            Arrays.fill(rowsMaxDisabled, -1);
        }
        
        public boolean isCellEditable(int rowIndex, int columnIndex) {
            return (rowIndex > rowsMaxDisabled[columnIndex]);
        }

        public void setRowMaxDisabled(int columnIndex, int rowMax) {
            rowsMaxDisabled[columnIndex] = rowMax;
        }

        public int getRowMaxDisabled(int columnIndex) {
            return rowsMaxDisabled[columnIndex];
        }
    }

    /**
     * Cell renderer used for tables of the Expressions panel.
     * <p>
     * The same as the DefaultTableCellRenderer but moreover grayed
     * the foreground color of the uneditables cells.
     */
    private class ExpressionsCellRenderer extends DefaultTableCellRenderer {

        public Component getTableCellRendererComponent(JTable table,
                                                       Object value,
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int row,
                                                       int column) {

            // get the result of the standard implementation of this method
            Component cr =
                super.getTableCellRendererComponent(table, value,
                                                    isSelected,
                                                    hasFocus, row, column);
            if (!table.isCellEditable(row, column)) {
                ((DefaultTableCellRenderer)cr).setForeground(Color.GRAY);

            } else {
                ((DefaultTableCellRenderer)cr).setForeground(Color.BLACK);
            }
            return cr;
        }
    }

    /**
     * Cell renderer for the iteration list. Show the iteration plus the
     * last relation extracted.
     */
    private class IterationCellRenderer extends JLabel
        implements ListCellRenderer {

        public IterationCellRenderer() {
            setOpaque(true);
        }

        public Component getListCellRendererComponent(JList list,
            Object value, int index, boolean isSelected,
            boolean cellHasFocus) {

            BufferedReader br;
            String fileName, line, previousLine, history;

            if (value == null) { return this; }

            setText(value.toString());
            setBackground(isSelected ? Color.red : Color.white);
            setForeground(isSelected ? Color.white : Color.black);

            line = ""; previousLine = ""; history = "";
            fileName = (String)optionsDirjComboBox.getSelectedItem()
                +System.getProperty("file.separator")+value.toString()
                +".hist";

            try {
                br = new BufferedReader(new FileReader(fileName));
                // if the .hist file is empty
                if (!br.ready()) { return this; }
                // read all lines of the .hist file
                line = br.readLine();
                while (line != null) {
                    previousLine = line;
                    history += line+", ";
                    line = br.readLine();
                }
                if (history.length() > 0) {
                    // remove the separator at the end of the history
                    history = history.substring(0, history.length()-2);
                }

            } catch (IOException ioe) {
                JOptionPane.showMessageDialog(frame,
                    ioe.getLocalizedMessage(),
                    messages.getString("Error_in_reading_file"),
                    JOptionPane.ERROR_MESSAGE);
                return this;
            }

            setText(value.toString()+" "+previousLine);
            setToolTipText(history);
            return this;
        }
    }


    /*-*************
     * Main method *
     ***************/

    /**
     * Launch the user interface.
     *
     * @param args no parameters exists.
     */
    public static void main(String args[]) {

        new ExtractTermsGUI().show();
    }


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JMenuItem aboutMenuItem;
    private javax.swing.JButton acceptListjButton;
    private javax.swing.JComboBox acceptListjComboBox;
    private javax.swing.JMenuItem acceptTypejMenuItem1;
    private javax.swing.JMenuItem acceptTypejMenuItem10;
    private javax.swing.JMenuItem acceptTypejMenuItem11;
    private javax.swing.JMenuItem acceptTypejMenuItem12;
    private javax.swing.JMenuItem acceptTypejMenuItem13;
    private javax.swing.JMenuItem acceptTypejMenuItem14;
    private javax.swing.JMenuItem acceptTypejMenuItem15;
    private javax.swing.JMenuItem acceptTypejMenuItem16;
    private javax.swing.JMenuItem acceptTypejMenuItem17;
    private javax.swing.JMenuItem acceptTypejMenuItem18;
    private javax.swing.JMenuItem acceptTypejMenuItem19;
    private javax.swing.JMenuItem acceptTypejMenuItem2;
    private javax.swing.JMenuItem acceptTypejMenuItem20;
    private javax.swing.JMenuItem acceptTypejMenuItem21;
    private javax.swing.JMenuItem acceptTypejMenuItem22;
    private javax.swing.JMenuItem acceptTypejMenuItem23;
    private javax.swing.JMenuItem acceptTypejMenuItem24;
    private javax.swing.JMenuItem acceptTypejMenuItem25;
    private javax.swing.JMenuItem acceptTypejMenuItem26;
    private javax.swing.JMenuItem acceptTypejMenuItem27;
    private javax.swing.JMenuItem acceptTypejMenuItem28;
    private javax.swing.JMenuItem acceptTypejMenuItem29;
    private javax.swing.JMenuItem acceptTypejMenuItem3;
    private javax.swing.JMenuItem acceptTypejMenuItem30;
    private javax.swing.JMenuItem acceptTypejMenuItem31;
    private javax.swing.JMenuItem acceptTypejMenuItem32;
    private javax.swing.JMenuItem acceptTypejMenuItem33;
    private javax.swing.JMenuItem acceptTypejMenuItem34;
    private javax.swing.JMenuItem acceptTypejMenuItem35;
    private javax.swing.JMenuItem acceptTypejMenuItem36;
    private javax.swing.JMenuItem acceptTypejMenuItem37;
    private javax.swing.JMenuItem acceptTypejMenuItem38;
    private javax.swing.JMenuItem acceptTypejMenuItem39;
    private javax.swing.JMenuItem acceptTypejMenuItem4;
    private javax.swing.JMenuItem acceptTypejMenuItem40;
    private javax.swing.JMenuItem acceptTypejMenuItem41;
    private javax.swing.JMenuItem acceptTypejMenuItem42;
    private javax.swing.JMenuItem acceptTypejMenuItem43;
    private javax.swing.JMenuItem acceptTypejMenuItem44;
    private javax.swing.JMenuItem acceptTypejMenuItem45;
    private javax.swing.JMenuItem acceptTypejMenuItem46;
    private javax.swing.JMenuItem acceptTypejMenuItem47;
    private javax.swing.JMenuItem acceptTypejMenuItem48;
    private javax.swing.JMenuItem acceptTypejMenuItem49;
    private javax.swing.JMenuItem acceptTypejMenuItem5;
    private javax.swing.JMenuItem acceptTypejMenuItem50;
    private javax.swing.JMenuItem acceptTypejMenuItem6;
    private javax.swing.JMenuItem acceptTypejMenuItem7;
    private javax.swing.JMenuItem acceptTypejMenuItem8;
    private javax.swing.JMenuItem acceptTypejMenuItem9;
    private javax.swing.JComboBox acceptedTermsIterationjComboBox;
    private javax.swing.JComboBox acceptedTermsRelationjComboBox;
    private javax.swing.JComboBox acceptedTermsTypesjComboBox;
    private javax.swing.JScrollPane acceptedTermsjScrollPane;
    private javax.swing.JTable acceptedTermsjTable;
    private javax.swing.JMenu acceptjMenu;
    private javax.swing.JList alreadyFormedTermsjList;
    private javax.swing.JScrollPane alreadyFormedTermsjScrollPane;
    private javax.swing.JRadioButtonMenuItem automaticjRadioButtonMenuItem;
    private javax.swing.JMenuItem contentsMenuItem;
    private javax.swing.JScrollPane contextjScrollPane;
    private javax.swing.JTextPane contextjTextPane;
    private javax.swing.JPanel contextualRulesjPanel;
    private javax.swing.JMenuItem contexualHelpjMenuItem;
    private javax.swing.JCheckBoxMenuItem displayExpertModejCheckBoxMenuItem;
    private javax.swing.JCheckBoxMenuItem displayMeasuresjCheckBoxMenuItem;
    private javax.swing.JCheckBoxMenuItem displayOccjCheckBoxMenuItem;
    private javax.swing.JCheckBoxMenuItem displayTagsjCheckBoxMenuItem;
    private javax.swing.JMenu displayjMenu;
    private javax.swing.JButton editTagsAlreadyFormedTermsjButton;
    private javax.swing.JRadioButtonMenuItem englishjRadioButtonMenuItem;
    private javax.swing.JMenuItem exitMenuItem;
    private javax.swing.JButton extractButton;
    private javax.swing.JSplitPane extraction2jSplitPane;
    private javax.swing.JProgressBar extractionjProgressBar;
    private javax.swing.JSplitPane extractionjSplitPane;
    private javax.swing.JMenuItem firstStartjMenuItem;
    private javax.swing.JScrollPane formulasjScrollPane;
    private javax.swing.JTable formulasjTable;
    private javax.swing.JRadioButtonMenuItem francaisjRadioButtonMenuItem;
    private javax.swing.JMenu helpMenu;
    private javax.swing.JButton helpTagsAlreadyFormedTermsjButton;
    private javax.swing.JCheckBoxMenuItem helpUnderCursorjCheckBoxMenuItem;
    private javax.swing.JLabel jLabel10;
    private javax.swing.JLabel jLabel11;
    private javax.swing.JLabel jLabel12;
    private javax.swing.JLabel jLabel13;
    private javax.swing.JLabel jLabel14;
    private javax.swing.JLabel jLabel15;
    private javax.swing.JLabel jLabel16;
    private javax.swing.JLabel jLabel17;
    private javax.swing.JLabel jLabel18;
    private javax.swing.JLabel jLabel19;
    private javax.swing.JLabel jLabel2;
    private javax.swing.JLabel jLabel20;
    private javax.swing.JLabel jLabel21;
    private javax.swing.JLabel jLabel22;
    private javax.swing.JLabel jLabel23;
    private javax.swing.JLabel jLabel24;
    private javax.swing.JLabel jLabel25;
    private javax.swing.JLabel jLabel26;
    private javax.swing.JLabel jLabel3;
    private javax.swing.JLabel jLabel4;
    private javax.swing.JLabel jLabel5;
    private javax.swing.JLabel jLabel6;
    private javax.swing.JLabel jLabel7;
    private javax.swing.JLabel jLabel8;
    private javax.swing.JLabel jLabel9;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JMenu languagejMenu;
    private javax.swing.JList localRules2UnitsjList;
    private javax.swing.JScrollPane localRules2UnitsjScrollPane;
    private javax.swing.JList localRules3UnitsjList;
    private javax.swing.JScrollPane localRules3UnitsjScrollPane;
    private javax.swing.JPanel localRulesActionsjPanel;
    private javax.swing.JPanel localRulesInputjPanel;
    private javax.swing.JPanel localRulesListsjPanel;
    private javax.swing.JPanel localRulesjPanel;
    private javax.swing.JButton lrAdd2UnitsjButton;
    private javax.swing.JButton lrAdd3UnitsjButton;
    private javax.swing.JButton lrAndjButton;
    private javax.swing.JButton lrDecomposejButton;
    private javax.swing.JButton lrDownjButton;
    private javax.swing.JButton lrNotjButton;
    private javax.swing.JButton lrOrjButton;
    private javax.swing.JComboBox lrPartjComboBox;
    private javax.swing.JButton lrRemovejButton;
    private javax.swing.JTextField lrStringjTextField;
    private javax.swing.JComboBox lrTypejComboBox;
    private javax.swing.JComboBox lrUnitjComboBox;
    private javax.swing.JButton lrUpjButton;
    private javax.swing.JTabbedPane mainjTabbedPane;
    private javax.swing.JComboBox measuresjComboBox;
    private javax.swing.JMenuBar menuBar;
    private javax.swing.JComboBox nameOfTextjComboBox;
    private javax.swing.JSpinner nbDecimalsSpinner;
    private javax.swing.JSpinner nbIntegersSpinner;
    private javax.swing.JCheckBox occFirstjCheckBox;
    private javax.swing.JComboBox optionsComboBox;
    private javax.swing.JComboBox optionsDirjComboBox;
    private javax.swing.JPanel optionsjPanel;
    private javax.swing.JButton outputFileChooserjButton;
    private javax.swing.JComboBox outputFileComboBox;
    private javax.swing.JPanel preProcessing11jPanel;
    private javax.swing.JPanel preProcessing12jPanel;
    private javax.swing.JPanel preProcessing2jPanel;
    private javax.swing.JPanel preProcessingjPanel;
    private javax.swing.JButton precedentTermsAcceptjButton;
    private javax.swing.JButton precedentTermsCanceljButton;
    private javax.swing.JButton precedentTermsRejectjButton;
    private javax.swing.JButton precedentTermsUndefinedTermsjButton;
    private javax.swing.JButton precedentTermsjButton;
    private javax.swing.JDialog precedentTermsjDialog;
    private javax.swing.JSpinner precedentTermsjSpinner;
    private javax.swing.JMenu programMenu;
    private javax.swing.JScrollPane regex2jScrollPane;
    private javax.swing.JTable regex2jTable;
    private javax.swing.JPanel regexFormulasTablesjPanel;
    private javax.swing.JPanel regexFormulasUtilsjPanel;
    private javax.swing.JPanel regexFormulasjPanel;
    private javax.swing.JButton regexHelpTagsjButton;
    private javax.swing.JButton regexHelpjButton;
    private javax.swing.JToggleButton regexTestjToggleButton;
    private javax.swing.JScrollPane regexjScrollPane;
    private javax.swing.JTable regexjTable;
    private javax.swing.JButton rejectListjButton;
    private javax.swing.JComboBox rejectListjComboBox;
    private javax.swing.JMenuItem rejectTypejMenuItem1;
    private javax.swing.JMenuItem rejectTypejMenuItem10;
    private javax.swing.JMenuItem rejectTypejMenuItem11;
    private javax.swing.JMenuItem rejectTypejMenuItem12;
    private javax.swing.JMenuItem rejectTypejMenuItem13;
    private javax.swing.JMenuItem rejectTypejMenuItem14;
    private javax.swing.JMenuItem rejectTypejMenuItem15;
    private javax.swing.JMenuItem rejectTypejMenuItem16;
    private javax.swing.JMenuItem rejectTypejMenuItem17;
    private javax.swing.JMenuItem rejectTypejMenuItem18;
    private javax.swing.JMenuItem rejectTypejMenuItem19;
    private javax.swing.JMenuItem rejectTypejMenuItem2;
    private javax.swing.JMenuItem rejectTypejMenuItem20;
    private javax.swing.JMenuItem rejectTypejMenuItem21;
    private javax.swing.JMenuItem rejectTypejMenuItem22;
    private javax.swing.JMenuItem rejectTypejMenuItem23;
    private javax.swing.JMenuItem rejectTypejMenuItem24;
    private javax.swing.JMenuItem rejectTypejMenuItem25;
    private javax.swing.JMenuItem rejectTypejMenuItem26;
    private javax.swing.JMenuItem rejectTypejMenuItem27;
    private javax.swing.JMenuItem rejectTypejMenuItem28;
    private javax.swing.JMenuItem rejectTypejMenuItem29;
    private javax.swing.JMenuItem rejectTypejMenuItem3;
    private javax.swing.JMenuItem rejectTypejMenuItem30;
    private javax.swing.JMenuItem rejectTypejMenuItem31;
    private javax.swing.JMenuItem rejectTypejMenuItem32;
    private javax.swing.JMenuItem rejectTypejMenuItem33;
    private javax.swing.JMenuItem rejectTypejMenuItem34;
    private javax.swing.JMenuItem rejectTypejMenuItem35;
    private javax.swing.JMenuItem rejectTypejMenuItem36;
    private javax.swing.JMenuItem rejectTypejMenuItem37;
    private javax.swing.JMenuItem rejectTypejMenuItem38;
    private javax.swing.JMenuItem rejectTypejMenuItem39;
    private javax.swing.JMenuItem rejectTypejMenuItem4;
    private javax.swing.JMenuItem rejectTypejMenuItem40;
    private javax.swing.JMenuItem rejectTypejMenuItem41;
    private javax.swing.JMenuItem rejectTypejMenuItem42;
    private javax.swing.JMenuItem rejectTypejMenuItem43;
    private javax.swing.JMenuItem rejectTypejMenuItem44;
    private javax.swing.JMenuItem rejectTypejMenuItem45;
    private javax.swing.JMenuItem rejectTypejMenuItem46;
    private javax.swing.JMenuItem rejectTypejMenuItem47;
    private javax.swing.JMenuItem rejectTypejMenuItem48;
    private javax.swing.JMenuItem rejectTypejMenuItem49;
    private javax.swing.JMenuItem rejectTypejMenuItem5;
    private javax.swing.JMenuItem rejectTypejMenuItem50;
    private javax.swing.JMenuItem rejectTypejMenuItem6;
    private javax.swing.JMenuItem rejectTypejMenuItem7;
    private javax.swing.JMenuItem rejectTypejMenuItem8;
    private javax.swing.JMenuItem rejectTypejMenuItem9;
    private javax.swing.JComboBox rejectedTermsIterationjComboBox;
    private javax.swing.JComboBox rejectedTermsRelationjComboBox;
    private javax.swing.JComboBox rejectedTermsTypesjComboBox;
    private javax.swing.JScrollPane rejectedTermsjScrollPane;
    private javax.swing.JTable rejectedTermsjTable;
    private javax.swing.JMenu rejectjMenu;
    private javax.swing.JButton removeAlreadyFormedTermsjButton;
    private javax.swing.JPopupMenu resultsjPopupMenu;
    private javax.swing.JScrollPane resultsjScrollPane;
    private javax.swing.JTable resultsjTable;
    private javax.swing.JPanel rulesjPanel;
    private javax.swing.JMenuItem saveResultsjMenuItem;
    private javax.swing.JButton savejButton;
    private javax.swing.JButton searchAlreadyFormedTermsjButton;
    private javax.swing.JMenuItem searchStringjMenuItem;
    private javax.swing.JMenuItem selectNewInputFilejMenuItem;
    private javax.swing.JComboBox sentPhrComboBox;
    private javax.swing.JTextField separatorjTextField;
    private javax.swing.JButton startAgainjButton;
    private javax.swing.JLabel statejLabel;
    private javax.swing.JPanel statejPanel;
    private javax.swing.JPanel termsButtonsjPanel;
    private javax.swing.JScrollPane termsContextjScrollPane;
    private javax.swing.JTextPane termsContextjTextPane;
    private javax.swing.JPanel termsTablesjPanel;
    private javax.swing.JScrollPane termsTypesjScrollPane;
    private javax.swing.JTable termsTypesjTable;
    private javax.swing.JPanel termsjPanel;
    private javax.swing.JSplitPane termsjSplitPane;
    private javax.swing.JCheckBox tfidfjCheckBox;
    private javax.swing.JComboBox threeUnitsComboBox;
    private javax.swing.JSpinner threeUnitsResultsSpinner;
    private javax.swing.JSpinner threeUnitsSpinner;
    private javax.swing.JComboBox twoUnitsComboBox;
    private javax.swing.JSpinner twoUnitsResultsSpinner;
    private javax.swing.JSpinner twoUnitsSpinner;
    private javax.swing.JPanel undefinedTermsButtonsjPanel;
    private javax.swing.JScrollPane undefinedTermsContextjScrollPane;
    private javax.swing.JTextPane undefinedTermsContextjTextPane;
    private javax.swing.JComboBox undefinedTermsIterationjComboBox;
    private javax.swing.JComboBox undefinedTermsRelationjComboBox;
    private javax.swing.JPanel undefinedTermsjPanel;
    private javax.swing.JScrollPane undefinedTermsjScrollPane;
    private javax.swing.JSplitPane undefinedTermsjSplitPane;
    private javax.swing.JTable undefinedTermsjTable;
    private javax.swing.JMenuItem undefinejMenuItem;
    private javax.swing.JComboBox unitComboBox;
    // End of variables declaration//GEN-END:variables
    
}
